package sim.app.traffic;
import sim.util.*;
import sim.engine.*;
import sim.field.network.*;
import sim.app.traffic.queuewealth.*;
import sim.app.traffic.rotation.*;
import sim.app.traffic.payment.*;
import sim.app.traffic.compensation.*;
import sim.app.traffic.car.*;
import ec.util.*;


/**
 * <code>Intersection</code> or its subclasses contain most elements of interest:
 * 
 * 
 * <dl>
 * <dt>Queue Groups</dt>
 * 		The queues in a GROUP gets the green light, all other queues get the red light.
 * 		Which of queues can send simultaneusly send cars into the intersection
 * 		(i.e. what maximal groups of queues can SHARE the green light)?
 *		<dd><ul>
 *		<li>all cars coming from the same street, regardless of destination (<code>OneQueueGroupIntersection</code>)
 * 		<li>all cars with non-conflicting source-destination streets (<code>OneLaneOutEdgeIntersection</code>).
 *  	</ul></dd>
 *  
 *  <dt>RotationPolicy: how do you switch between red and green lights</dt>
 *  	<dd><ul>
 *  	<li>round robin between the entire set of queue groups (<code>RoundRobin</code>)
 *  	<li>same as before, but skip a group if all queues in it are empty
 *  		(<code>NonEmptyRoundRobin</code>)
 *  	<li>the winner is the group of non-conflicitng queues with the largest "wealth"
 *  		(<code>WealthiestFirst</code>)
 *  	<li>the winner group is selected using roulette selection, and each
 *  		group has a chance proportional to its wealth (<code>WealthProportionate</code>).
 *  		Clearly, this does't play nice with QueueWealthMetrics that allow negative values.
 *  	</ul></dd>
 *  
 * 	<dt>Wealth policy</dt>  
 *  The wealth of a queue is encoded in a <code>QueueWealthMetric</code>:
 *		<dd><ul>
 *		<li>sum of credit wealths for the winnig party (<code>SumOfCarWealths</code>)
 *		<li>sum of bids (<code>SumOfCarBids</code>)
 *		<li>number of cars (<code>CountingCars</code>)
 *		<li>etc.
 * 		</ul></dd>
 * 
 * 	<dt>How much you pay for Green (<code>PaymentPolicy</code>)</dt>
 * 		<dd><ul>
 * 		<li>fixed amount
 * 		<li>collective bidding: every car pays how much it bid.
 * 		</ul></dd>
 * 
 * 
 *	<dt>Who Pays for Green</dt>
 * 		<dd><ul>
 * 		<li>all the cars in the winning queues
 * 		(as specified by the <code>PaymentPolicy</code> in <code>Intersection.greenCollector</code>; can be null).
 * 		<li>the cars in the winning queus that actually made it through the intersection 
 * 		(as specified by the <code>PaymentPolicy</code> in <code>Intersection.passageCollector</code>; can be null)
 * 		</ul></dd>
 * 	<p>The first policy specifies that the winners are buying passage, 
 * 	the second specifies the winners are buys phase switching/extensions.
 * 	It makes sense to buy passage: you only pay for what you use.
 * 	It also makes sense to buy green time even if you don't get to use it:
 * 	before you can cross the intersection, you need the cars in front of you
 * 	in the queue to cross.
 * 
 * 	<p>Also, note that there's a difference between phase switching and phase extension.
 * They have different costs: it takes longer to go through when you're stopped 
 * than when you're at cruising speed. Plus, there is a <code>Intersection.takeOffTime</code>
 * time loss a car suffers waiting for the car in front of it to take off (and reach nominal speed).
 * Therefore, <code>Intersection.carsPassingPerFreshGreenPhasePerQueue</code> cars
 * can cross an intersection when the light switches from red to green, and 
 * <code>Intersection.carsPassingPerOldGreenPhasePerQueue</code> when the green light phase is extended.
 * 
 * 
 * <dt>What do you Earn at Red (i.e. when you lose) (<code>CompensationPolicy</code>)</dt>
 * 		<dd><ul>
 * 		<li>fixed amount (<code>FixedCompensation</code>)
 * 		<li>divide the winners' payments equally among losers (<code>ZeroSumEqualCompensation</code>)
 * 		<li>divide the winners' payments to the losers, proportional to each losing bid.
 * 		(<code>ZeroSumBidProportionalCompensation</code>)
 * 		<li>etc.
 * 		</ul></dd>
 * </dl>
 *
 * <p>Intersection knows about the "getting payed at red" policy, TrafficQueue knows about the "paying for green" policy
 * 
 * @author Gabriel Balan
 */
public abstract class Intersection implements Parametrizable, Steppable
{
	public static final String P_GREEN_COLLECTOR = "pay4green-policy";
	public static final String P_PASSAGE_COLLECTOR = "pay4passage-policy";
	public static final String P_COMPENSATION_POLICY = "compensation-policy";
	public static final String P_ROTATION_POLICY = "rotation-policy";
	public static final String P_QWEALTH_METRIC = "queue-wealth-metric";
	
	
	public static final String P_PHASE_OFFSET = "phase-offset";
	public static final String P_PHASE_DURATION = "phase-duration";
	public static final String P_CARS_PER_FRESH_GREEN = "cars/fresh-green/queue";
	public static final String P_CARS_PER_OLD_GREEN = "cars/old-green/queue";	
	public static final String P_DT_BETWEEN_CAR_TAKEOFFS = "takeoff-time";
	
	public static final String P_FIX_NOCNOC_BUG = "fix-nocnoc-bug";
	public static final String P_DONT_COLLECT_UNLESS_LOSERS = "dont-collect-unless-losers";
	
	
	public Network.IndexOutIn ioi; //to avoid hash-table look-ups
	// I don't want to use an Intersection in multiple Networks, so this is fine.
	// Of course, it breaks cloneGraph().


	/** 
	 * this refers to the following bug: when all cars have 0 credits, the losers don't get 
	 * compensated under non-zero compensation policies
	 **/
	boolean fixNoCNoCBug = false;

	QueueWealthMetric qWealthMetric;
	public QueueWealthMetric getQueueWealthMetric(){return qWealthMetric;}
	
	//each Intersection implementation is responsible for allocating these arrays and 
	// setting the id field of the Queues
	public TrafficQueue[][] groups;
	//each intersection decides the number of queues and handles the linearization
	//I choose this solution to avoid the overhead of the Iterator
	public TrafficQueue[] queues;
	public boolean[] winnersMask;	
	public float[] groupWealths;
	
	PaymentPolicy greenCollector;
	PaymentPolicy passageCollector;
	CompensationPolicy compensationPolicy;
	double sourceProb;
	double destinationProb;	
	protected TrafficSim tSimState;
	
	//inspector clues
	public TrafficQueue[] getQueues(){return queues;}
	
	
	//default behavior specified by these constants. 
	//Overwrite start and reschedule to model for instance 
	//phase extensions smaller than regular phaseDuration 
	double phaseOffset = 0;
	public double phaseDuration;
	public void start()
	{
		tSimState.schedule.scheduleOnce(phaseOffset, TrafficSim.INTERSECTION_ORDERING, this); 
		lastWinningGroupIndex = 0;//so that the system behaves the same when you press start the second time
		numberOfPatternChanges = 0;
		for(int i=queues.length-1;i>=0;i--)
			queues[i].start();
		revenueThisPhase = 0;
	}
	
	public void resetStats()
	{
		for(int i=queues.length-1;i>=0;i--)
			queues[i].resetStats();
		
	}
	
	//TODO i should let others wake me up sooner if less than n were let through
	public void reschedule(){tSimState.schedule.scheduleOnceIn(phaseDuration, this, TrafficSim.INTERSECTION_ORDERING);}
	protected int carsPassingPerFreshGreenPhasePerQueue;
	protected int carsPassingPerOldGreenPhasePerQueue;
	protected double takeOffTime;//this includes the wave effect too.

	RotationPolicy rotationPolicy;//round robin or Wealthiest-first
	public void netSetup(Network.IndexOutIn ioi)
	{
		this.ioi = ioi;
	}
	
	public void setup(ParamSimState state, Parameter base)
	{
		tSimState = (TrafficSim)state;
		if(ioi==null)
		{//this is a prototype intersection, the default intersection of the city
			Parameter paramThis = base;
			Parameter paramDef = defaultBase();
			Parameter p, pd;

			greenCollector = null;
			if(state.parameters.exists(p = paramThis.push(P_GREEN_COLLECTOR), pd=paramDef.push(P_GREEN_COLLECTOR)))
			{
				greenCollector = (PaymentPolicy)state.parameters.getInstanceForParameter(p,pd,PaymentPolicy.class);
				greenCollector.setup(state, p);
			}

								
			passageCollector = null;
			if(state.parameters.exists(p = paramThis.push(P_PASSAGE_COLLECTOR), pd=paramDef.push(P_PASSAGE_COLLECTOR)))													
			{
				passageCollector = (PaymentPolicy)state.parameters.getInstanceForParameter(	p,pd,PaymentPolicy.class);
				passageCollector.setup(state, p);
			}

			compensationPolicy = null;
			if(state.parameters.exists(p=paramThis.push(P_COMPENSATION_POLICY), pd=paramDef.push(P_COMPENSATION_POLICY)))
			{
				compensationPolicy = (CompensationPolicy)state.parameters.getInstanceForParameter(p, pd, CompensationPolicy.class);
				compensationPolicy.setup(state, p);
			}

			rotationPolicy = (RotationPolicy)state.parameters.getInstanceForParameter(	p=paramThis.push(P_ROTATION_POLICY), 
																						paramDef.push(P_ROTATION_POLICY), 
																						RotationPolicy.class);

			phaseOffset = state.parameters.getDoubleWithDefault(paramThis.push(P_PHASE_OFFSET), 
																paramDef.push(P_PHASE_OFFSET), phaseOffset);
			phaseDuration = state.parameters.getDouble(base.push(P_PHASE_DURATION), paramDef.push(P_PHASE_DURATION), 0);
			
			carsPassingPerFreshGreenPhasePerQueue = state.parameters.getInt( paramThis.push(P_CARS_PER_FRESH_GREEN), 
																paramDef.push(P_CARS_PER_FRESH_GREEN),1);
																//at least one car

			carsPassingPerOldGreenPhasePerQueue = state.parameters.getInt( paramThis.push(P_CARS_PER_OLD_GREEN), 
																paramDef.push(P_CARS_PER_OLD_GREEN),carsPassingPerFreshGreenPhasePerQueue);

			takeOffTime = state.parameters.getDouble( p = paramThis.push(P_DT_BETWEEN_CAR_TAKEOFFS), 
														paramDef.push(P_DT_BETWEEN_CAR_TAKEOFFS), 0);
			if(takeOffTime<=0)	throw new RuntimeException(p+" must be >0");
													
			qWealthMetric = (QueueWealthMetric)state.parameters.getInstanceForParameter(p=paramThis.push(P_QWEALTH_METRIC), 
																	 paramDef.push(P_QWEALTH_METRIC), 
																	 QueueWealthMetric.class);
			qWealthMetric.setup(state, p);
			
			fixNoCNoCBug = state.parameters.getBoolean( paramThis.push(P_FIX_NOCNOC_BUG),
														paramDef.push(P_FIX_NOCNOC_BUG),
														fixNoCNoCBug);
		}
		else
		{//this is not a prototypeIntersection, it's actually overridding the prototypeIntersection 
		 // that's why no defaults are considered.
			String indexStr = ioi.index+"";
			Parameter paramThis = base.push(indexStr);
			Parameter p;
			
			if(state.parameters.exists(p = paramThis.push(P_ROTATION_POLICY)))
				rotationPolicy = (RotationPolicy)state.parameters.getInstanceForParameter(p, null, RotationPolicy.class);

			if(state.parameters.exists(p = paramThis.push(P_GREEN_COLLECTOR)))
			{
				greenCollector = (PaymentPolicy)state.parameters.getInstanceForParameter(p, null, PaymentPolicy.class);
				greenCollector.setup(state, p);
			}
			if(state.parameters.exists(p = paramThis.push(P_PASSAGE_COLLECTOR)))
			{
				passageCollector = (PaymentPolicy)state.parameters.getInstanceForParameter(p, null, PaymentPolicy.class); 
				passageCollector.setup(state, p);
			}
			
			if(state.parameters.exists(p = paramThis.push(P_COMPENSATION_POLICY)))
			{
				compensationPolicy = (CompensationPolicy)state.parameters.getInstanceForParameter(p, null, CompensationPolicy.class);
				compensationPolicy.setup(state, p);
			}
			
			if(state.parameters.exists(p=paramThis.push(P_PHASE_OFFSET)))
				phaseOffset = state.parameters.getDouble(p, null, 0);
				
			if(state.parameters.exists(p=paramThis.push(P_PHASE_DURATION)))
				phaseDuration = state.parameters.getDouble(p, null, 0);
				
			if(state.parameters.exists(p=paramThis.push(P_CARS_PER_FRESH_GREEN)))
				carsPassingPerFreshGreenPhasePerQueue = state.parameters.getInt(p, null, 1);

			if(state.parameters.exists(p=paramThis.push(P_CARS_PER_OLD_GREEN)))
				carsPassingPerFreshGreenPhasePerQueue = state.parameters.getInt(p, null, carsPassingPerFreshGreenPhasePerQueue);
				
			if(state.parameters.exists(p=paramThis.push(P_QWEALTH_METRIC)))
			{
				qWealthMetric = (QueueWealthMetric)state.parameters.getInstanceForParameter(p, null, QueueWealthMetric.class);
				qWealthMetric.setup(state, p);
			}
				

		}
		if(destinationProb<0) throw new RuntimeException("Invalid destination probability");
		if(sourceProb<0) throw new RuntimeException("Invalid source probability");
		if(phaseDuration<=0) throw new RuntimeException("Invalid phase duration");
		if(phaseOffset<0) throw new RuntimeException("Invalid phase offset");

		if(carsPassingPerOldGreenPhasePerQueue<=0) throw new RuntimeException("Invalid cars/per old green");
		if(carsPassingPerFreshGreenPhasePerQueue<=0) throw new RuntimeException("Invalid cars/per fresh green");

		//these three can be null, meaning no payment is required/no compensation is provided 
//		if(greenCollector==null) throw new RuntimeException("Pay-per-green policy missing");
//		if(passageCollector==null) throw new RuntimeException("Pay-per-passage policy missing");
//		if(compensationPolicy==null) throw new RuntimeException("Compensation-for-waiting policy missing");
		if(rotationPolicy==null) throw new RuntimeException("Rotation policy missing");
		if(qWealthMetric==null) throw new RuntimeException("Assessing queue wealth policy missing");
		
		
		//default value
		doNotCollectUnlessLosers = (compensationPolicy==null)? false: (compensationPolicy instanceof ZeroSumCompensationPolicy);
		//here's your chance to force a value from outside.
		doNotCollectUnlessLosers = state.parameters.getBoolean(	base.push(P_DONT_COLLECT_UNLESS_LOSERS),
																null,
																doNotCollectUnlessLosers);
		//if I have a zeroSum compensation policy, I do no collect from winners if there are no losers to give the "money" to.
		//if there is no compensation policy, I can collect whenever.
	}	
	
	
	public void allocateData(int nQueues, int nGroups)
	{
		groups = new TrafficQueue[nGroups][];
		winnersMask = new boolean[nQueues];
		groupWealths = new float[nGroups];
		queues = new TrafficQueue[nQueues];
	}

	public boolean isEmpty()
	{
		for(int i=queues.length-1;i>=0;i--)
		{
			TrafficQueue q = queues[i];
			if(q.inRangeCars.size()>0) return false;
		}
		return true;
	}

	//incommigEdge is Edge(streets the car is coming from).indexTo 
	//OutgoingEdge is Edge(streets the car is coming from).indexFrom
	//Sean's method would ignore the last argument
	// mine would not, but how would you get this index? 
	public abstract TrafficQueue getQueue(int incomingEdgeIndex , int OutgoingEdgeIndex);
	public abstract long getTotalVehicles();
	public abstract long getStoppedVehicles();
	public abstract TrafficTally getDelayTally();
	
	
	public void updateQueueWealths()
	{
		int n = queues.length;
		for(int i=0;i<n;i++)
		{
			TrafficQueue q = queues[i];
			q.currentWealth = qWealthMetric.getQueueWealth(q);
		}
	}
	

	
	/**
	 * for portraying purposes.
	 * In the general case, no promises are made relative to the order of the Queues
	 */
	public abstract Bag getQueuesForInEdge(int index, Bag b);


	/**
	 * I could use a Tally of pattern length instead. But the average pattern length = T/n
	 * (T=run length, n=number of pattern changes), so one wouldn't get much out of it.
	 * 
	 * A histogram might be better.
	 * @see TrafficSim.getUniterruptedRedsTally
	 * @see TrafficSim.getUniterruptedGreensTally
	 * (the unused queues are not counted in uninterruptedXXXTally)
	 */
	public int numberOfPatternChanges;
	
	
	public int lastWinningGroupIndex;
	float revenueThisPhase;
	boolean losersThisPhase;
	boolean doNotCollectUnlessLosers;

	private boolean collect()
	{
		return (losersThisPhase || !doNotCollectUnlessLosers);
		//either the "don't collect unless losers" does not apply
		//or there are losers.
	}
	
	public void collectForGreen(Car c)
	{
		if(greenCollector!= null && collect())
			revenueThisPhase += greenCollector.collect(c);
	}
	
	public void collectForPassage(Car c)
	{
		if(passageCollector!= null && collect())
			revenueThisPhase += passageCollector.collect(c);
	}
	

	public void step(SimState state)
	{
		TrafficSim ts = (TrafficSim)state;
		
//		double now = ts.doubleTimeSchedule.getTime(); 
		
		if(	compensationPolicy!=null && 
			losersThisPhase && 
			(revenueThisPhase>0 ||
			 	(fixNoCNoCBug && !doNotCollectUnlessLosers)))
		//cause doNotCollectUnlessLosers == compensationPolicy !=null && 
		//									compensationPolicy instanceof ZeroSumCompensationPolicy
		{
			//pay the losers from last phase
			// (those that lost last time are still in the losing queues
			//  however, new ones might have come since)
			
			// later-on winners do not pay for green //TODO finese this issue, remember if its a new green or not
			// later-on winners do pay for passage
			// later-on losers are compensated the same as those present in the beggining of the phase
			// (for implementation reasons, but it also makes a little sense).
	

			compensationPolicy.distribute(this, revenueThisPhase);
			revenueThisPhase = 0;
		} 


		//Decide the winner.
		int winningGroupIndex = rotationPolicy.getWinningGroupIndex(ts, this);
		if(lastWinningGroupIndex != winningGroupIndex)
			numberOfPatternChanges++;
		lastWinningGroupIndex = winningGroupIndex;
		
		for(int i=0;i<winnersMask.length;i++)
			winnersMask[i]=false;
		TrafficQueue[] winners = groups[lastWinningGroupIndex];
		for(int i=0;i<winners.length;i++)
		{
			winnersMask[winners[i].index]=true;	
		}
		
//		if(toString().equals("Intersection27") && now>=110)
//			System.out.println("\tI27 auction, now q has "+queues[12].getColorName());

		
		//are there any losers?
		losersThisPhase = false;
		for(int i=0;i<winnersMask.length;i++)
			if(!winnersMask[i] && queues[i].inRangeCars.size()!=0)//if a losing Q is not empty
			{
				losersThisPhase = true;
				break;
			}
		//later on new losers might join, but old losers cannot leave. 

		//collect credits from the winners.
		revenueThisPhase = 0;
		for(int i=0;i<queues.length;i++)
			if(winnersMask[i])	
				queues[i].winningNotification();
			else
				queues[i].losingNotification();
				
		reschedule();
	}
	

	public Object simpleClone()
	{
		try{return clone();}catch(CloneNotSupportedException ex){return null;}
	}

	
	
	//slow like hell, for debugging purposes only
	public String tableCellDescription(int width)
	{
		String	s = ioi.index+"[";
		if(ioi.out!=null)
			for(int i=0;i<ioi.out.numObjs;i++)
			{
				Edge e= ((Edge)ioi.out.objs[i]);
				if(i>0)s+=" ";
				s+= ((Intersection)e.getOtherNode(this)).ioi.index;

			}
		s+="]";
		for(int i=s.length();i<width;i++)
			s+=" ";
		return s;
	}
	public String toString(){return (ioi!=null)?("Intersection"+ioi.index):null;}
}
