package sim.app.traffic;
import java.io.*;

import sim.engine.*;
import ec.util.*;
import ec.multirun.*;
/**
 * It executes multiple runs of a TrafficSim, each with 
 * a new random seed. Each run of the TrafficSim stores 
 * the results as a row in an Averager.
 * 
 * <p>It should work with a generic class for any ParamSimState but I don't
 * know how to construct an instance of a superclass of ParamSimState
 * that takes parameters. Also I need to accomodate long and double time schedules, 
 * so ... too much trouble.
 * 
 * <p>It uses a .params file to get the number of runs and the length of each run
 * 
 * 
 * {Int|Double}Distribution uses SimState.random to generate a seed, so I cannot reuse
 * TrafficSim between runs. Even if no Distribution is used right now, I don't want to 
 * risk forgeting about this.
 * 
 * <p>It generates the seeds, so you better redirect the output into a file.
 * 
 * @author Gabriel Balan
 */
public abstract class MultiRun 
{
	public static final String A_FILE = "-file";
	
	public static final Parameter P_RUNS = new Parameter("runs");
	public static final Parameter P_LENGTH = new Parameter("run-length");
	public static final Parameter P_ANOVA_FILE = new Parameter("anova-file");
	
	public static void main(String[] args)
	{
		ParameterDatabase parameters=null;
		for(int x=0;x<args.length-1;x++)
			if (args[x].equals(A_FILE))
			{
				try
				{
					parameters=new ParameterDatabase(new File(new File(args[x+1]).getAbsolutePath()),args);
					break;
				}catch(FileNotFoundException e)
				{ 
					throw new RuntimeException("A File Not Found Exception was generated upon" +
						"reading the parameter file \"" + args[x+1] + "\".\nHere it is:\n" + e); 
				}catch(IOException e)
				{
					throw new RuntimeException("An IO Exception was generated upon reading the" +
						"parameter file \"" + args[x+1] +"\".\nHere it is:\n" + e);
				}
			}
		if (parameters==null)
			throw new RuntimeException("No parameter file was specified." );
			
		int nRuns = parameters.getInt(P_RUNS, null, 1);
		if(nRuns<1) throw new RuntimeException("Too few runs.");


		double runLen = parameters.getDouble(P_LENGTH, null, 0);
		if(runLen<=0) throw new RuntimeException("Too short of a run.");
		
		File anovaFile = parameters.getFile(P_ANOVA_FILE, null);
		if(anovaFile == null)
			throw new RuntimeException("No anova storage file specified.");

		AnovaDataStorage anovaStorage = new ec.multirun.AnovaDataStorage(nRuns);
		int anoveStorageId = ec.multirun.AnovaDataStorage.getId(anovaStorage);
		parameters.set(new Parameter(TrafficSim.P_ANOVASTORAGE_ID), ""+anoveStorageId);
		
		
		MersenneTwisterFast rng = new MersenneTwisterFast();
		
		for(int i=0;i<nRuns;i++)
		{
			long seed_i = rng.nextLong();
			System.out.println("\tRun "+i+"\tSeed "+seed_i);
			parameters.set(new Parameter(TrafficSim.P_SEED), seed_i+"");
			TrafficSim ts = new TrafficSim(parameters);
			Schedule schedule = ts.schedule;
			ts.start();
			while(schedule.time()<runLen)
				if (!schedule.step(ts)) break;
			ts.finish();
		}

		try
		{
			FileWriter fw =new FileWriter(anovaFile); 
			PrintWriter pw =  new PrintWriter(fw);
			anovaStorage.print(pw);
			pw.flush();
			pw.close();
			fw.close();
		}catch(IOException ex){ex.printStackTrace();}
	}
}
	
