package sim.app.traffic;
import sim.field.network.*;
import sim.util.*;
import ec.util.*;

/**
 * Generalized Intersection: there is a TrafficQueue for each pair
 * incoming edge - outgoing edge. 
 * 
 * <p>This is more sensible than <code>OneQeueGroupIntersection</code>: 
 * a set of non-conflicting cars are allowed in the intersection, 
 * regardless of the street they are coming from. 
 * 
 * <p>There is a limitation, however: two queues flowing into the same out edge
 * is considered a conflict. So, the incoming edges are split into as many lanes as
 * outgoing edges, but every outgoing edge has only one lane.
 * 
 * <p>Unlike the simplified mechanism, this one allows two way traffic
 * (e.g. N and S)
 * 
 * <p>However, this one requires the planar graph information.
 * 
 * @author Gabriel Balan
 */
public class OneLaneOutEdgeIntersection extends Intersection 
{	
	public int nQueues;
	
	public static class EdgeIndex
	{
		public EdgeIndex(int i, boolean b){index = i;in=b;}
		int index;
		boolean in;
	}

	public void netSetup(TrafficSim ts, Network.IndexOutIn ioi, EdgeIndex[] order)
	{
		this.netSetup(ts, ioi, getGroups(ioi.in.numObjs, ioi.out.numObjs, order));
	}
	
	/**
	 * use this to reuse a precomputed <code>groupSpecs</code> 
	 */
	public void netSetup(TrafficSim ts, Network.IndexOutIn ioi, int[][] groupSpecs)
	{
		super.netSetup(ioi);
		int ins = ioi.in.numObjs;
		int outs = ioi.out.numObjs;
		nQueues = ins*outs;
		int nGroups = groupSpecs.length;		
		allocateData(nQueues, nGroups);
		for(int i=0;i<nQueues;i++)
				queues[i]= new TrafficQueue(ts, this, i, carsPassingPerFreshGreenPhasePerQueue, carsPassingPerOldGreenPhasePerQueue, takeOffTime);
		for(int i=0;i<nGroups;i++)
		{
			int[] group_i_specs = groupSpecs[i];
			int group_i_size = group_i_specs.length;
			TrafficQueue[] group_i = new TrafficQueue[group_i_size];
			groups[i]=group_i;
			for(int j=0;j<group_i_size;j++)
				group_i[j]=queues[group_i_specs[j]];
		}
	}
	
	public static boolean[][] getConflictsMatrix(int ins, int outs, EdgeIndex[] order, boolean[] forbiddenQueueMask)
	{
		int nQueues = ins*outs;
		int nEdges = ins+outs;
		if(nEdges!=order.length) throw new RuntimeException("order spec has the wrong size");
		
		boolean[][] conflicts = new boolean[nQueues][nQueues];
		for(int i=0;i<nQueues;i++)
		{
			boolean[] conflicts_i = conflicts[i];
			for(int j=0;j<nQueues;j++) conflicts_i[j]=true;
		}
		

		EdgeIndex tmp;
		IntBag insB = new IntBag(ins);
		IntBag outsB = new IntBag(outs);
		
		for(int i=0;i<nEdges;i++)
		{
			tmp = order[i];
			if(!tmp.in) continue;
			EdgeIndex e1In = tmp;
			for(int j=1;j<nEdges;j++)
			{
				tmp = order[(i+j+nEdges)%nEdges];
				if(tmp.in) continue;
				EdgeIndex e1Out = tmp;
				int queue_1_index = e1In.index*outs+e1Out.index;
				if(forbiddenQueueMask[queue_1_index])
					continue;//everything is in conflict with this guy
					
				boolean[] conflicts_line = conflicts[queue_1_index];				
				//extract the in and out edges on the same side of e1In-eiOut
				for(int k=0;k<j;k++)
				{
					tmp = order[(k+i+nEdges)%nEdges];
					(tmp.in?insB:outsB).add(tmp.index);
				}
				// and sign them as no conflict
				for(int x=0;x<insB.numObjs;x++)
				{
					int e_x = insB.objs[x];
					for(int y=0;y<outsB.numObjs;y++)
					{
						int queue_2_index = e_x*outs+outsB.objs[y];
						conflicts_line[queue_2_index]=false;
						conflicts[queue_2_index][queue_1_index]=false;
//						System.out.println("No conflict between "+e1In.index+"->"+e1Out.index+" and "+e_x+"->"+outsB.objs[y]);
					}
				}
				insB.clear();outsB.clear();
				//same for the other side.
				for(int k=j+1;k<=nEdges;k++)
				{
					tmp = order[(k+i+nEdges)%nEdges];
					(tmp.in?insB:outsB).add(tmp.index);
				}
				for(int x=0;x<insB.numObjs;x++)
				{
					int e_x = insB.objs[x];
					for(int y=0;y<outsB.numObjs;y++)
					{
						int queue_2_index = e_x*outs+outsB.objs[y];
						conflicts_line[queue_2_index]=false;
						conflicts[queue_2_index][queue_1_index]=false;
//						System.out.println("No conflict between "+e1In.index+"->"+e1Out.index+" and "+e_x+"->"+outsB.objs[y]);
					}
				}
				insB.clear();outsB.clear();
			}	
		}
	
		return conflicts;
	}
	
	public void printConflicMatrix(int outs, boolean[][] conflicts)
	{
			System.out.print(" \t\t\t|");
			for(int i=0;i<nQueues;i++)
			{
				System.out.print((i%10)+" ");
			}
			System.out.println();
			for(int i=0;i<nQueues;i++)
			{
				System.out.print(i+"\t("+(i/outs)+"-> "+(i%outs)+")\t|");
				for(int j=0;j<nQueues;j++)
					System.out.print(conflicts[i][j]?"  ":"x ");
				System.out.println("|");
			}
	}
	
	
	public static int[][] getGroups(int ins, int outs, EdgeIndex[] order, int[] forbiddenQueueIndices)
	{	
		int nQueues = ins*outs;
		boolean[] forbiddenQueueMask = new boolean[nQueues];//by default nobody's forbidden
		if(forbiddenQueueIndices!=null)
			for(int i=0;i<forbiddenQueueIndices.length;i++)
				forbiddenQueueMask[forbiddenQueueIndices[i]]=true; 
		return getGroups(ins, outs, getConflictsMatrix(ins, outs, order, forbiddenQueueMask), forbiddenQueueMask);
		
	}	
	public static int[][] getGroups(int ins, int outs, EdgeIndex[] order)
	{	
		int nQueues = ins*outs;
		boolean[] forbiddenQueueMask = new boolean[nQueues];//by default nobody's forbidden
		return getGroups(ins, outs, getConflictsMatrix(ins, outs, order, forbiddenQueueMask), forbiddenQueueMask);
	}

	public static int[][] getGroups(int ins, int outs, boolean[][]conflicts, boolean[] forbiddenQueueMask)
	{
		int nQueues = conflicts.length;
		Bag groupBag = new Bag(ins);
		Stack stack = new Stack(outs);
		stack.push(-1);
		boolean	pile = false;	
		do
		{
			int top = stack.top;
			if(pile)
			{
				if(top == outs)
				{
					int[] group = new int[outs];
					System.arraycopy(stack.data, 0, group, 0, outs);
					groupBag.add(group);
				}
				else
				{
					stack.push(stack.data[top-1]);
				}
				pile = false;
			}
			else
			{//increment top.
				int i= stack.pop();
				while((++i)<nQueues)
				{		
					boolean conflict = false;
					if(!forbiddenQueueMask[i])
					{
						boolean[] conflicts_i = conflicts[i];
						
						for(int j=0;j<stack.top;j++)
							if(conflicts_i[stack.data[j]])
							{	
								conflict = true;
								break;
							}
						if(!conflict)
						{
							stack.push(i);
							pile = true;
							break;
						}
					}
				}
			}
			 
		}while(stack.top>0);//not empty
		
		int[][] groupSpecs = new int[groupBag.numObjs][];
		for(int i=groupBag.numObjs-1;i>=0;i--)
			groupSpecs[i]=(int[])groupBag.objs[i];
		return groupSpecs;
	}
	
	public Bag getQueuesForInEdge(int index, Bag b)
	{
		int outs = ioi.out.numObjs;
		if(b.objs.length>outs) b.resize(outs);
		try{
			System.arraycopy(queues, index*outs, b.objs, 0, outs);
			b.numObjs = outs;
		}catch(ArrayIndexOutOfBoundsException ex)
		{
			outs = ioi.out.numObjs;//blablabla, just a line to stick a break point into.
		}
		return b;
	}

	
	public TrafficQueue getQueue(
		int incomingEdgeIndex,
		int outgoingEdgeIndex)
	{
		return queues[incomingEdgeIndex*ioi.out.numObjs+outgoingEdgeIndex];
	}
	
	public static int getQueueId(	int incomingEdgeIndex,
									int OutgoingEdgeIndex,
									int outs)
	{ 
		return incomingEdgeIndex*outs+OutgoingEdgeIndex;
	}
	

//TODO I should store the results along with the time
//they were computed. 
//I can also do it in an inlineable fashion.
	public long getTotalVehicles() {
		long sum = 0;
		for(int i=0;i<nQueues;i++)
			sum+= queues[i].totalVehicles;
		return sum;
	}

	public long getStoppedVehicles() {
		long sum = 0;
		for(int i=0;i<nQueues;i++)
			sum+= queues[i].stoppedVehicles;
		return sum;
	}

	public TrafficTally getDelayTally() 
	{
		//if this gets called at the end, its fine.
		//TODO if this gets called often, use a static Tally to avoid "new" ops
		TrafficTally tally = new TrafficTally();
		for(int i=0;i<nQueues;i++)
			tally.add(queues[i].delayTally);
		return tally;
	}

//	public void updateQueueWealths()
//	{
//		for(int i=0;i<nQueues;i++)
//		{
//			TrafficQueue q = queues[i];
//			q.currentWealth = qWealthMetric.getQueueWealth(q);
//		}
//	}
	
	
	public Parameter defaultBase(){return null;}
	static class Stack
	{
		final int[] data;
		int top = 0;
		public Stack(int n){data = new int[n];}
		public void push(int datum){data[top++]=datum;}
		public int pop(){return data[--top];}
		public boolean empty(){return top==0;}
		public String toString()
		{
			String s = "";
			for(int i=0;i<top;i++)
				s+=data[i]+", ";
			return s;
		}
	}
}
