package sim.app.traffic;

import sim.util.*;
import sim.engine.*;
import sim.app.traffic.analysis.CreditCheckCar;
import sim.app.traffic.car.*;
import sim.app.traffic.payment.*;
import sim.util.stats.doubletime.*;

/**
 * 
 * This is a Queue keeping track of a couple of statistics, such as average delay and percentage of stopped cars.
 * 
 * <p>If you only care about the overall stats for the intersection or entire
 * simulation, thes stats at the queue level are not needed.
 * 
 * <p>It should help with the "granularity of double" issue: after addind lots of
 * small delays, adding another one would make no change. If we add the tiny
 * delays togher into groups (Queues and then Intersections), the result would
 * be more accurate.
 * 
 * <p>On the down side, if you need the stats at every time step (e.g. when you turn
 * on the GUI), then a ton of numbers will have to be added together over and
 * over again.
 * 
 * <code>Intersection</code> knows about the "getting payed at red" policy, <code>TrafficQueue</code> knows
 * about the "paying for green" policy
 * 
 * @author Gabriel Balan
 */
public class TrafficQueue extends SteppableQueue
{
	public int stoppedVehicles;

	public int totalVehicles;

	/**
	 * Used to ignore the unused queues, (e.g. Uturn queues). It's not reset in
	 * resetStats, but it is in start.
	 */
	boolean unused = true;

	public TrafficTally delayTally = new TrafficTally();

	// this is a memoization of the QueueWealthMetric
	public float currentWealth;

	public CarBagWithO1Remove inRangeCars = new CarBagWithO1Remove(
			CarBagWithO1Remove.QUEUE_BAG);

	/** Number of diary-keeping cars that are in the sensor range of this queue. */
	int diaryKeepingCarsInRange = 0;

	// I will look for diary-keeping cars (and keep them informed) only if there
	// are such cars present.
	boolean lookForDiaryKeepingCars()
	{
		return diaryKeepingCarsInRange != 0;
	}

	public void registerDiaryKeepingCars()
	{
		diaryKeepingCarsInRange++;
	}

	public void unregisterDiaryKeepingCars()
	{
		diaryKeepingCarsInRange--;
	}

	public TimeWeightedAverager queueLengthCalculator;

	public void queueLenghChanged()
	{
		queueLengthCalculator.addSample(size);
	}

	public void queueLenghChangedWithCorrection(double correction)
	{
		queueLengthCalculator.addSample(size);
		queueLengthCalculator.performCorrection(correction);
	}

	// TODO pack these in an object?
	// TODO I might need the actual numbers for a Box&Whiskers plot!!
	TrafficTally uninterruptedGreenTally = new TrafficTally();

	TrafficTally uninterruptedRedTally = new TrafficTally();

	int consecutivePhasesWithNoColorChange = 0;

	// public double getAverageUniterruptedReds()
	// {
	// if(consecutivePhasesWithNoColorChange<2 || greenLastTime)
	// //it's green or the phase just changed
	// return uninterruptedRedTally.getAverage();
	//
	// //otherwise it's been red for a while
	// return
	// uninterruptedRedTally.getAverageWithExtraValue(consecutivePhasesWithNoColorChange);
	//		
	// //if I just changed to red, should I return AverageWithExtraValue(1)
	// instead of getAverage()?
	// //maybe I should consistently return getAverage. For enough samples it
	// would not matter.
	// }
	// public double getAverageUniterruptedGreens()
	// {
	//		
	// if(consecutivePhasesWithNoColorChange<2 || !greenLastTime)
	// //it's red or the phase just changed
	// return uninterruptedRedTally.getAverage();
	//
	// //otherwise it's been red for a while
	// return
	// uninterruptedRedTally.getAverageWithExtraValue(consecutivePhasesWithNoColorChange);
	//		
	// //if I just changed to red, should I return AverageWithExtraValue(1)
	// instead of getAverage()?
	// //maybe I should consistently return getAverage. For enough samples it
	// would not matter.
	// }

	/** the Queue's index within the Intersection */
	public final int index;

	public final Intersection owner;

	public TrafficSim state;

	public final double takeOffTime;

	public TrafficQueue(TrafficSim ts, Intersection inter, int index,
			int carsPerFreshGreen, int carsPerOldGreen, double takeOffTime)
	{
		super();
		state = ts;
		this.index = index;
		owner = inter;
		this.maxCarsPassingPerFreshGreenPhase = carsPerFreshGreen;
		this.maxCarsPassingPerOldGreenPhase = carsPerOldGreen;
		this.takeOffTime = takeOffTime;
		queueLengthCalculator = new TimeWeightedAverager(ts.schedule,0);
	}

	public void start()
	{
		resetStats();
		clear();
		inRangeCars.clear();
		currentWealth = 0;// for the GUI
		unused = true;
	}

	public void resetStats()
	{
		stoppedVehicles = 0;
		totalVehicles = 0;
		delayTally.reset();
		uninterruptedGreenTally.reset();
		uninterruptedRedTally.reset();
		queueLengthCalculator.start();		
	}

	public void recordCarDelay(double delay)
	{
		totalVehicles++;
		if (delay != 0)
			stoppedVehicles++;
		delayTally.add(delay);
	}

	/**
	 * returns the total wealth of the cars waiting in the queue. way faster
	 * than getWaitingCarsTotalWealth(1)
	 */
	public float getSumOfCarsAvgWT(double defaultValue)
	{
		float sum = 0;
		Bag b = inRangeCars.bag;
		for (int i = 0; i < b.numObjs; i++)
		{
			Car c = (Car) b.objs[i];
			sum += c.waitingTimeTally.getAverageWithDefault(defaultValue);
		}
		return sum;
	}

	public float getSumOfCarsInRangeTimes()
	{
		float sum = 0;
		double now = state.schedule.time();
		Bag b = inRangeCars.bag;
		for (int i = 0; i < b.numObjs; i++)
		{
			Car c = (Car) b.objs[i];
			sum += c.getTimeInCurrentRange(now);
		}
		return sum;
	}

	public float getSumOfCarsAvgWTPlusCurrent(double now)
	{
		float sum = 0;
		Bag b = inRangeCars.bag;
		for (int i = 0; i < b.numObjs; i++)
		{
			Car c = (Car) b.objs[i];
			sum += c.waitingTimeTally.getAverageWithExtraValue(c
					.getTimeInCurrentQueue(now));
		}
		return sum;
	}

	/**
	 * returns the total wealth of the cars waiting in the queue. way faster
	 * than getWaitingCarsTotalWealth(1)
	 */
	public float getSumOfCarsWealth()
	{
		float sum = 0;
		Bag b = inRangeCars.bag;
		for (int i = 0; i < b.numObjs; i++)
		{
			Car c = (Car) b.objs[i];
			sum += c.wealth;
		}
		return sum;
	}

	/**
	 * returns the some of squared wealth of the cars waiting in the queue. way
	 * faster than getWaitingCarsTotalWealth(2)
	 */
	public float getSumOfSqCarWealth()
	{
		float sum = 0;
		Bag b = inRangeCars.bag;
		for (int i = 0; i < b.numObjs; i++)
		{
			Car c = (Car) b.objs[i];
			double w = c.wealth;
			sum += w * w;
		}
		return sum;
	}

	/**
	 * returns the sum of squared roots of wealth of the cars waiting in the
	 * queue. way faster than getWaitingCarsTotalWealth(.5)
	 */
	public float getSumOfSqrtCarWealth()
	{
		float sum = 0;
		Bag b = inRangeCars.bag;
		for (int i = 0; i < b.numObjs; i++)
		{
			Car c = (Car) b.objs[i];
			double w = c.wealth;
			sum += Math.sqrt(w);
		}
		return sum;
	}

	public float getSumOfPowCarWealth(double power)
	{
		float sum = 0;
		Bag b = inRangeCars.bag;
		for (int i = 0; i < b.numObjs; i++)
		{
			Car c = (Car) b.objs[i];
			sum += Math.pow(c.wealth, power);
		}
		return sum;
	}

	/**
	 * returns the total wealth of the cars waiting in the queue.
	 */
	public float getSumOfCarBids()
	{
		float sum = 0;
		Bag b = inRangeCars.bag;
		for (int i = 0; i < b.numObjs; i++)
		{
			Car c = (Car) b.objs[i];
			sum += c.getBid(state);
		}
		return sum;
	}

	// TODO you might want to make a "function" of time
	// (interface) out of it. This way you can have
	// distributions, situations like 2.5 cars, etc.
	int maxCarsPassingPerFreshGreenPhase;
	int maxCarsPassingPerOldGreenPhase;
	public int getMaxCarsPassingPerOldGreenPhase(){return maxCarsPassingPerOldGreenPhase;}
	public int getMaxCarsPassingPerFreshGreenPhase(){return maxCarsPassingPerFreshGreenPhase;}

	float creditBalance = 0;

	// if you had red last time, you got to let those that did had green clear
	// the intersection
	public boolean greenLastTime = false;

	// no more than maxCarsPassingPerGreenPhase can pass in one phase
	// regardless of whether the car was stopped at the switch or not.
	int crossingOpportunities;
	public int getCrossingOpportunities()
	{
		return crossingOpportunities;
	}

	public int color;

	public static final int RED = 0;
	public static final int YELLOW = 1;
	public static final int GREEN = 2;

	static final String colorNames[] = { "red", "yellow", "green" };

	public String getColorName()
	{
		return colorNames[color];
	}

	/*
	 * The queue gets the verdict intersection_Xing_time in advance!!! If the
	 * queue wins and used to be red then [it turns YELLOW and] it waits
	 * intersection_Xing_time before releasing the cars, so the the intersection
	 * gets clear. it turns GREEN and releases the cars.
	 * 
	 * If the queue wins and used to be green then releases cars right away (it
	 * was safe before, it's safe now) ALSO let more cars through.
	 * 
	 * if the queue loses and used to be green [it turns YELLOW for
	 * intersection_Xing_time]. it turns RED
	 * 
	 * YELLOW IS JUST A WHIM. It will only slow things down.
	 */

	/**
	 * If there are less than n cars in the queue, the guys in the queue pay up.
	 * 
	 * If another comes a little later and still catches the "green", it should
	 * not pay, since the green has been already paid for.
	 * 
	 * TODO TrafficQueue wakes up at most n cars that go through the
	 * intersection. "n" should be a function of the duration of the green
	 * phase.
	 * 
	 */
	public void winningNotification()
	{
		PaymentPolicy greenCollector = owner.greenCollector;

		if (greenCollector != null)
		// TODO who pays for green: all in sensing radius or all waited in the
		// Q?
		// for(int i=0;i<size;i++)
		// owner.collectForGreen((Car)array[(i+front)%array.length]);
		{
			Bag b = inRangeCars.bag;
			for (int i = 0; i < b.numObjs; i++)
				owner.collectForGreen((Car) b.objs[i]);
		}

		if (greenLastTime)
		{
			// here's one more phase to the uninterrupted sequence of greens
			consecutivePhasesWithNoColorChange++;

			// Maybe I shouldn't count this green phase if no cars used it.
			// i.e. crossingOpportunities == maxCarsPassingThisGreenPhase

			popCars(maxCarsPassingPerOldGreenPhase);
		} else
		{
			color = YELLOW;
			// it's the end of a sequence of red phases.
			uninterruptedRedTally.add(consecutivePhasesWithNoColorChange);
			// and the beggining of a green one.
			consecutivePhasesWithNoColorChange = 1;

			state.schedule.scheduleOnceIn(
					state.roadNet.intersectionCrossingDuration,
					new Steppable()
					{
						public void step(SimState simState)
						{
							color = GREEN;
							popCars(maxCarsPassingPerFreshGreenPhase);
						}
					},
					TrafficSim.QUEUE_ORDERING);
		}

		if (lookForDiaryKeepingCars())
		{
			inRangeCars.broadcastDiaryEntryWithQposition("auction won");
		}

	}

	/**
	 * if the queue loses and used to be green it turns YELLOW for
	 * intersection_Xing_time, then it turns RED
	 */
	public void losingNotification()
	{
		if (greenLastTime)
		{
			greenLastTime = false;
			crossingOpportunities = 0;
			color = YELLOW;
			// it's the end of a sequence of green phases.
			uninterruptedGreenTally.add(consecutivePhasesWithNoColorChange);
			// and the beggining of a green one.
			consecutivePhasesWithNoColorChange = 1;
			state.schedule.scheduleOnceIn(
					state.roadNet.intersectionCrossingDuration,
					new Steppable()
					{
						public void step(SimState simState)
						{
							color = RED;
						}
					},
					TrafficSim.QUEUE_ORDERING);
		} else
		{
			// here's one more phase to the uninterrupted sequence of reds
			consecutivePhasesWithNoColorChange++;
		}
		if (lookForDiaryKeepingCars())
		{
			inRangeCars.broadcastDiaryEntryWithQposition("auction lost");
		}
	}

	private void popCars(int maxCarsPassingThisGreenPhase)
	{

		int n = (maxCarsPassingThisGreenPhase < size) ? maxCarsPassingThisGreenPhase
				: size;
		crossingOpportunities = maxCarsPassingThisGreenPhase - n;

		boolean lookForDiaryKeepingCars = lookForDiaryKeepingCars();

		for (int i = 1; i <= n; i++)
		{
			Car c = (Car) dequeue();

			if (lookForDiaryKeepingCars)
				((CreditCheckCar) c).addLogEntry("scheduled for release in "
						+ (i * takeOffTime));

			owner.collectForPassage(c);
			state.schedule.scheduleOnceIn(i*takeOffTime,c,TrafficSim.CAR_ORDERING);

			queueLenghChangedWithCorrection(i * takeOffTime);
			// with correction, since the queue length WILL change in
			// i*takeOffTime, not right away.
		}
		greenLastTime = true;
	}

	/**
	 * Adds a car to this queue, unless the queue is empty and the light is
	 * green, when it goes right through.
	 * 
	 * @param c =
	 *            the car
	 * @return whether the car got stuck in the queue.
	 */
	public boolean enqueueCar(Car c)
	{
		if (color == GREEN)
		{
			if (size == 0 && crossingOpportunities > 0)
			{
				crossingOpportunities--;
				owner.collectForPassage(c);
				return false;
			}
		} else
			owner.losersThisPhase = true;
		enqueue(c);
		queueLenghChanged();
		return true;
	}

	public void registerInRange(Car c)
	{
		unused = false;
		inRangeCars.add(c);
	}

	public void unregisterInRange(Car c)
	{
		inRangeCars.remove(c);
	}

	// High speed compensation methods
	// handed over to the bag to access the data directly)

	public float compensateFixedAmount(float payoff, boolean stoppedOnly)
	{
		boolean look4DiaryKeepingCars = lookForDiaryKeepingCars();
		if (stoppedOnly)
		{
			for (int i = 0; i < size; i++)
			{
				int offset = (front + i) % array.length;
				Car c = (Car) array[offset];
				if (c == null)
					System.out.println();
				c.wealth += payoff;
				if (look4DiaryKeepingCars && c instanceof CreditCheckCar)
					((CreditCheckCar) c).addLogEntry("Compensated $" + payoff
							+ " [$" + c.wealth + "]");
			}
			return size * payoff;
		}

		if (look4DiaryKeepingCars)
			return inRangeCars.compensateFixedAmountWithDiary(payoff);
		return inRangeCars.compensateFixedAmount(payoff);
	}

	/**
	 * I assume that Intersection refreshed tmpTotalBids by calling
	 * getWaitingCarsTotalBids() this time step to determine the winner
	 * 
	 * @param share =
	 *            how much it pays per unit of bid.
	 * 
	 */
	public void compensateBidPropotionally(float share, boolean stoppedOnly)
	{
		boolean look4DiaryKeepingCars = lookForDiaryKeepingCars();
		if (stoppedOnly)
			for (int i = 0; i < size; i++)
			{
				int offset = (front + i) % array.length;
				Car c = (Car) array[offset];

				float lumpSum = c.currentBid * share;
				c.wealth += lumpSum;
				if (look4DiaryKeepingCars && c instanceof CreditCheckCar)
					((CreditCheckCar) c).addLogEntry("Compensated $" + lumpSum
							+ " [$" + c.wealth + "]");

			}

		if (look4DiaryKeepingCars)
			inRangeCars.compensateBidPropotionallyWithDiary(share);
		else
			inRangeCars.compensateBidPropotionally(share);
	}

	/**
	 * @return -1 if the car's not [stopped] in this queue. So you better make
	 *         sure your car is in the queue.
	 */
	public int getCarPositionInQueue(Car c)
	{
		for (int i = 0; i < size; i++)
		{
			int offset = (front + i) % array.length;
			if (c == array[offset])
				return i;
		}
		return -1;
	}

	/**
	 * My intersection is considering giving me the RED/GREEN light. How much
	 * bigger is my delay at this intersection going to be if I get RED
	 * instead of GREEN?
	 */
	public double getRedOverGreenTotalDelayIncrease()
	{
		double totalExtraDelay = 0;
		if(inRangeCars.bag.numObjs==0)
			return totalExtraDelay;
			

		double nextPhaseDuration = owner.phaseDuration;

		// /upper limit on the number of cars I'll be able to
		// get through next phase (assuming I win)
		int crossingOpportunitiesNextPhase = (color == GREEN) ? maxCarsPassingPerOldGreenPhase
				: maxCarsPassingPerFreshGreenPhase;

		// for each car C already stopped
		// if C could made it through [if it gets green]
		// -add 5secs of yellow and 3secs*position_of_C.
		// else add 20secs

		// <code>takeoffDelay</code> includes the yellow duration and the
		// <code>takeoffTime</code>s
		double takeoffDelay = (color == GREEN) ? state.roadNet.intersectionCrossingDuration: 0;
		for (int i = 0; i < size; i++)
		{
			// int offset = (front+i)%array.length;
			// Car c = (Car)array[offset];

			double delay_i = 0;
			if (crossingOpportunitiesNextPhase == 0)
			{
				delay_i = nextPhaseDuration;
			} else
			{
				crossingOpportunitiesNextPhase--;
				takeoffDelay += takeOffTime;
				delay_i = takeoffDelay;
			}
			totalExtraDelay += delay_i;
		}

		//
		// for each car C moving in-range
		// there's a delay of
		// (20secs-the time it takes for C to reach the intersection) if C would
		// make it through
		// //the first _crossing-opportunities_ cars
		// 0 secs for the rest (they won't make it through either way).
		double now = state.schedule.time();
		int numberInRangeCars = inRangeCars.bag.numObjs;
		int numberInRangeMovingCars = inRangeCars.bag.numObjs-size;
		if (numberInRangeMovingCars != 0 && crossingOpportunitiesNextPhase != 0)
		{
			if(numberInRangeMovingCars <=crossingOpportunitiesNextPhase)
				//all cars in range could pass through.
				for (int i = 0; i < numberInRangeCars; i++)
				{
					Car c = (Car) inRangeCars.bag.objs[i];
					if(c.isStoppedInQueue()) continue;
					double delay_i = Math.max(nextPhaseDuration-c.getTimeTillIntersection(now), 0);
					totalExtraDelay += delay_i;
				}
			else
			{
				// only the first "crossingOpportunitiesNextPhase" cars
				// would be affected by the next RED/GREEN decision

				Double[] keys = tmpHeapKeys;
				Car[] values = tmpHeapValues;
				if(keys.length<numberInRangeCars)
				{//resize
					keys = new Double[numberInRangeCars];
					values = new Car[numberInRangeCars];
				}

				int index = 0;
				for (int i = 0; i < numberInRangeCars; i++)
				{
					Car c = (Car) inRangeCars.bag.objs[i];
					if(c.isStoppedInQueue()) continue;
					values[index] = c;
					keys[index] = new Double(c.getTimeTillIntersection(now));
					index++;
				}
				Heap h = new Heap(keys, values, numberInRangeMovingCars);
				for (int x = 0; x < crossingOpportunitiesNextPhase; x++)
				{
					double timeTillIntersection = ((Double)h.getMinKey()).doubleValue();
					// Car c = (Car)
					h.extractMin();
					double delay_i = Math.max(nextPhaseDuration	- timeTillIntersection, 0);
					totalExtraDelay += delay_i;
				}
			}
		}
		if(totalExtraDelay<0)
			throw new RuntimeException("Error: getRedOverGreenTotalDelayIncrease<0");
		return totalExtraDelay;
	}
	
	
	
	/**
	 * My intersection is considering giving me the RED/GREEN light. How much
	 * bigger would my sum of trip-avereaged delays be if I get RED instead of GREEN?
	 */
	public double getRedOverGreenTripAvgdDelayIncrease()
	{
		double totalTripAvgdDelay = 0;

		double nextPhaseDuration = owner.phaseDuration;

		// /upper limit on the number of cars I'll be able to
		// get through next phase (assuming I win)
		int crossingOpportunitiesNextPhase = (color == GREEN) ? maxCarsPassingPerOldGreenPhase
				: maxCarsPassingPerFreshGreenPhase;

		// for each car C already stopped
		// if C could made it through [if it gets green]
		// -add 5secs of yellow and 3secs*position of C.
		// else add 20secs

		// <code>takeoffDelay</code> includes the yellow duration and the
		// <code>takeoffTime</code>s
		double takeoffDelay = (color == GREEN) ? state.roadNet.intersectionCrossingDuration
				: 0;
		for (int i = 0; i < size; i++)
		{
			 int offset = (front+i)%array.length;
			 Car c = (Car)array[offset];

			double delay_i = 0;
			if (crossingOpportunitiesNextPhase == 0)
			{
				delay_i = nextPhaseDuration;
			} else
			{
				crossingOpportunitiesNextPhase--;
				takeoffDelay += takeOffTime;
				delay_i = takeoffDelay;
			}

			totalTripAvgdDelay += c.waitingTimeTally.getAverageWithExtraValue(delay_i);
		}

		//
		// for each car C moving in-range
		// there's a delay of
		// (20secs-the time it takes for C to reach the intersection) if C would
		// make it through
		// //the first _crossing-opportunities_ cars
		// 0 secs for the rest (they won't make it through either way).
		double now = state.schedule.time();
		int numberInRangeCars = inRangeCars.bag.numObjs;
		int numberInRangeMovingCars = numberInRangeCars - size;
		if (numberInRangeMovingCars != 0 && crossingOpportunitiesNextPhase != 0)
		{
			if(numberInRangeMovingCars <=crossingOpportunitiesNextPhase)
			//all cars in range could pass through.
				for (int i = 0; i < numberInRangeCars; i++)
				{
					Car c = (Car) inRangeCars.bag.objs[i];
					if(c.isStoppedInQueue()) continue;
					double delay_i = Math.max(nextPhaseDuration-c.getTimeTillIntersection(now), 0);
					totalTripAvgdDelay += c.waitingTimeTally.getAverageWithExtraValue(delay_i);
				}
			else
			{
				// i need the first "crossingOpportunitiesNextPhase"
				// to reach the intersection.

				Double[] keys = tmpHeapKeys;
				Car[] values = tmpHeapValues;
				if(keys.length<numberInRangeCars)
				{//resize
					keys = new Double[numberInRangeCars];
					values = new Car[numberInRangeCars];
				}
				
				int index =0;
				for (int i = 0; i < numberInRangeCars; i++)
				{
					Car c = (Car) inRangeCars.bag.objs[i];
					if(c.isStoppedInQueue()) continue;
					values[index] = c;
					keys[index] = new Double(c.getTimeTillIntersection(now));
					index++;
				}
				Heap h = new Heap(keys, values, numberInRangeMovingCars);
				for (int x = 0; x < crossingOpportunitiesNextPhase; x++)
				{
					double timeTillIntersection = ((Double)h.getMinKey()).doubleValue();
					Car c = (Car)h.extractMin();
					double delay_i = Math.max(nextPhaseDuration - timeTillIntersection, 0);
					totalTripAvgdDelay += c.waitingTimeTally.getAverageWithExtraValue(delay_i);
				}
			}
		}
		
		if(totalTripAvgdDelay<0)
			throw new RuntimeException("Error: RedOverGreenTripAvgdDelayIncrease<0");
		return totalTripAvgdDelay;
	}

	static Double[] tmpHeapKeys = new Double[0];
	static Car[] tmpHeapValues = new Car[1];

	
//	/**
//	 * My intersection is considering giving me the RED/GREEN light. How much
//	 * bigger is my delay at this intersection going to be if i get RED
//	 * instead of GREEN?
//	 */
//	public double getRedOverGreenTotalDelayIncrease()
//	{
//		double totalExtraDelay = 0;
//		if(inRangeCars.bag.numObjs==0)
//			return totalExtraDelay;
//			
//
//		double nextPhaseDuration = owner.phaseDuration;
//
//		// for each car C already stopped the extra delay
//		// cause by getting a red instead of a green is 20secs
//		totalExtraDelay += nextPhaseDuration * size;
//		
//		//
//		// for each car C moving in-range
//		// there's a delay of
//		// (20secs-the time it takes for C to reach the intersection) if C would
//		// make it through
//		// //the first _crossing-opportunities_ cars
//		// 20 secs for the rest (cause the cars in front won't move)
//		double now = state.doubleTimeSchedule.time();
//		int numberInRangeCars = inRangeCars.bag.numObjs;
//		int numberInRangeMovingCars = inRangeCars.bag.numObjs-size;
//		if (numberInRangeMovingCars != 0 && crossingOpportunitiesNextPhase != 0)
//		{
//			if(numberInRangeMovingCars <=crossingOpportunitiesNextPhase)
//				//all cars in range could pass through.
//				for (int i = 0; i < numberInRangeCars; i++)
//				{
//					Car c = (Car) inRangeCars.bag.objs[i];
//					if(c.isStoppedInQueue()) continue;
//					double delay_i = Math.max(nextPhaseDuration-c.getTimeTillIntersection(now), 0);
//					totalExtraDelay += delay_i;
//				}
//			else
//			{
//				// only the first "crossingOpportunitiesNextPhase" cars
//				// would be affected by the next RED/GREEN decision
//				Double[] keys = tmpHeapKeys;
//				Car[] values = tmpHeapValues;
//				if(keys.length<numberInRangeCars)
//				{//resize
//					keys = new Double[numberInRangeCars];
//					values = new Car[numberInRangeCars];
//				}
//
//				int index = 0;
//				for (int i = 0; i < numberInRangeCars; i++)
//				{
//					Car c = (Car) inRangeCars.bag.objs[i];
//					if(c.isStoppedInQueue()) continue;
//					values[index] = c;
//					keys[index] = new Double(c.getTimeTillIntersection(now));
//					index++;
//				}
//				Heap h = new Heap(keys, values, numberInRangeMovingCars);
//				for (int x = 0; x < crossingOpportunitiesNextPhase; x++)
//				{
//					double timeTillIntersection = ((Double)h.getMinKey()).doubleValue();
//					// Car c = (Car)
//					h.extractMin();
//					double delay_i = Math.max(nextPhaseDuration	- timeTillIntersection, 0);
//					totalExtraDelay += delay_i;
//				}
//				for(int i= numberInRangeCars - crossingOpportunitiesNextPhase;i>=0;i--)
//				{
//					double delay_i = nextPhaseDuration;
//					totalExtraDelay += delay_i;
//				}
//			}
//		}
//		if(totalExtraDelay<0)
//			throw new RuntimeException("Error: getRedOverGreenTotalDelayIncrease<0");
//		return totalExtraDelay;
//	}
}
