package sim.app.traffic;

import java.io.*;
import sim.app.traffic.load.*;
import sim.app.traffic.route.*;
import sim.engine.*;
import sim.util.*;
import ec.util.*;
import ec.multirun.*;

/**
 * This simulation is meant as a study of history-based urban traffic control.
 * Traffic lights can recognize cars, retrieve info about their past
 * (e.g. the history is stored on a black-box in your car talking to the traffic light over radio) 
 * and use that info to decide over red light-green light patterns. 
 * 
 * <p>I'm interested in fairness and paying schemes. Ultimately, I'd like to study drivers biding for the green light.
 * The code allows for a number of issues to be varied (most of them are embedded in intersections, 
 * and a few are in road links and cars):
 * <dl>
 * <dt>Party composition</dt>
 * 	<dd><ul> 
 * 	<li>all cars coming from the same street, regardless of destination (<code>OneQueueGroupIntersection</code>)
 * 	<li>all cars with non-conflicting source-destination street (<code>OneLaneOutEdgeIntersection</code>).
 * 	</ul>
 *  Each intersection has its own strategy, independent of the other intersections (the intersections and the roads are grouped together in a <code>RoadNet</code>).
 * </dd>
 * 
 * <dt>Winning policy</dt> 
 * 	<dd><ul>
 * 	<li>sum of bids
 * 	<li>sum of credit wealths for the winnig party
 *  <li>sum of cars' total waiting times (in the past)
 *  <li>sum of cars' average waiting times
 * 	</ul>
 * This is held in cars. Different cars could have different policies for bidding; otherwise all cars must be honest about waiting time stats.</dd>
 * 
 * <dt>Who Pays for Green</dt> 
 * 	<dd><ul>
 * 	<li>all the winners
 * 	<li>only the winners that made it through the intersection.
 * 	</ul></dd>
 * 
 * <dt>Green Price</dt> 
 * 	<dd><ul>
 * 	<li>fixed amount
 * 	<li>collective bidding: every car pays how much it bids (wants).
 * 	</ul></dd>
 * 
 * <dt>What the winners are buying</dt> 
 * 	<dd><ul>
 * 	<li>passage 
 * 	<li>green phase switching or extension
 * </ul> The two are quite different things (they cost differently, too) because it takes longer to go
 * through if the initial speed is 0 than when it's the cruising speed. Plus you may need more than a 
 * green phase to pass through.
 * </dd>
 * 
 * <dt>What losers Earn at Red lights</dt>
 * 	<dd><ul>
 * 	<li>a fixed amount
 * 	<li>an equal share of the winners' bids.
 * 	<li>a share of the winners' bids proportional to the losing bid
 * 	</ul></dd>
 * </dl>
 * 
 * <p><code>TraficSim</code> is the entry point of the GUI-less simulation. 
 * It holds on to the scheduler, the <code>RoadNet</code>, the <code>TrafficLoadManager</code> 
 * and a bunch of statistics tallies.
 * 
 * @author Gabriel Balan
 */
public class TrafficSim extends ParamSimState {
	public static final String P_ROOT = "trafficsim";

	public static final String P_ROAD_NET = "road-net";

	public static final String P_ITINERARY_FACTORY = "itinerary-factory";

	public static final String P_LOAD_MANAGER = "load-manager";

	public static final String P_INITIAL_WEALTH = "initial-car-wealth";

	public static final String P_WARMUP_PERIOD = "warmup-period";

	public static final String P_ANOVASTORAGE_ID = "anova-storage-id";

	public static final String P_OVERALL_DELAY = "overall-x-delay";

	public static final String P_CAR_DELAY = "car-x-delay";
	
	public static final String P_SLOWDOWN = "car-x-slowdown";
	public static final String P_TRIP_STOP_FRACTION = "car-x-stop-fraction";
	
	
	public static final String P_INDIVIDUAL_DELAY_LOG = "individual-delay-log";

	public static final String P_LOG_INDIVIDUAL_DELAYS = "log-individual-delays";

	public static final String P_TRIP_AVGD_DELAY_LOG = "trip-averaged-delay-log";

	public static final String P_LOG_TRIP_AVGD_DELAYS = "log-trip-averaged-delays";

	public static final String P_RESET_WEALTH_EVERY_INTERSECTION = "reset-wealth-after-every-intersection";
	
	public static final String P_NO_PRINTS = "no-prints";

	public static final int CAR_ORDERING = 0;

	// I want the cars stepped first, to be ready when the queues decide to pop
	// out cars.
	public static final int INTERSECTION_ORDERING = 1;

	public static final int QUEUE_ORDERING = 2;

	public static final int CHECK_ORDERING = 3;

	public RoadNet roadNet;

	/** the average delay experienced by a car between two wealth resets. */
	public TrafficTally carXDelays;
	public Bag carXDelaysByTripLength;
	public TrafficTally carXSlowdowns;
	//XP_stopping_prob per intersection
	public TrafficTally carXStopFraction;
	
	public TrafficTally overallXDelays = new TrafficTally("Overall Intersection Delay");


	public boolean printsAllowed=true;
	public DelayLog individualDelayLog = null;

	public DelayLog tripAvgdDelayLog = null;
	
	
	public void recordTripWaitingTimes(TrafficTally tripWaitingTimeTally)
	{
		double avgWT = tripWaitingTimeTally.getAverage();
		
		//carXDelays is not really needed.
		carXDelays.add(avgWT);
		
		int tripLen = tripWaitingTimeTally.count;
		carXDelaysByTripLength.resize(tripLen);
		TrafficTally tripLenTally = (TrafficTally)carXDelaysByTripLength.objs[tripLen-1];
		if(tripLenTally==null)
			carXDelaysByTripLength.objs[tripLen-1]=tripLenTally= new TrafficTally();
		if(tripLen>carXDelaysByTripLength.numObjs)
			carXDelaysByTripLength.numObjs=tripLen;
		tripLenTally.add(avgWT);
		
		if (tripAvgdDelayLog != null)
			tripAvgdDelayLog.add(avgWT);

//		tripCountSanityCheck();
	}

	public static interface DelayLog {
		void add(double delay);
	}

	public class DelayFileLog implements DelayLog {
		final PrintWriter pw;

		public DelayFileLog(File f) throws FileNotFoundException {
			pw = new PrintWriter(new FileOutputStream(f), false);
		}

		public void add(double delay) {
			if (schedule.time() > warmupPeriod)
				pw.println(delay);
		}

		public void finalize() {
			pw.close();
		}
	}

	// same pointer as schedule, but it saves time on casts
//	public DoubleTimeSchedule doubleTimeSchedule;

	public ItineraryFactory itineraryFactory;

	public TrafficLoadManager loadManager;

	public float initialWealth;

	public boolean resetWealthAfterEveryIntersection = false;

	public double warmupPeriod;

	AnovaDataStorage multirunDataStorage = null;

//static long retrieveSeed(ParameterDatabase parameters)
//{
//	String seedImg = parameters.getStringWithDefault(new Parameter(P_SEED), null, "time");
//	long seed = seedImg.equalsIgnoreCase("time")?
//					System.currentTimeMillis():
//					Long.parseLong(seedImg);
//	System.out.println("Seed=\t"+seed);
//	return seed;
//}
	
	public TrafficSim(ParameterDatabase parameters)
	{
		super(parameters);
		
		Parameter base =new Parameter(P_ROOT);
		Parameter p;
// long seed = parameters.getLongWithDefault(new Parameter(P_SEED), null,
// System.currentTimeMillis());
// random.setSeed(seed);
	   printsAllowed = ! parameters.getBoolean(new Parameter(P_NO_PRINTS),null, !printsAllowed);
		

		initialWealth = parameters.getFloat(base.push(P_INITIAL_WEALTH), null, 0);
		if(initialWealth<0)
			System.err.println("Warning: negative initial wealth");

		
		roadNet = (RoadNet)parameters.getInstanceForParameter(p = base.push(P_ROAD_NET), null, RoadNet.class);
		roadNet.setup(this, p);
		itineraryFactory = (ItineraryFactory)parameters.getInstanceForParameter(p = base.push(P_ITINERARY_FACTORY), null, ItineraryFactory.class);
		itineraryFactory.setup(this, p);
		
			
		resetWealthAfterEveryIntersection = parameters.getBoolean(base.push(P_RESET_WEALTH_EVERY_INTERSECTION), null, resetWealthAfterEveryIntersection);
		
		loadManager = (TrafficLoadManager)parameters.getInstanceForParameter(p=base.push(P_LOAD_MANAGER), null, TrafficLoadManager.class);
		loadManager.setup(this, p);
		
		warmupPeriod = parameters.getDoubleWithDefault(base.push(P_WARMUP_PERIOD), null, 0);
				
		carXDelays = new TrafficTally("Trip Delay/Intersection");
		carXDelays.setup(this, base.push(P_CAR_DELAY));
		
		carXDelaysByTripLength = new Bag();
		
		carXSlowdowns = new TrafficTally("Slowdowns");
		carXSlowdowns.setup(this, base.push(P_SLOWDOWN));
		
		carXStopFraction = new TrafficTally("XP_stopping_prob");
		carXStopFraction.setup(this, base.push(P_TRIP_STOP_FRACTION));
		
		
		overallXDelays = new TrafficTally("Overall Delay/Intersection");
		overallXDelays.setup(this, base.push(P_OVERALL_DELAY));

		if(parameters.exists(p=new Parameter(P_ANOVASTORAGE_ID)))
		{
			int instanceId = parameters.getInt(p, null);
			multirunDataStorage = AnovaDataStorage.getAnovaDataStorage(instanceId);
			java.util.Vector v = new java.util.Vector();
			overallXDelays.addToHeader(v);
			carXDelays.addToHeader(v);
			v.add("stoppedCar%");
			v.add("X_stoppedFraction");
			v.add("tripCount");
			v.add("crossingCount");
			carXSlowdowns.addToHeader(v);
			v.add("Avg_PatternChanges");
			v.add("Avg_UninterruptedGreens");
			v.add("Avg_UninterruptedReds");
			
			multirunDataStorage.initialize(v);
		}
		
		if(parameters.getBoolean(new Parameter(P_LOG_INDIVIDUAL_DELAYS), null, false))
		{
			java.io.File log = parameters.getFile(new Parameter(P_INDIVIDUAL_DELAY_LOG), null);
			try{individualDelayLog = new DelayFileLog(log);}
			catch(FileNotFoundException ex){individualDelayLog=null; ex.printStackTrace();}
		}
		
		if(parameters.getBoolean(new Parameter(P_LOG_TRIP_AVGD_DELAYS), null, false))
		{
			java.io.File log = parameters.getFile(new Parameter(P_TRIP_AVGD_DELAY_LOG), null);
			try{tripAvgdDelayLog = new DelayFileLog(log);}
			catch(FileNotFoundException ex){tripAvgdDelayLog =null; ex.printStackTrace();}
		}
		
	}	
	
	public void start() {
		if(printsAllowed)System.out.println("---------------------");
		super.start();
		resetStats();

		if (warmupPeriod > 0)
			schedule.scheduleOnce(warmupPeriod, new Steppable() {
				public void step(SimState state) {
					resetStats();
				}
			});
		roadNet.start();
		loadManager.start();
		schedulePhaseDisplayers();
	}

	public void resetStats() {
		carXDelays.reset();
		for(int i=0;i<carXDelaysByTripLength.objs.length;i++)
		{
			Object o = carXDelaysByTripLength.objs[i];
			if(o==null) continue;
			TrafficTally tally = (TrafficTally)o;
			tally.reset();
		}
		carXDelaysByTripLength.numObjs=0;
		
//		tripCountSanityCheck();
		
		carXSlowdowns.reset();
		carXStopFraction.reset();
		for (int i = 0; i < roadNet.allNodes.numObjs; i++)
			((Intersection) roadNet.allNodes.objs[i]).resetStats();
	}

	public void finish() 
	{
		loadManager.finish();

		long totalCarXings = getTotalVehicles();// I don't need this, it's
												// overallDelay.getCount()
		TrafficTally overallDelay = getOverallDelayTally();
		long totalCarStops = getTotalStoppedVehicles();

		double stoppedCarPercentage = 100d * totalCarStops / totalCarXings;

		if(printsAllowed)
		{
			System.out.println("\nSimulated time\t" + schedule.time());
			System.out.println("Ignored warmup\t" + warmupPeriod);
			System.out.println("stopped cars\t" + stoppedCarPercentage + "%");
			System.out.println("XP_stopping_prob\t" + carXStopFraction.getAverage());
		}			
		
		overallDelay.print(multirunDataStorage, printsAllowed);
		carXDelays.print(multirunDataStorage, printsAllowed);
		
		if (multirunDataStorage != null)
		{
			multirunDataStorage.add(stoppedCarPercentage); 
			multirunDataStorage.add(carXStopFraction.getAverage());	
			multirunDataStorage.add(carXDelays.count); 
			multirunDataStorage.add(overallDelay.count); 

			carXSlowdowns.print(multirunDataStorage, printsAllowed);
		
			multirunDataStorage.add(getAverageNumberOfPatternChanges());
			multirunDataStorage.add(getUniterruptedGreensTally().getAverage());
			multirunDataStorage.add(getUniterruptedRedsTally().getAverage());
		}

		// if some trips did not end, so some of the delay values reported in
		// the overall waiting time
		// were not considered for avg waiting time per intersection per trip.

		// -at the limit, the average should be the same for overall and
		// avg/trip ...
		// (the numbers come from the same distribtuion, and the extra averaging
		// changes the confidence intervals, not the expected value).
		// -but the variation should be smaller in the latter case (since I do
		// an extra level of averaging).
		// PLUS it should be closer to a normal distribution!

		// ANYWAY the average can go either way
		//
		// 10 10 10
		// 0 0 0
		// --------- ----------
		// overallAvg = 3.(3) overallAvg = 6.(6)
		// avg(avg) = 5 avg(avg) = 5
		// (plus for short runs unfinished trips are noticeable)

		// overall max should be at least the max avg!!!!
		// unless the max happens right before the warmup resetStats,
		// which clears the global stats, but not the cars
		// TODO use some sort of flag inside Car to submit avg delays
		// only for trips that started after the warmup?

		// COMPARING CC /w WFx
		// the average and stdev should be smaller across the table!!!

		// ////// System.out.println("ERR
		// "+ZeroSumEqualCompensation.TOTAL_ERROR);

		if(printsAllowed)
		{

			System.out.println();				
			Tally queueLenTally = getAverageQueueLengthTally();
			System.out.println("AVG_Q_AVG_LEN\t"+queueLenTally.getAverage());
			System.out.println("AVG_LIVE_CAR_COUNT\t"+loadManager.getAverageNumberOfLifeCars());
			System.out.println("Avg_PatternChanges\t"+getAverageNumberOfPatternChanges());
			System.out.println("Avg_UninterruptedGreens\t"+getUniterruptedGreensTally().getAverage());
			System.out.println("Avg_UninterruptedReds\t"+getUniterruptedRedsTally().getAverage());
	
			printStatsTripLength();
		}		
		//this is the last cause it clears the "time".
		super.finish();
	}

	// iterate over the intersections and add the numbers together
	public long getTotalVehicles() {
		long sum = 0;
		for (int i = 0; i < roadNet.allNodes.numObjs; i++) {
			Intersection intersection = (Intersection) roadNet.allNodes.objs[i];
			sum += intersection.getTotalVehicles();
		}
		return sum;
	}


//	public void tripCountSanityCheck()
//	{
//		int totalTripCount = carXDelays.count;
//		int totalTripCount_test = 0;
//
//		for(int i=0; i< carXDelaysByTripLength.numObjs;i++)
//		{
//			TrafficTally t = (TrafficTally)carXDelaysByTripLength.objs[i];
//			if(t==null) continue;
//			totalTripCount_test +=t.count;
//		}
//
//
//		if(totalTripCount_test != totalTripCount)
//			throw new RuntimeException("Trip count sanity check failed: "+totalTripCount_test+"!="+totalTripCount);
//	}

	public void printStatsTripLength()
	{
		int totalTripCount = carXDelays.count;
		//it better be equal to the sum of counts for all these 
		//tallies in the bag.
		int totalTripCount_test = 0;

		double globalAverage = carXDelays.getAverage();



		System.out.print("AvgDelay-byTripLen\t");
		for(int i=0; i< carXDelaysByTripLength.numObjs;i++)
		{
			TrafficTally t = (TrafficTally)carXDelaysByTripLength.objs[i];
			if(t!=null)
			{
				System.out.print(t.getAverage()+"\t");
				totalTripCount_test +=t.count;
			}
			else
				System.out.print(Double.NaN+"\t");
		}
		System.out.println();


		if(totalTripCount_test != totalTripCount)
			throw new RuntimeException("Trip count sanity check failed: "+totalTripCount_test+"!="+totalTripCount);
		/*
		 * I want to see the contribution of each trip-length class to the overall average.
		 * -They have to be proportional to the averages inside a class times the number of trips in that class.
		 * -they have to be comparable in magnitude to the overall average 
		 * 
		 * So, the "contribution" values are so that their average is the oveall average. 
		 * 
		 * It breaks if the last category has zero points in one treatment and non-zero in another.
		 */
		double correctionCoefficient = ((double)carXDelaysByTripLength.numObjs)/totalTripCount;
		double meanSanitySum = 0;
		System.out.print("Contributions2AvgDelay-byTripLen\t");
		for(int i=0; i< carXDelaysByTripLength.numObjs;i++)
		{
			
			TrafficTally tally = (TrafficTally)carXDelaysByTripLength.objs[i];
			if(tally!=null)
			{
				double contribution = tally.getAverage()* tally.count*correctionCoefficient;
				System.out.print(contribution+"\t");
				meanSanitySum+=contribution;
			}
			else
				System.out.print(Double.NaN+"\t");

		}
		System.out.println();
//		if(meanSanitySum != globalAverage * carXDelaysByTripLength.numObjs)
//			throw new RuntimeException("Mean sanity check failed: "+(meanSanitySum/carXDelaysByTripLength.numObjs)+"!="+globalAverage);
// the sanity check should allow for rounding errors (eg E-12)
		
		double globalAverageSq = globalAverage * globalAverage;
		double varianceSanitySum = 0;
		double scaling = ((double)carXDelaysByTripLength.numObjs)/(totalTripCount-1);
	
		/*
		 * I want to see the contribution of trips of a given length
		 * to the variance (NOT stdev) of trip-averaged delay.
		 * 
		 * I go with variance cause of the summation, so it's easy to identify the 
		 * contributions; not so easy to extract them from under the squared root. 
		 * Sqrt(avg x^2) >= avg x <=> avg x^2 >=(avg x)^2.
		 */
		//I also want the these values normalized by # trips in each class
		//so instead of computing them twice, I store them in this array.
		//[i guess I should have printed all these stats transposed; simpler code, simpler calculations]
		double[] avgContrib2Var = new double[carXDelaysByTripLength.numObjs];

		System.out.print("Contributions2VarianceInAvgDelay-byTripLen\t");
		for(int i=0; i< carXDelaysByTripLength.numObjs;i++)
		{
			TrafficTally t = (TrafficTally)carXDelaysByTripLength.objs[i];
			if(t==null)
			{
					System.out.print(Double.NaN+"\t");
					avgContrib2Var[i]=Double.NaN;
					continue;
			}
			double contribution= (t.sumSq - 2 * t.sum * globalAverage + t.count * globalAverageSq);
			//this has to be divided by globalN-1; 
			//Also, I want the contributios to average (not sum) to the global variance.
			contribution *=scaling;
			
			System.out.print(contribution+"\t");
			varianceSanitySum+=contribution;
			avgContrib2Var[i]=contribution/t.count;
		}
		System.out.println();
//		if(carXDelays.getVariance()*carXDelaysByTripLength.numObjs != varianceSanitySum)
//			throw new RuntimeException("Variance sanity check failed: "+carXDelays.getVariance()+"!="+(varianceSanitySum/carXDelaysByTripLength.numObjs));

		System.out.print("AvgCarContribution2VarianceInAvgDelay-byTripLen\t");
		for(int i=0; i< carXDelaysByTripLength.numObjs;i++)
		{
			System.out.print(avgContrib2Var[i]+"\t");
		}
		System.out.println();

		System.out.print("InClassVariance-byTripLen\t");
		for(int i=0; i< carXDelaysByTripLength.numObjs;i++)
		{
			TrafficTally t = (TrafficTally)carXDelaysByTripLength.objs[i];
			if(t==null)
			{
					System.out.print(Double.NaN+"\t");
					continue;
			}
			System.out.print(t.getVariance()+"\t");
		}
		System.out.println();
		
		
	}

	public long getTotalStoppedVehicles() {
		long sum = 0;
		for (int i = 0; i < roadNet.allNodes.numObjs; i++) {
			Intersection intersection = (Intersection) roadNet.allNodes.objs[i];
			sum += intersection.getStoppedVehicles();
		}
		return sum;
	}
	
	/**
	 * There are some intersections (the corners) where the pattern
	 * never changes. I could skip them when I normalize the number of
	 * pattern changes by the number of intersections.
	 */
	public double getAverageNumberOfPatternChanges() {
		long sum = 0;
		int nIntersections = roadNet.allNodes.numObjs;
		for (int i = 0; i < nIntersections; i++) {
			Intersection intersection = (Intersection) roadNet.allNodes.objs[i];
			sum += intersection.numberOfPatternChanges;
		}
		return ((double)sum)/nIntersections;
	}	

	public TrafficTally getOverallDelayTally() {
		overallXDelays.reset();
		for (int i = 0; i < roadNet.allNodes.numObjs; i++) {
			Intersection intersection = (Intersection) roadNet.allNodes.objs[i];
			overallXDelays.add(intersection.getDelayTally());
		}
		return overallXDelays;
	}
	
	public Tally queueLengths = new Tally("Queues Average Length");
	public Tally getAverageQueueLengthTally()
	{
		queueLengths.reset();
		for (int i = 0; i < roadNet.allNodes.numObjs; i++) {
			TrafficQueue[] queues = ((Intersection) roadNet.allNodes.objs[i]).getQueues();
			for(int j=queues.length-1;j>=0;j--)
			{
				TrafficQueue q = queues[j];
				if(!q.unused)
					queueLengths.add(q.queueLengthCalculator.getAverage());
			}
		}
		return queueLengths;		
	}

	public Tally queueRecentLengths = new Tally("Recent Queues Average Length");
	public Tally getRecentAverageQueueLengthTally(boolean resetBookmark)
	{
		queueRecentLengths.reset();
		for (int i = 0; i < roadNet.allNodes.numObjs; i++) {
			TrafficQueue[] queues = ((Intersection) roadNet.allNodes.objs[i]).getQueues();
			for(int j=queues.length-1;j>=0;j--)
			{
				TrafficQueue q = queues[j];
				if(!q.unused)
				{
					queueRecentLengths.add(q.queueLengthCalculator.getPostBookMarkAverage());
					if(resetBookmark) q.queueLengthCalculator.setBookMark();
				}
			}
		}		
		return queueRecentLengths;		
	}

	public TrafficTally uniterruptedGreens = new TrafficTally("Avg # consecutive green phases");
	public TrafficTally getUniterruptedGreensTally()
	{
		uniterruptedGreens.reset();
		for (int i = 0; i < roadNet.allNodes.numObjs; i++) {
			TrafficQueue[] queues = ((Intersection) roadNet.allNodes.objs[i]).getQueues();
			for(int j=queues.length-1;j>=0;j--)
			{
				TrafficQueue q = queues[j];
				if(!q.unused)
				{
					uniterruptedGreens.add(q.uninterruptedGreenTally);
				}
			}
		}		
		return uniterruptedGreens;		
	}

	public TrafficTally uniterruptedReds = new TrafficTally("Avg # consecutive red phases");
	public TrafficTally getUniterruptedRedsTally()
	{
		uniterruptedReds.reset();
		for (int i = 0; i < roadNet.allNodes.numObjs; i++) {
			TrafficQueue[] queues = ((Intersection) roadNet.allNodes.objs[i]).getQueues();
			for(int j=queues.length-1;j>=0;j--)
			{
				TrafficQueue q = queues[j];
				if(!q.unused)
				{
					uniterruptedReds.add(q.uninterruptedRedTally);
				}
			}
		}		
		return uniterruptedReds;		
	}
	// TODO the problem is that I do class casts (Intersection) n times for each
	// function
	// when a single loop would save time. Moreover, I need totalVehicles
	// both for average delay and % of stopped cars.
	//
	// On the third hand if I need average (total trip duration/trip duration
	// without delays)...

	public static void main(String[] args) {
		ParameterDatabase parameters = null;
		for (int x = 0; x < args.length - 1; x++)
			if (args[x].equals(TrafficSim.A_FILE)) {
				try {
					parameters = new ParameterDatabase(new java.io.File(
							new java.io.File(args[x + 1]).getAbsolutePath()),
							args);
				} catch (java.io.IOException ex) {
					ex.printStackTrace();
				}
				break;
			}
		if (parameters == null)
			throw new RuntimeException("No parameter file was specified.");
		double runLen = parameters.getDoubleWithDefault(MultiRun.P_LENGTH,
				null, 1000000);

		// at every moment in time, the cars will be stepped before the
		// intersections
		TrafficSim sim = new TrafficSim(parameters);
		sim.start();

		while (sim.schedule.time() < runLen)
		{			
			if (!sim.schedule.step(sim))
				break;
		}
		sim.finish();
	}

	Bag phaseDisplayers = new Bag();// TODO clean in on start

	void schedulePhaseDisplayers() {
		if (phaseDisplayers.numObjs > 0) {
			for (int i = 0; i < phaseDisplayers.numObjs; i++)
				((InfoDisplayUpdater) phaseDisplayers.objs[i]).reset();

			schedule.scheduleOnce(0.0, TrafficSim.CHECK_ORDERING,
					new Steppable() {
						public void step(sim.engine.SimState sim) {
							for (int i = 0; i < phaseDisplayers.numObjs; i++)
								((InfoDisplayUpdater) phaseDisplayers.objs[i]).add(TrafficSim.this);
							schedule.scheduleOnceIn(
									roadNet.intersectionPrototype.phaseDuration,
									this,
									TrafficSim.CHECK_ORDERING);
						}
					});
		}
	}

	static public interface InfoDisplayUpdater {
		public void reset();

		public void add(TrafficSim sim);
	}

	public void addPhaseListener(InfoDisplayUpdater listener) {
		phaseDisplayers.add(listener);
	}
	

}
