package sim.app.traffic;
import sim.util.*;
import java.util.*;
import ec.multirun.*;

/**
 * The simplest most basic functionality of SSJ's Tally.
 * 
 * You add numbers, it maintains the count, sum and sum of squared values
 * in order to compute the sample mean and stdev.
 * 
 * What's new and special is that tallies can be combined.
 * 
 * I also maintain min and max, but they might just be dead weight slowing me down.
 * 
 * @author Gabriel Balan
 */
public class TrafficTally extends Tally{
	
	public TrafficTally(){super();}
	public TrafficTally(String name){super(name);}

	public void add(TrafficTally t)
	{
		sum		+= t.sum;
		count	+= t.count;
		sumSq 	+= t.sumSq;
		
		if(t.min<min) min = t.min;
		if(t.max>max) max = t.max;
	}
	
	/** clears this Tally and fills it with the 
	 * "difference between NEW_TALLY and OLD_TALLY.
	 * (this assumes that OLD_TALLY is an older version of NEW_TALLY,
	 * i.e. NEW_TALLY is an OLD_TALLY to wich samples were added)
	 * 
	 * min (max) might not be accurate if oldT.min==newT.min (max)
	 **/
	public void minus(TrafficTally newT, TrafficTally oldT)
	{
		if(oldT.count>newT.count || oldT.min<newT.min || oldT.max>newT.max ||  newT.sumSq<oldT.sumSq)
		{
			throw new RuntimeException("newT is not compatible with oldT");
		}
		count = newT.count-oldT.count;
		sum = newT.sum-oldT.sum;
		sumSq = newT.sumSq-oldT.sumSq;
		min = newT.min;//not accurate
		max = newT.max;//not accurate
	}
	public void set(TrafficTally t)
	{
		count = t.count;
		min = t.min;
		max = t.max;
		sum = t.sum;
		sumSq = t.sumSq;
	}
	
	/** 
	 * In cause you don't want to get NAN if count is 0, provide that value as the parameter.
	 */
	public double getAverageWithDefault(double defaultValue){return (count==0) ? defaultValue : sum/count;}

	/**
	 * @return the average of the sample U{value}
	 * NON-DESTRUCTIVE: the value is not actually added to the sample
	 * 
	 * if the sample points are durations (time interval lengths), 
	 * you might need the average including the current interval, even if it's not over yet.
	 * ("assuming it would be over now, what would be the average?") 
	 * 
	 * I don't need a "default value" in this case.
	 */
	public double getAverageWithExtraValue(double extraValue){return (sum+extraValue)/(count+1);}

	public void addToHeader(Vector header)
	{
		if(printAVG) 	header.add(name+"_avg");
		if(printSTDEV)	header.add(name+"_stdev");
		if(printMIN)	header.add(name+"_min");
		if(printMAX)	header.add(name+"_max");		
		if(printCV)		header.add(name+"_CV");		
		if(printFI)		header.add(name+"_FI");
	}
	public void print(AnovaDataStorage ads)	
	{
		print(ads, true);
	}
	public void print(AnovaDataStorage ads, boolean printsAllowed)
	{
		if(printsAllowed) print();
		if(ads==null) return;
		if(printAVG) 	ads.add(getAverage());
		if(printSTDEV)	ads.add(getStDev());
		if(printMIN)	ads.add(getMin());
		if(printMAX)	ads.add(getMax());		
		if(printCV)		ads.add(getCoefVar());
		if(printFI)		ads.add(getFairnessIndex());
	}

}
