package sim.app.traffic.analysis;

/**
 * This graph shows the average score of a car
 * at the k-th stop in its trip under the 1/TL score-fn.
 * 
 * <p>Score<sub>TL(k)</sub> = sum<sub>n=k ... max trip length</sub> P(n|k)/n
 * 
 * <p>P(n|k) = P(n & k) / P(k) <br>
 * P(k) = sum<sub>j=k...n</sub> P(k|j) = sum<sub>j=k...n</sub> P( k & j) Pj <br>
 * P(n&k) = P(n) P(k|n) <br>
 * 
 * P(k|n) = 1/n (when k<=n)<br>
 * "You're on a n-intersection trip; you get to an intersection...
 * what's the probability that's k-th one?"
 * 
 * <p>If you accept this, then 										<br>
 * P(n&k) = Pn/n													<br>
 * P(k) = sum_{j=k}^{max} Pj^2 / j									<br>
 * P(n|k) = P(n & k) / P(k) = Pn/n  / [ sum<sub>m=k ... max</sub> (Pm^2 / m) ]<br>
 * Let's call  [ sum_{m=k}^{max} (Pm^2 / m) ]  D(k)					<br>
 * Score_TL(k) = sum_{n=k}^{max} Pn/(n^2)  / D(k) 
 * <--!
 * 					max
 * 					---  Pn     1
 * 					>	 --- x ---
 * 					---  n      n
 * 					n=k
 * Score_TL(k) =  -----------------		
 *  					max		 
 *  					---  Pm^2	 
 *  					>	 ----
 *  					---   m
 * 					    m=k
 * -->
 * 
 * <p>Now I can tell the story differenty: a car's saying it's on
 * a n-intersection trip. What's the chance it's at the k-th 
 * intersection? Well, under LSF a car's priority gets smaller and 
 * smaller, so a car is expected to spend more and more time at 
 * later intersections. Therefore a car at the last intersection
 * might be asked more times than at the beginning of it.
 * 
 * <p>Sanity check: k = max trip length.
 * <p>Score_TL(max) should be TLF(max) = 1/max
 * <p>Score_TL(max)=Pmax/max^2/Pmax/Pmax *max = 1/Pmax/max
 * <p>So THE ABOVE FORMULA IS WRONG at least a little!!!!!
 * 
 * 			sum_{j=k}^{max} TLfn(j) x Pj
 * fn(k) = -----------------------
 * 			   sum_{j=k}^{max} Pj
 * 
 * 
 * @author Gabriel Balan
 *
 */
public class AvgTLofLSF
{
	static double[] TRIPLENPROBS_10x10G = new double[] { 6.75, 8.98, 10.47, 11.36,
		11.53, 11.14, 10.21, 8.82, 6.92, 5.01, 3.54, 2.35, 1.47, 0.84,
		0.42, 0.17, 0.04 };
	
	
	
	
	/**
	 * Computes the expected trip length of a car given that it reached K-th 
	 * intersection of its trip.
	 */
	public static double[] computeExpected_TL_ofLSF(double[] tripLenProbs, double[] results)
	{
		int n=tripLenProbs.length;
		double[] function = new double[tripLenProbs.length];		
		for(int i=0; i<n;i++) function[i]=i+1;
		return computeExpected_TLfn_ofLSF(tripLenProbs, function, results);
	}


	/**
	 * Computes the expected value of 1/TL for a car that reached K-th intersection of its trip.
	 */
	public static double[] computeExpected_1OverTL_ofLSF(double[] tripLenProbs, double[] results)
	{
		int n=tripLenProbs.length;
		double[] function = new double[tripLenProbs.length];		
		for(int i=0; i<n;i++) function[i]=1d/(i+1);
		return computeExpected_TLfn_ofLSF(tripLenProbs, function, results);
	}

	/**
	 * Computes the the expected value of a given function of the trip-length of a car given that it reached K-th 
	 * intersection of its trip.
	 * 
	 * 			sum_{j=k}^{n} TLfn(j) x W_j
	 * fn(k) = -----------------------
	 * 			   sum_{j=k}^{n} W_j
	 * 
	 * Where wj = tripLenProbs[j]
	 * and they don't have to add to 1.
	 */
	public static double[] computeExpected_TLfn_ofLSF(double[] tripLenProbs, double[] tlScore, double[] results)
	{
		int n=tripLenProbs.length;
		if(results==null)
			results = new double[n];
		double weightedSumSoFar=0;
		double weightSumSoFar=0;
		for(int j=n-1;j>=0;j--)
		{
			double weight_j = tripLenProbs[j];
			weightSumSoFar+=weight_j;
			weightedSumSoFar+=weight_j*tlScore[j];
			results[j]=(weightSumSoFar==0)?0:weightedSumSoFar/weightSumSoFar;
		}
		return results;
	}


	public static void print(double[] values)
	{
		for (int i = 0; i < values.length; i++)
		{
			System.out.println(values[i]);
		}
	}
	
	public static void printNormalized(double[] values)
	{
		double max = values[0];
		for (int i = 0; i < values.length; i++)
		{
			System.out.println((values[i] / max));
		}
	}
	
	public static void printNormalizedOn1Line(double[] values)
	{
		double max = values[0];
		for (int i = 0; i < values.length; i++)
		{
			System.out.print((values[i] / max)+"\t");
		}
		System.out.println();
	}	
	public static void main(String[] args)
	{
//		double[] values = computeExpected_1OverTL_ofLSF(TRIPLENPROBS_10x10G, null);
//		printNormalizedOn1Line(values);
		double[] values = computeExpected_TL_ofLSF(TRIPLENPROBS_10x10G, null);
		print(values);

	}
}
