package sim.app.traffic.analysis;




import ec.util.*;
import sim.app.traffic.*;
import sim.app.traffic.car.Car;
import sim.engine.*;
import sim.field.network.*;
import sim.util.*;
import java.io.*;
import java.util.*;

/**
 * <code>Car</code> extension used for debugging and 
 * getting stats on the amount of credits a car end its trip with.
 * 
 * <p>My purpose for this is to establish a correlation between
 * cars' average waiting times (or whatever stats you use to evaluate controllers)
 * and the amount of credits cars finish their trips with.
 * 
 * @author Gabriel Balan
 *
 */
public class CreditCheckCar extends Car {

	Bag log;

	TrafficSim tState;
	
	public void setup(ParamSimState state, Parameter base) 
	{
		super.setup(state, base);
		tState = (TrafficSim)state;
	}


	private class LogEntry
	{
		final int date;
		final double wealth;
		final String obs;
		public LogEntry(String obs)
		{
			this.date = (int)tState.schedule.time();
			this.obs=obs;
			this.wealth= CreditCheckCar.this.wealth;
		}
		public String toString()
		{
			return date+"\t[$"+wealth+"]\t"+obs;
		}
	}
	public void addLogEntry(String text)
	{
		if(!traceable)return;
		//if(time < xxxx) return;	
		log.add(new LogEntry(text));
	}
	
	
	static int ID = 0;
	String name;
	boolean traceable = false;
	public Object simpleClone()
	{
		CreditCheckCar newc = (CreditCheckCar)super.simpleClone();
		if(ID==1988)
		{
			newc.traceable=true;
			int iniBagSize=10;
			newc.log = new Bag(iniBagSize);
		}
		newc.name = "C"+(ID++);
		return newc;
	}
	
	public int getPosInQueue()
	{
		if(fsaState!=WAIT_IN_QUEUE) return -1;//not waiting in queue just yet.
		Iterator i = currentQueue.iterator();
		int pos=1;
		while(i.hasNext())
		{
			if(i.next()==this) 
				return pos;
			pos++;
		}
		throw new RuntimeException("Car not found");
	}
	
	protected void getInTheQueue(TrafficSim ts) {
		Edge e = getCurrentSegment();
		super.getInTheQueue(ts);
		if(e== getCurrentSegment())
		{
			//I'm still in the queue.
			addLogEntry("got stuck in "+queueDescription()+" @pos"+getPosInQueue());
		}
	}
	
	protected void getOutOfTheQueue(TrafficSim ts)
	{
		if(traceable)
			currentQueue.unregisterDiaryKeepingCars();
		super.getOutOfTheQueue(ts);
		addLogEntry("get out of queue");
	}
	
	protected void travelSegmentInsideSensorRange(TrafficSim ts) 
	{
		super.travelSegmentInsideSensorRange(ts);
		if(traceable && currentQueue !=null)
		{
			currentQueue.registerDiaryKeepingCars();
			addLogEntry("in sensor range of "+queueDescription());
		}
	}
	
	protected String queueDescription()
	{
		return currentIntersection+
				"-q"+ ((currentQueue==null)? 
						"--": 
						""+currentQueue.index+"-"+currentQueue.getColorName());
	}

		
	public void resetWealthAndWaitingTimes(TrafficSim ts, float newWealth)
	{
		double awt = waitingTimeTally.getAverage();
		if(awt>0)
		{
			//initialWealth+compensation-5Xtrip_len = final wealth.
			double compensation = wealth -newWealth+5*waitingTimeTally.count;
			double compPerIntersection = compensation/waitingTimeTally.count;

			boolean printLog = false;
			if(compPerIntersection>(Math.ceil(awt/20)*20))
			{
				printLog = true;
				//print !!!!!
			}
			if(wealth>=90)	printLog=true;
			
			if(printLog)
			{
				System.out.println(name+"\t["+waitingTimeTally.count+"x]$"+compPerIntersection+"/x<-?->"+awt+"sec/x");
				for(int i=0;i<log.numObjs;i++)
					System.out.println("\t"+log.objs[i]);
				System.out.println("------------------");
			}
		}
		if(log!=null)log.clear();
		super.resetWealthAndWaitingTimes(ts, newWealth);
	}
	
//	//called just once, for the prototype.
//	public void finish()
//	{
//		if(multiRunHistogram.numObjs<histoCounts.numObjs)
//		{
//			multiRunHistogram.resize(histoCounts.numObjs);
//			multiRunHistogram.numObjs=histoCounts.numObjs;
//		}
//		int total = 0;
//		for(int i=0;i<histoCounts.numObjs;i++)
//			total+=histoCounts.objs[i];
//		System.out.println(total);
//		for(int i=0;i<histoCounts.numObjs;i++)
//			multiRunHistogram.objs[i]+=((double)histoCounts.objs[i])/total;
//	}


	public static void main(String[] args)
	{
		ParameterDatabase parameters=null;
		for(int x=0;x<args.length-1;x++)
			if (args[x].equals(MultiRun.A_FILE))
			{
				try
				{
					parameters=new ParameterDatabase(new File(new File(args[x+1]).getAbsolutePath()),args);
					break;
				}catch(FileNotFoundException e)
				{ 
					throw new RuntimeException("A File Not Found Exception was generated upon" +
						"reading the parameter file \"" + args[x+1] + "\".\nHere it is:\n" + e); 
				}catch(IOException e)
				{
					throw new RuntimeException("An IO Exception was generated upon reading the" +
						"parameter file \"" + args[x+1] +"\".\nHere it is:\n" + e);
				}
			}
		if (parameters==null)
			throw new RuntimeException("No parameter file was specified." );

		parameters.set(new Parameter("trafficsim.load-manager.car-generator.car"),
						"sim.app.traffic.analysis.CreditCheckCar");
	
		int nRuns = parameters.getInt(MultiRun.P_RUNS, null, 1);
		if(nRuns<1) throw new RuntimeException("Too few runs.");


		double runLen = parameters.getDouble(MultiRun.P_LENGTH, null, 0);
		if(runLen<=0) throw new RuntimeException("Too short of a run.");
		
//		MersenneTwisterFast rng = new MersenneTwisterFast();
		
		for(int i=0;i<nRuns;i++)
		{
			long seed_i = 1;//rng.nextLong();
			System.err.println("DON'T FORGET TO MAKE NEW SEEDS!!!!");
			System.out.println("\tRun "+i+"\tSeed "+seed_i);
			parameters.set(new Parameter(TrafficSim.P_SEED), seed_i+"");
			TrafficSim ts = new TrafficSim(parameters);
			Schedule schedule = ts.schedule;
			ts.start();
			while(schedule.time()<runLen)
				if (!schedule.step(ts)) break;
			ts.finish();
		}
//
//		//average the percentages: so far I added them up, now I divide by their number
//		for(int i=0;i<multiRunHistogram.numObjs;i++)
//		{
//			multiRunHistogram.objs[i]/=nRuns;
//		}
//		
//		
//		XYSeries dataSeries = new XYSeries("avg-%",false, true);
//		//i provide the data in increasing X order, so:
//		//autosort-false, allow duplicates-true
//		for(int i=0;i<multiRunHistogram.numObjs;i++)
//			dataSeries.add(i+1, multiRunHistogram.objs[i]);
//		
//		//use this one to output the histogram
//		XYSeriesCollection data = new XYSeriesCollection(dataSeries);
//		
//		JFreeChart chart = ChartFactory.createXYBarChart("Distribution", "trip length", false, "%", data, PlotOrientation.VERTICAL, false, false, false);
//		XYPlot plot = chart.getXYPlot();
//		((NumberAxis)plot.getRangeAxis()).setAutoRangeIncludesZero(false);
//		((NumberAxis)plot.getDomainAxis()).setTickUnit(new NumberTickUnit(1));
//
//
//		XYItemRenderer itemrenderer = plot.getRenderer();
//		itemrenderer.setItemLabelsVisible(true);
//		itemrenderer.setItemLabelGenerator(new XYItemLabelGenerator()
//		{
//////			NumberFormat formatter = NumberFormat.getPercentInstance();
//  				DecimalFormat formatter = new DecimalFormat("#0.00%");
//			public String generateLabel(XYDataset dataset, int series,  int item) 
//			{
//				return  formatter.format(dataset.getY(series, item));
//			}
//		});
//
//		print the average and median.
//
//		PDFPrinter.printChartToPDF(chart, "credit2grave-distrib.pdf");
	}
}
