package sim.app.traffic.analysis;
import java.io.*;
import java.util.zip.*;
import org.jfree.chart.*;
import org.jfree.data.xy.*;
import org.jfree.chart.axis.LogarithmicAxis;
import org.jfree.chart.axis.NumberAxis;
import org.jfree.chart.axis.NumberTickUnit;
import org.jfree.chart.plot.*;

import sim.util.jfreechart.PDFPrinter;
/**
 * This takes the following java properties as switches:
 * <dl>
 * <dt>extension</dt><dd>e.g. ".txt" or ".txt.gz" (without the quotes)</dd>
 * <dt>log-range-axis</dt><dd>"true" or "false" (without the quotes).</dd>
 * </dl>
 *
 * @author Gabriel Balan
 */
public class CrossTreatmentTripDelayByLengthAnalysis
{
	XYSeriesCollection plots[];

	boolean	logRangeAxis = false;

	/** override this */
	protected TripDelayByLength getTreatmentAnalyzer(BufferedReader br)throws IOException
	{
		return new TripDelayByLength(br);
	}

	/** override this */
	protected int getPlotCount(){return 5;}
	/** override this */
	protected void addTreatmentToPlots(BufferedReader br, String name)throws IOException
	{
		TripDelayByLength analyzer = getTreatmentAnalyzer(br);
		plots[0].addSeries(analyzer.getAveragesLine(name));
		plots[1].addSeries(analyzer.getContributionsToAverageLine(name));
		plots[2].addSeries(analyzer.getStdevLine(name));
		plots[3].addSeries(analyzer.getContributionsToStdevLine(name));
		plots[4].addSeries(analyzer.getMedianLine(name));
	}
	/** override this */
	protected void writePlots()
	{
		PDFPrinter.printChartToPDF(getChart(plots[0], "Expected Trip-averaged Delay"), "TripAvgd-AvgsLinePlot.pdf");
		PDFPrinter.printChartToPDF(getChart(plots[1], "Contributions to Average Trip-averaged Delay"), "Contribs2TripAvgd-AvgsLinePlot.pdf");

		PDFPrinter.printChartToPDF(getChart(plots[2], "Stdev Trip-averaged Delay"), "TripAvgd-StdevsLinePlot.pdf");
		PDFPrinter.printChartToPDF(getChart(plots[3], "Contributions to Stdev Trip-averaged Delay"), "Contribs2TripAvgd-StdevsLinePlot.pdf");


		PDFPrinter.printChartToPDF(getChart(plots[4], "Median of Trip-averaged Delay"), "TripAvgd-MediansLinePlot.pdf");
	}

	public CrossTreatmentTripDelayByLengthAnalysis(String args[])throws IOException
	{
		plots = new XYSeriesCollection[getPlotCount()];
		for(int i=0;i<plots.length;i++) plots[i]=new XYSeriesCollection();
		String extension = System.getProperty("extension", ".txt.gz");
		int extensionLength = extension.length();

		logRangeAxis = Boolean.getBoolean("log-range-axis");
		
		int n = args.length;
		
		for(int i=0;i<n;i++)
		{
			String fileName = args[i];
			if(fileName.endsWith(extension))
			{
				addTreatmentToPlots(fileName.endsWith(".gz") ?
										new BufferedReader(new InputStreamReader(new GZIPInputStream(new FileInputStream (fileName)))):
										new BufferedReader(new FileReader(fileName)),
									fileName.substring(0, fileName.length()-extensionLength));

			}
			else throw new RuntimeException("Extension ["+extension+"] not matching "+fileName);


		}
		writePlots();
	}
	protected JFreeChart getChart(XYSeriesCollection dataset, String title)
	{
		return getChart(dataset, title, "Delay (sec)");
	}
	
	protected JFreeChart getChart(XYSeriesCollection dataset, String title, String yAxisCaption)
	{
		
		JFreeChart chart = ChartFactory.createXYLineChart(
											title, "Trip length", yAxisCaption,
											dataset, PlotOrientation.VERTICAL, true, false, false);
											//legend yes, tooltips no, url... no
		if(logRangeAxis)
		{
			LogarithmicAxis logAx=new LogarithmicAxis("Delay (sec)");
			logAx.setStrictValuesFlag(false);
//			logAx.setAllowNegativesFlag();
			chart.getXYPlot().setRangeAxis(logAx);
		}
		((NumberAxis)chart.getXYPlot().getRangeAxis()).setAutoRangeIncludesZero(false);
		((NumberAxis)chart.getXYPlot().getDomainAxis()).setTickUnit(new NumberTickUnit(1));
		return chart;
	}
	
	public static void main(String args[])throws IOException
	{
		new CrossTreatmentTripDelayByLengthAnalysis(args);
	}

}
