package sim.app.traffic.analysis;


import sim.util.jfreechart.*;
import sim.util.*;
import sim.util.stats.*;

import ec.util.*;
import java.io.*;
import org.jfree.chart.*;
import org.jfree.chart.axis.*;
import org.jfree.chart.plot.*;
import org.jfree.chart.labels .*;
import org.jfree.chart.renderer.xy.*;
import org.jfree.data.statistics.*;
import org.jfree.data.xy.*;

import javax.swing.*;
import java.awt.*;
import java.util.zip.*;

/**
 * This takes the file name of a TripData log as argument and the following java properties as switches:
 * <dl>
 * <dt>tripData.extended-format</dt><dd> (default = false)</dd>
 * <dt>outlier.coefficient</dt><dd> numerical value (default = 1.5);</dd>
 * 
 * <dt>show.gui</dt><dd> put up a jfreechart graph, or just produce stat outputs in files (default = false)</dd>
 * <dt>outlier.ignore-3+s</dt><dd> skip outlier triplets; that is you don't print than two occurences of an outlier in a row. 
 * 		Use it if you want to speed jfreechart up (default = false);</dd>
 * 
 * <dt>compare.means</dt><dd>Does a anova-like stat analysis (aside from box-and-whisker plots);</dd>
 * <dt>number.comparisons</dt><dd>it refers to the bonferroni correction (default = number.treatments * (number.treatments-1)/2 </dd>
 * <dt>number.treatments</dt><dd></dd>
 * <dt>alpha</dt><dd>significance level (default = 0.01)</dd>
 * <dt>verbosity</dt><dd>get printouts of what's going on under the hood (default = false)</dd>
 * </dl>
 * 
 *	
 * @author Gabriel Catalin Balan
 */
public class TripDelayByLength implements TripDataAnalyzer
{
	Bag tallies;


	public void add(TripData tripRecord)
	{
		DoubleBag delayBag = tripRecord.delayBag;
		int n = delayBag.numObjs;
		if(n>0)
		{
			add(n, tripRecord.getAverageDelay());
		}
	}
	
	public void add(int n, double value)
	{
		tallies.resize(n);
		MyTallyInArray tally_n;
		if ((tally_n=(MyTallyInArray)tallies.objs[n-1])==null)
		{
			tallies.objs[n-1] = tally_n = new MyTallyInArray();
			tally_n.setName(n+"Xings");
		}
		tally_n.add(value);

	}

//	static final int AVERAGES = 0;
//	static final int MEDIANS = 1;
//
//	public XYSeries getLine(int statistic, String seriesName)
//	{
//		switch (statistic)
//		{
//			case AVERAGES: return getAveragesLine(seriesName);
//			case MEDIANS:  return getMedianLine(seriesName);
//			default: throw new RuntimeException("Invalid statistic ID: "+statistic);
//		}
//	}

	
	/**
	 * I want to see the contribution of each trip-length class to the overall average.
	 * -They have to be proportional to the averages inside a class times the number of trips in that class.
	 * -they have to be comparable in magnitude to the overall average 
	 * 
	 * So, the "contribution" values are so that their average is the oveall average. 
	 * 
	 * It breaks if the last category has zero points in one treatment and non-zero in another.
	 */
	public XYSeries getContributionsToAverageLine(String seriesName)
	{
		XYSeries data = new XYSeries(seriesName,false,true);
		//-No thanx, I'll give you the data already sorted.
		//-I don't actually plan to have duplicates, 
		//but I don't want JFreeChart to get even slower by activelly preventing them.
		
		int totalDataPoints = 0;
		for(int i=0;i<tallies.numObjs;i++)
		{
			MyTallyInArray tally=(MyTallyInArray)tallies.objs[i];
			if(tally!=null)
				totalDataPoints+=tally.numberObs();
		}
		double correctionCoefficient = ((double)tallies.numObjs)/totalDataPoints;

		for(int i=0;i<tallies.numObjs;i++)
		{
			MyTallyInArray tally=(MyTallyInArray)tallies.objs[i];
			if(tally!=null)
				data.add(i+1, tally.average()* tally.numberObs()*correctionCoefficient, false);//don't notify.
		}
		return data;
	}


	/**
	 * I want to see the contribution of each trip-length class to the overall average.
	 * -They have to be proportional to the averages inside a class times the number of trips in that class.
	 * -unitary area under the curve.
	 * 
	 */
	public XYSeries getNormalizedContributionsToAverageLine(String seriesName)
	{
		XYSeries data = new XYSeries(seriesName,false,true);
		//-No thanx, I'll give you the data already sorted.
		//-I don't actually plan to have duplicates, 
		//but I don't want JFreeChart to get even slower by activelly preventing them.
		
		double sum = 0;
		for(int i=0;i<tallies.numObjs;i++)
		{
			MyTallyInArray tally=(MyTallyInArray)tallies.objs[i];
			if(tally!=null)
				sum+=tally.sum();//tally.average()* tally.numberObs()
		}
		double correctionCoefficient = 1d/sum;

		for(int i=0;i<tallies.numObjs;i++)
		{
			MyTallyInArray tally=(MyTallyInArray)tallies.objs[i];
			if(tally!=null)
				data.add(i+1, tally.sum()*correctionCoefficient, false);//don't notify.
		}
		return data;
	}	
	
//	/**
//	 * I want to see the contribution of each trip-length class to the "snap-shot"  efficiency
//	 * (longer trips take longer to finish, so the distribution of trip-lengths of the cars
//	 * you see at some moment in time is different than the one you get from getContributionsToAverageLine).
//	 * 
//	 * -They have to be proportional to the averages inside a class times the average number of cars on the field on trips of that length. 
//	 * 
//	 * -they have to be comparable in magnitude to the overall average 
//	 * OR
//	 * -they can be normalized to unitary area under the curve, so I can put cross-load lines on the same plot.
//	 * 
//	 */
//	public XYSeries getContributionsToEnvironmentNormalizedLine(String seriesName)
//	{
//		XYSeries data = new XYSeries(seriesName,false,true);
//		//-No thanx, I'll give you the data already sorted.
//		//-I don't actually plan to have duplicates, 
//		//but I don't want JFreeChart to get even slower by activelly preventing them.
//		
//		int totalDataPoints = 0;
//		for(int i=0;i<tallies.numObjs;i++)
//		{
//			MyTallyInArray tally=(MyTallyInArray)tallies.objs[i];
//			if(tally!=null)
//				totalDataPoints+=tally.numberObs();
//		}
//		double correctionCoefficient = ((double)tallies.numObjs)/totalDataPoints;
//
//		for(int i=0;i<tallies.numObjs;i++)
//		{
//			MyTallyInArray tally=(MyTallyInArray)tallies.objs[i];
//			if(tally!=null)
//				data.add(i+1, tally.average()* tally.numberObs()*correctionCoefficient, false);//don't notify.
//		}
//		return data;
//	}

	public XYSeries getAveragesLine(String seriesName)
	{
		XYSeries data = new XYSeries(seriesName,false,true);
		//-No thanx, I'll give you the data already sorted.
		//-I don't actually plan to have duplicates, 
		//but I don't want JFreeChart to get even slower by activelly preventing them.
		
		for(int i=0;i<tallies.numObjs;i++)
		{
			MyTallyInArray tally=(MyTallyInArray)tallies.objs[i];
			if(tally!=null)
				data.add(i+1, tally.average(), false);//don't notify.
		}
		return data;
	}


	public XYSeries getMedianLine(String seriesName)
	{
		XYSeries data = new XYSeries(seriesName,false,true);
		//-No thanx, I'll give you the data already sorted.
		//-I don't actually plan to have duplicates, 
		//but I don't want JFreeChart to get even slower by activelly preventing them.
		
		for(int i=0;i<tallies.numObjs;i++)
		{
			MyTallyInArray tally=(MyTallyInArray)tallies.objs[i];
			if(tally!=null)
				data.add(i+1, tally.quantile(.5), false);//don't notify.
		}
		return data;
	}
	
	public XYSeries getStdevLine(String seriesName)
	{
		XYSeries data = new XYSeries(seriesName,false,true);
		//-No thanx, I'll give you the data already sorted.
		//-I don't actually plan to have duplicates, 
		//but I don't want JFreeChart to get even slower by activelly preventing them.
		
		for(int i=0;i<tallies.numObjs;i++)
		{
			MyTallyInArray tally=(MyTallyInArray)tallies.objs[i];
			if(tally!=null)
				data.add(i+1, tally.standardDeviation(), false);//don't notify.
		}
		return data;
	}
	
	public XYSeries getContributionsToStdevLine(String seriesName)
	{
		XYSeries data = new XYSeries(seriesName,false,true);
		//-No thanx, I'll give you the data already sorted.
		//-I don't actually plan to have duplicates, 
		//but I don't want JFreeChart to get even slower by activelly preventing them.
		
		int totalDataPoints = 0;
		for(int i=0;i<tallies.numObjs;i++)
		{
			MyTallyInArray tally=(MyTallyInArray)tallies.objs[i];
			if(tally!=null)
				totalDataPoints+=tally.numberObs();
		}
		double correctionCoefficient = ((double)tallies.numObjs)/totalDataPoints;

		for(int i=0;i<tallies.numObjs;i++)
		{
			MyTallyInArray tally=(MyTallyInArray)tallies.objs[i];
			if(tally!=null)
				data.add(i+1, tally.standardDeviation()* tally.numberObs()*correctionCoefficient, false);//don't notify.
		}
		return data;
	}
	
	
	public XYSeries getContributionsToVarianceLine(String seriesName)
	{
		double allTallySum = 0;
		double allTallySampleCount=0;

		for(int i=0;i<tallies.numObjs;i++)
		{
			MyTallyInArray tally=(MyTallyInArray)tallies.objs[i];
			if(tally!=null)
			{
				allTallySum += tally.sum();
				allTallySampleCount += tally.numberObs();
				
			}
		}
		double allTallyAverage = allTallySum / allTallySampleCount;
		double scaling  = ((double)tallies.numObjs)/(allTallySampleCount-1);
		//variance = Sum[(x-m)^2]/(N-1)
		//so scaling = 1/(N -1).
		// Also, I want the contributions to average to the global variance, 
		//not to sum to the global variance. => scaling = #tallies/(N-1).
		
		XYSeries data = new XYSeries(seriesName,false,true);
		//-No thanx, I'll give you the data already sorted.
		//-I don't actually plan to have duplicates, 
		//but I don't want JFreeChart to get even slower by activelly preventing them.
		
		
		
		for(int i=0;i<tallies.numObjs;i++)
		{
			MyTallyInArray tally=(MyTallyInArray)tallies.objs[i];
			if(tally!=null)
			{
				double[] observations = tally.getArray().elements();
				double contrib=0;
				int n_i = observations.length;
				for(int j=n_i-1;j>=0;j--)
				{
					double x = observations[j];
					double diff = x - allTallyAverage;
					contrib+=diff*diff;
				}
//				if(sum!=tally.sum())
//					throw new RuntimeException("Sum error!!");
//				if((sumsq-sum*sum/n_i)/(n_i-1)!=tally.variance())
//					throw new RuntimeException("Variance error!!");
					
					
				contrib*=scaling;
				data.add(i+1, contrib, false);//don't notify.
			}
		}
		return data;
	}
	
	public BoxAndWhiskerSeries getBoxAndWhiskerDataSet(double outlierCoefficient, boolean skipOutliersTriplets)
	{
//		if(dataSet==null) 
		BoxAndWhiskerSeries dataSet = new BoxAndWhiskerSeries("",tallies.numObjs);
		dataSet.setOutlierCoefficient(outlierCoefficient);
		BoxAndWhiskerItem bnwi;
		for(int i=0;i<tallies.numObjs;i++)
		{
			MyTallyInArray tally=(MyTallyInArray)tallies.objs[i];
			if(tally!=null)
			{
				bnwi=sim.util.jfreechart.BoxAndWhiskerCalculator.calculateBoxAndWhiskerStatistics(tally, outlierCoefficient, skipOutliersTriplets);
				dataSet.add(new Integer(i+1), 
							bnwi,
							false);
				System.out.println((i+1)+"-"+tally.numberObs()+"->"+bnwi.getMean());
			}
		}
		return dataSet;
	}
	
	public void compareMeans()
	{
		int tallyCount = 0;
		for(int i=0;i<tallies.numObjs;i++)
			if(tallies.objs[i] !=null)
				tallyCount++;
		StatAnalysis sa = new StatAnalysis(tallyCount);
		for(int i=0;i<tallies.numObjs;i++)
		{
			MyTallyInArray tally=(MyTallyInArray)tallies.objs[i];
			if(tally!=null)
				sa.addTreatment(tally);
		}
		double alpha = Double.parseDouble(System.getProperty("alpha", "0.01"));
		int ntreatments = Integer.parseInt(System.getProperty("number.treatments", ""+tallyCount));
		int ncomparisons = Integer.parseInt(System.getProperty("number.comparisons", ""+(ntreatments*(ntreatments-1)/2)));
		boolean verbosity = Boolean.getBoolean("verbosity");
		sa.pseudoANOVAEquivalenceClasses(alpha, true, StatAnalysis.BONFERRONI_CORRECTION, ncomparisons, verbosity);
	}

	
	public String getValueAxisCaption(){return "Average Waiting Time per intersection";}
	public String getChartTitle(){return "Trip waiting-time vs. trip length";}
	
	
	//TODO I should add a line of stdevs to the Box&Whisker set.
	public JFreeChart createBoxAndWiskerChart(double outlierCoefficient, boolean skipOutliersTriplets)
	{
		
		NumberAxis timeAxis = new NumberAxis("Trip length (number of crossed intersections)");		
		NumberAxis valueAxis = new NumberAxis(getValueAxisCaption());
		timeAxis.setAutoRangeIncludesZero(false);
		timeAxis.setTickUnit(new NumberTickUnit(1));
		
		XYBoxAndWhiskerRenderer renderer = new XYBoxAndWhiskerRenderer(20);

		BoxAndWhiskerXYToolTipGenerator bwTooltipgenerator = new BoxAndWhiskerXYToolTipGenerator();
		renderer.setToolTipGenerator(bwTooltipgenerator);
		
		XYPlot plot = new XYPlot(getBoxAndWhiskerDataSet(outlierCoefficient, skipOutliersTriplets), timeAxis, valueAxis, renderer);
		JFreeChart chart = new JFreeChart(	getChartTitle(), 
											JFreeChart.DEFAULT_TITLE_FONT, plot, false);//no legend
		return chart;
	}


	public void showAndPrint(double outlierCoefficient, boolean skipOutliersTriplets, boolean showGUI, String pdfName)
	{
		JFreeChart chart = createBoxAndWiskerChart(outlierCoefficient, skipOutliersTriplets);
		ChartPanel panel = new ChartPanel(chart);
		

		
		if(pdfName!=null)
		{
			PDFPrinter.printChartToPDF(panel, pdfName);
//			System.exit(0);

		}
		if(showGUI)
		{
			JFrame frame = new JFrame();

			frame.setResizable(true);
			frame.getContentPane().setLayout(new BorderLayout());
			frame.getContentPane().add(panel, BorderLayout.CENTER);
			frame.setTitle(getChartTitle());
			frame.pack();
			frame.setVisible(true);
		}
		else
			System.exit(0);	
	}
	
	
	
	public void setup(ParamSimState state, Parameter base){}
	public Parameter defaultBase(){return null;}
	public Object simpleClone(){return null;}

	public TripDelayByLength()
	{
		tallies = new Bag();
	}
	
	public TripDelayByLength(BufferedReader br)throws IOException
	{
		this();
		int max = 0;
		
		boolean tripDataExtendedFormat = Boolean.getBoolean("tripData.extended-format");

	
		TripData td;
		while((td=TripData.read(br,tripDataExtendedFormat))!=null)
		{
			add(td);
			if(td.delayBag.numObjs>max) max = td.delayBag.numObjs;
		}
		tallies.numObjs = max;
//		tallies.shrink(max);

		br.close();
	}

	public TripDelayByLength(String[] args)throws FileNotFoundException, IOException
	{
		this(args[0].endsWith(".gz") ?
				new BufferedReader(new InputStreamReader(new GZIPInputStream(new FileInputStream (args[0])))):
				new BufferedReader(new FileReader(args[0])));
		
		double outlierCoefficient = Double.parseDouble(System.getProperty("outlier.coefficient", "1.5"));
		
		boolean compareMeans = Boolean.getBoolean("compare.means");
		if(compareMeans)
			compareMeans();
		boolean showGui = Boolean.getBoolean("show.gui");
		// JFreeChart shows 2 circles no matter how many equal outliers you have, 
		// so one can choose to ignore the rest, to speed JFreeChart's chart-building 
		boolean skipOutliersTriplets = Boolean.getBoolean("outlier.ignore-3+s");
		showAndPrint(outlierCoefficient,skipOutliersTriplets, showGui, (args.length==1)? null:args[1]);
	}
	public void finish(){}
	
	public static void main(String[] args) throws FileNotFoundException, IOException
	{
		new TripDelayByLength(args);
	}
}
