package sim.app.traffic.analysis;


import org.jfree.chart.*;
import org.jfree.chart.plot.*;
import org.jfree.chart.axis.*;
import org.jfree.chart.labels.*;
import org.jfree.data.xy.*;
import org.jfree.chart.renderer.xy.*;


import ec.util.MersenneTwisterFast;
import ec.util.Parameter;
import ec.util.ParameterDatabase;
import sim.app.traffic.*;
import sim.app.traffic.car.Car;
import sim.app.traffic.route.Itinerary;
import sim.engine.*;
import sim.field.network.Edge;
import sim.util.*;
import sim.util.jfreechart.*;

import java.io.*;
import java.text.*;

/**
 * @author Gabriel Balan
 */
public class TripLenDistribAnalyzerCar extends Car {

	IntBag histoCounts;
	static DoubleBag multiRunHistogram = new DoubleBag();
	
	//there are no 0-length trips, so first counter is for 1-stop trips.

	public void setup(ParamSimState state, Parameter base) 
	{
		super.setup(state, base);
		//allocate the histogram data thing
		
		histoCounts = new IntBag(16);
		
	}
	
	int tripLen = -1;
		
	public void resetWealthAndWaitingTimes(TrafficSim ts, float wealth) 
	{
		super.resetWealthAndWaitingTimes(ts, wealth);
		//no point in logging just build up the histogram
		//I need a data structured shared between the instances 
						
		if(tripLen> histoCounts.numObjs)
		{
			histoCounts.resize(tripLen);
			histoCounts.numObjs = tripLen;
		}
		
		if(tripLen>0)//we don't care about the 0-stop trips
			histoCounts.objs[tripLen-1]++;
		tripLen = -1;
	}

	public void deploy(TrafficSim tState, Intersection startingPoint, Itinerary itinerary)
	{
		currentItinerary = itinerary;
		currentIntersection = startingPoint;

		//the first item in the itinerary is for the starting intersection
		//where you don't wait, so it does not count
		int nextsegIndex;
		while((nextsegIndex=itinerary.getNextSegmentIndex())!=Itinerary.NO_NEXT_SEGMENT)
		{
			tripLen++;
			Edge segment = (Edge)currentIntersection.ioi.out.objs[nextsegIndex];	
			currentIntersection = (Intersection)(segment.to());
		}
		
		
		tState.schedule.scheduleOnceIn(1, new Steppable()
		{
			public void step(SimState s)
			{
				//sort of re-writing finishTrip()
				destinationsReached++;
				TrafficSim ts = (TrafficSim)s;
				ts.loadManager.destinationReached(TripLenDistribAnalyzerCar.this);						
			}
		});
	}
	
	//called just once, for the prototype.
	public void finish()
	{
		super.finish();
		if(multiRunHistogram.numObjs<histoCounts.numObjs)
		{
			multiRunHistogram.resize(histoCounts.numObjs);
			multiRunHistogram.numObjs=histoCounts.numObjs;
		}
		int total = 0;
		for(int i=0;i<histoCounts.numObjs;i++)
			total+=histoCounts.objs[i];
//		System.out.println(total);
		for(int i=0;i<histoCounts.numObjs;i++)
			multiRunHistogram.objs[i]+=((double)histoCounts.objs[i])/total;
	}


	public static DoubleBag getTripLenDistribution(String[] args)
	{
		ParameterDatabase parameters=null;
		for(int x=0;x<args.length-1;x++)
			if (args[x].equals(MultiRun.A_FILE))
			{
				try
				{
					parameters=new ParameterDatabase(new File(new File(args[x+1]).getAbsolutePath()),args);
					break;
				}catch(FileNotFoundException e)
				{ 
					throw new RuntimeException("A File Not Found Exception was generated upon" +
						"reading the parameter file \"" + args[x+1] + "\".\nHere it is:\n" + e); 
				}catch(IOException e)
				{
					throw new RuntimeException("An IO Exception was generated upon reading the" +
						"parameter file \"" + args[x+1] +"\".\nHere it is:\n" + e);
				}
			}
		if (parameters==null)
			throw new RuntimeException("No parameter file was specified." );
			
		int nRuns = parameters.getInt(MultiRun.P_RUNS, null, 1);
		if(nRuns<1) throw new RuntimeException("Too few runs.");


		double runLen = parameters.getDouble(MultiRun.P_LENGTH, null, 0);
		if(runLen<=0) throw new RuntimeException("Too short of a run.");
		
		MersenneTwisterFast rng = new MersenneTwisterFast();

		for(int i=0;i<multiRunHistogram.numObjs;i++)
			multiRunHistogram.objs[i]=0;

		
		for(int i=0;i<nRuns;i++)
		{
			long seed_i = rng.nextLong();

			parameters.set(new Parameter(TrafficSim.P_SEED), seed_i+"");
			TrafficSim ts = new TrafficSim(parameters);
			Schedule schedule = ts.schedule;
//			System.out.println("\tRun "+i+"\tSeed "+seed_i);
			ts.start();
			while(schedule.time()<runLen)
				if (!schedule.step(ts)) break;
			ts.finish();
		}


		//average the percentages: so far I added them up, now I divide by their number
		for(int i=0;i<multiRunHistogram.numObjs;i++)
		{
			multiRunHistogram.objs[i]/=nRuns;
		}

		return 	multiRunHistogram;
	}


	public static void main(String[] args)
	{

		getTripLenDistribution(args);
		
		XYSeries dataSeries = new XYSeries("avg-%",false, true);
		//i provide the data in increasing X order, so:
		//autosort-false, allow duplicates-true
		for(int i=0;i<multiRunHistogram.numObjs;i++)
			dataSeries.add(i+1, multiRunHistogram.objs[i]);
		
		//use this one to output the histogram
		XYSeriesCollection data = new XYSeriesCollection(dataSeries);
		
		JFreeChart chart = ChartFactory.createXYBarChart("Distribution", "trip length", false, "%", data, PlotOrientation.VERTICAL, false, false, false);
		XYPlot plot = chart.getXYPlot();
		((NumberAxis)plot.getRangeAxis()).setAutoRangeIncludesZero(false);
		((NumberAxis)plot.getDomainAxis()).setTickUnit(new NumberTickUnit(1));


		XYItemRenderer itemrenderer = plot.getRenderer();
		itemrenderer.setItemLabelsVisible(true);
		itemrenderer.setItemLabelGenerator(new XYItemLabelGenerator()
		{
////			NumberFormat formatter = NumberFormat.getPercentInstance();
  				DecimalFormat formatter = new DecimalFormat("#0.00%");
			public String generateLabel(XYDataset dataset, int series,  int item) 
			{
				return  formatter.format(dataset.getY(series, item));
			}
		});


		PDFPrinter.printChartToPDF(chart, "trip-len-distrib.pdf");
	}
}
