package sim.app.traffic.car;

import sim.app.traffic.route.*;
import sim.app.traffic.info.*;
import sim.app.traffic.*;
import sim.field.network.*;
import sim.engine.*;
import sim.util.*;
import ec.util.*;

/**
 * This is the base class for cars (drivers). 
 * <code>Car</code> interacts with queues at the intersections 
 * and with the links (roads) between them.
 * The traffic light moves cars from queues at its intersection.
 * 
 * <code>Car</code> objects do not interact with each other; they travel 
 * road segments by scheduling themselves to be woken up at the other
 * end after a delay equal to the product of their max speed and length of the road segment.
 * 
 * Actually, the traffic light controlling the intersection at the end of a road segment 
 * might not sense cars as soon as they enter the segment. This distinction
 *  divides a road segment into 2 parts: the first part is "out of sensor range," 
 *  the second part is "in sensor range." Therefore a car entering a road segment 
 *  schedules itself to be awoken when it enters the sensor range (so it can inform
 *  the traffic light of its presence, intensions, bid/wealth, etc), and then 
 *  it schedules again for when it actually reaches the intersection (and 
 *  enters a queue). 
 *  
 * <ul>
 * <li>There problably should be a <code>Car</code> subclass could set the speed based on how many other cars
 * are on the road (the more congested the road, the slower you go).
 * 
 * <li>Another <code>Car</code> subclass could use a MICRO simulation along the roads
 * (one would also need a different kind of roads, so cars can "see" other cars, 
 * change lanes, etc.).
 * </ul>
 * 
 * <p>I use TRIP with 2 meanings:
 * <ul>
 * <li>ODTrip: what's going in between the Origin and Destination
 * <li>WRTrip: a sequence of trips (in the O&rarr;D sense) between two wealth resets.
 * </ul>
 * 
 * <p>Note: On the last segment of a trip I am NOT picked up by the sensors.
 * 
 * 
 * @author Gabriel Balan
 */
public class Car implements Steppable, Parametrizable {
	public final static String P_CAR = "car";

	double oneOverSpeed;

	double intersectionCrossingDuration;
	
	//TODO should I move it somewhere else? AS it only makes sense for closed modes, so...
	public static final String P_SKIP_LAST_SEGMENT = "skip-last-segment";
	boolean skipLastSegment=false;




	// this is essentially TrafficSim.intersectionCrossingDuration, but if that
	// turns into a distribution, this is already deployed
	// on a car-by-car basis.
	/* this is needed by the load manager to decide when to reset the wealth. */
	public int destinationsReached = 0;

	public int[] indicesInBags = new int[CarBagWithO1Remove.NUM_BAGS];

	public Parameter defaultBase() {
		return new Parameter(P_CAR);
	}

	public Object simpleClone() {
		try {
			Car newc = (Car) clone();
			newc.indicesInBags = new int[CarBagWithO1Remove.NUM_BAGS];
			newc.waitingTimeTally = new TripWaitingTimesTally();
			return newc;
		} catch (CloneNotSupportedException ex) {
			return null;
		}
	}

	// int speed (preffered / max speed?)
	// the segments have int lengths, but
	// the time to traverse a segment might not be int.
	// see particle-hopping for values/ideas.

	// /
	// / STATE related fields and constants
	// /
	public static final int TRANSIT_SEGMENT_OUTSIDE_SENSOR_RANGE = 0;

	public static final int TRANSIT_SEGMENT_INSIDE_SENSOR_RANGE = 1;

	public static final int TRANSIT_LAST_SEGMENT = 3;

	public static final int WAIT_IN_QUEUE = 2;

	public static final String[] stateNames = new String[] {
			"outside sensor range", "inside sensor range", "in a queue",
			"last segment" };

	protected int fsaState;
	protected String getStateName(){ return stateNames[fsaState];}
	

	double momentEnteringRange;

	double momentEnteringQueue;

	// used to detect if the vehicle stopped or not
	// (for percentage of stopped vehicles)
	protected double getMomentEnteringQueue() {return momentEnteringQueue;}

	protected TrafficQueue currentQueue;

	// right now the queue schedules the cars
	// when they can go though the intersection
	// so the payment is not handeled though step().

	public float wealth;// number of credits.

	/** @deprecated */
	float MIN_BID = 1;

	public float currentBid = 1;

	// I went with floats so I can implement "split the credits from the winners
	// to the losers"
	// 1. the flops are slower than int ops, but I already work with doubles for
	// time, so...
	// 2. I could use a special CreditManager object in each car (with different
	// implementations for int and float and different biding behaviors

	/**
	 * This could be an increasing function of the waiting duration. Right now
	 * it bids minimum all the time.
	 * 
	 * Somebody should make sure that no car goes into debt (bidding more than
	 * its wealth).
	 * 
	 * TODO make sure the bids are stricty positive under
	 * ZeroSumBidProportionalCompensation, cause you'll stay on 0 forever; and
	 * don't bid everything on PayBid either.
	 */
	public float getBid(TrafficSim ts) {
		return (wealth > MIN_BID) ? MIN_BID : wealth;
	}

	public void setup(ParamSimState state, Parameter base) {
		TrafficSim ts = (TrafficSim) state;
		destinationsReached = 0;
		if (!ts.roadNet.directed) {
			System.err
					.println("This is not yet working well with an undirected graph");
		}
		// speed = ts.carSpeedsDistribution.getDouble();
		// the speed will be set by the CarGenerator.
		intersectionCrossingDuration = ts.roadNet.intersectionCrossingDuration;
		
		skipLastSegment = state.parameters.getBoolean(base.push(P_SKIP_LAST_SEGMENT), 
													  defaultBase().push(P_SKIP_LAST_SEGMENT), 
													  skipLastSegment);
	}
	
	//refers to the sequence of trips since the last wealth reset.
	double wrtrip_DurationWhenAlone=0;
	double wrtrip_StartingMoment = Double.NaN;

	public void resetWealthAndWaitingTimes(TrafficSim ts, float wealth) {
		this.wealth = wealth;
		double now = ts.schedule.getTime();

		if (waitingTimeTally.getCount() > 0) {
			
			ts.recordTripWaitingTimes(waitingTimeTally);
//			double avgWT = waitingTimeTally.getAverage();
//			ts.carXDelays.add(avgWT);
//			if (ts.tripAvgdDelayLog != null)
//				ts.tripAvgdDelayLog.add(avgWT);
				
			double slowdown = (now - wrtrip_StartingMoment)/wrtrip_DurationWhenAlone;
			if(ts.carXSlowdowns!=null)
				ts.carXSlowdowns.add(slowdown);
				
			ts.carXStopFraction.add(waitingTimeTally.getStoppedPercentage());
			
			waitingTimeTally.reset();
		}
		//this got moved inside the if
		//(no reason to reset it if nothing happend to the tally)
//		waitingTimeTally.reset();

		wrtrip_DurationWhenAlone = 0;
		wrtrip_StartingMoment = now;
	}

	protected void recordCarDelay(TrafficSim ts, double delay) {
		currentQueue.recordCarDelay(delay);
		waitingTimeTally.add(delay);
		if (ts.individualDelayLog != null)
			ts.individualDelayLog.add(delay);
	}

	/** this wakes up the car and gives it a new destination */
	public void deploy(TrafficSim tState, Intersection startingPoint,
			Itinerary itinerary) {
		currentItinerary = itinerary;
		currentIntersection = startingPoint;
		
		
		nextSegmentFromIndex = currentItinerary.getNextSegmentIndex();

		currentSegment = (Edge) currentIntersection.ioi.out.objs[nextSegmentFromIndex];
		currentSegmentToIndex = currentSegment.indexTo();
		// won't work well /w undirected graphs

		SegmentInfo si = (SegmentInfo) currentSegment.info;
		double segLen = si.length;
		double sensorRange = si.sensorRange;

		// schedule this car to be stepped when it's done traversing the segment.
		double now = tState.schedule.time();
		// System.out.println(now+"\tdeploy");

		if(skipLastSegment && currentItinerary.lastSegment())//0-intersection trip
		{
			//Got to skip the last segment
			//call finishTrip()
			//but finishTrip(tState) does destinationReached++, unregisterSegment.
			//TODO I should have a switch whether 0 intersection trips should count.
			finishTrip(tState);
			return;
		}

		// // the waiting time tally gets reset when the wealth is reset.
		si.registerOutOfRange(this);

		if (segLen == sensorRange) {
			travelSegmentInsideSensorRange(tState);
			return;
		} else
			fsaState = TRANSIT_SEGMENT_OUTSIDE_SENSOR_RANGE;

		if(Double.isNaN(wrtrip_StartingMoment))
			wrtrip_StartingMoment = now;
		double durationWhenAlone = (segLen - sensorRange) * oneOverSpeed;
		wrtrip_DurationWhenAlone += durationWhenAlone;
		double actualDuration = durationWhenAlone;//here you can make the speed depend on density.
		tState.schedule.scheduleOnce(	now + actualDuration, 
										TrafficSim.CAR_ORDERING, 
										this);
	}

	public class TripWaitingTimesTally extends TrafficTally
	{
		int nonZeroValues = 0;
		public void add(double value)
		{
			super.add(value);
			if(value!=0) nonZeroValues++;
		}
		public void reset()
		{
			super.reset();
			nonZeroValues=0;	
		}
		public float getStoppedPercentage()
		{
			return (float)nonZeroValues/count;
		}
	}
	public TripWaitingTimesTally waitingTimeTally;


	private Edge currentSegment;
	protected Edge getCurrentSegment(){return currentSegment;}

	private int currentSegmentToIndex;

	private int nextSegmentFromIndex;

	public Intersection currentIntersection;

	protected Itinerary currentItinerary;
	public Itinerary getItinerary(){ return currentItinerary;}

	public void step(SimState state) {
		TrafficSim ts = (TrafficSim) state;
		// System.out.println(ts.doubleTimeSchedule.getTime()+"\t"+this+" post
		// "+stateNames[fsaState]+"\t"+wealth+" credits");
		switch (fsaState) {
		case WAIT_IN_QUEUE:// I was waiting, now I am out of the Queue
			getOutOfTheQueue(ts);
			break;
		case TRANSIT_SEGMENT_OUTSIDE_SENSOR_RANGE:// I was travelling though
													// the segment, now I'll
													// register into a queue
			travelSegmentInsideSensorRange(ts);
			break;
		case TRANSIT_SEGMENT_INSIDE_SENSOR_RANGE:
			getInTheQueue(ts);
			break;
		case TRANSIT_LAST_SEGMENT:
			finishTrip(ts);
			break;
		}
	}

	protected void getOutOfTheQueue(TrafficSim ts) {
		double now = ts.schedule.time();
		recordCarDelay(ts, now - momentEnteringQueue);
		currentQueue.unregisterInRange(this);

		currentSegment = (Edge) currentIntersection.ioi.out.objs[nextSegmentFromIndex];
		currentSegmentToIndex = currentSegment.indexTo();

		SegmentInfo si = (SegmentInfo) currentSegment.info;
		double segLen = si.length;
		double sensorRange = si.sensorRange;
		
		// I should move it when the car just gets in range.
		if (ts.resetWealthAfterEveryIntersection)
			wealth = ts.initialWealth;

		if(skipLastSegment && currentItinerary.lastSegment())
		{
			//Got to skip the last segment
			//schedule finishTrip() for after crossing the intersection
			fsaState = TRANSIT_LAST_SEGMENT;
			double durationWhenAlone = intersectionCrossingDuration;
			wrtrip_DurationWhenAlone += durationWhenAlone;
			double actualDuration = durationWhenAlone;//here you can make the speed depend on density.
			ts.schedule.scheduleOnce(now + actualDuration, TrafficSim.CAR_ORDERING, this);
			return;
		}

		si.registerOutOfRange(this);

		fsaState = TRANSIT_SEGMENT_OUTSIDE_SENSOR_RANGE;
		// assuming intersectionCrossingDuration !=0
		
		
		double durationWhenAlone = (segLen - sensorRange) * oneOverSpeed + intersectionCrossingDuration;
		wrtrip_DurationWhenAlone += durationWhenAlone;
		double actualDuration = durationWhenAlone;//here you can make the speed depend on density.

		ts.schedule.scheduleOnce(now + actualDuration, TrafficSim.CAR_ORDERING, this);
		// see me when I get picked up by the next intersection's sensors


	}

	protected void travelSegmentInsideSensorRange(TrafficSim ts) {
		momentEnteringRange = ts.schedule.time();
		((SegmentInfo) currentSegment.info).registerInRange(this);
		currentIntersection = (Intersection) (currentSegment.to());
		nextSegmentFromIndex = currentItinerary.getNextSegmentIndex();

		double sensorRange = ((SegmentInfo) currentSegment.info).sensorRange;
		if (nextSegmentFromIndex == Itinerary.NO_NEXT_SEGMENT) {
			if (sensorRange == 0) {
				finishTrip(ts);
				return;
			}
			// else TRANSIT_LAST_SEGMENT for the next sensorRange*oneOverSpeed
			// ... seconds
			fsaState = TRANSIT_LAST_SEGMENT;
		} else {
			fsaState = TRANSIT_SEGMENT_INSIDE_SENSOR_RANGE;
			currentQueue = currentIntersection.getQueue(currentSegmentToIndex, nextSegmentFromIndex);
			currentQueue.registerInRange(this);
			if (sensorRange == 0) {
				getInTheQueue(ts);
				return;
			}
		}
		
		
		double durationWhenAlone = sensorRange * oneOverSpeed;
		wrtrip_DurationWhenAlone += durationWhenAlone;
		double actualDuration = durationWhenAlone;//here you can make the speed depend on density.

		ts.schedule.scheduleOnceIn(actualDuration, this, TrafficSim.CAR_ORDERING);
	}

	protected void getInTheQueue(TrafficSim ts) {
		((SegmentInfo) currentSegment.info).unregister(this);

		momentEnteringQueue = ts.schedule.time();// now

		fsaState = WAIT_IN_QUEUE;
		if (!currentQueue.enqueueCar(this))//if I did not get stopped, I go ahead.
			getOutOfTheQueue(ts);
	}

	public double getTimeInCurrentQueue(double now) {
		if (fsaState == WAIT_IN_QUEUE)
			return now - momentEnteringQueue;
		return 0;
	}

	public double getTimeInCurrentRange(double now) {
		return now - momentEnteringRange;
	}

	/**
	 * This refers to the OD trip.
	 */
	private void finishTrip(TrafficSim ts) {
		destinationsReached++;
		
		if(!skipLastSegment)
			((SegmentInfo) currentSegment.info).unregister(this);
		// if(numberIntersectionsThisTrip>0)
		// ts.tripXDelays.add(totalDelayThisTrip/numberIntersectionsThisTrip);
		ts.loadManager.destinationReached(this);
	}

	/**
	 * This is for cleanup. If there's no info reset, the manager already visits
	 * the live cars to make them report their xp-stats
	 * 
	 * I only added this cause I want to close a GZIP stream in a extension of
	 * Car. (the car prototype, not all cars)
	 */
	public void finish() {}
	
	/**
	 * I assume the car's in range. 
	 * (fsa==TRANSIT_SEGMENT_INSIDE_SENSOR_RANGE)
	 * but not already stopped.
	 */
	public double getTimeTillIntersection(double now)
	{
		double sensorRange = ((SegmentInfo) currentSegment.info).sensorRange;
		double inRangeDuration = sensorRange * oneOverSpeed;
		return momentEnteringRange + inRangeDuration - now;
	}
	
	public boolean isStoppedInQueue(){return (fsaState == WAIT_IN_QUEUE);}	

}
