package sim.app.traffic.car;
import sim.app.traffic.analysis.CreditCheckCar;
import sim.util.*;

/**
 * This offers O(1) remove because the objects keep their indices in the Bag.
 * <code>Bag.remove()</code> moves the last element in the place of the deleted one;
 * here the moved element is informed of the change.
 * 
 * <p>Objects might be included in more such bags, so each object should maintain multiple indices.
 * Theoretically they  should implement an interface such as <code>int getIndex(BagWithO1Remove b)</code>.
 * 
 * <p>I want to avoid the hash look-ups, so I associate indices with <code>Bag</code> instances, 
 * so the objects might use arrays rather than hash tables.
 * In general the arrays might be full of holes, but here I sacrifice generality completely and 
 * have it specifically taylored for this traffic app:
 * 
 * Instead of <code>int getIndex(int bagIndex)</code>, 
 * the cars offer their arrays of indices [in each Bag] for direct access.
 * 
 * So far cars can be in 3 (types of) bags:
 * <ul>
 *  <li> All cars in sensor-range of a <code>TrafficQueue</code> are held in a <code>CarBagWithO1Remove</code> with 
 *  	<code>bagIndex = CarBagWithO1Remove.QUEUE_BAG</code>
 * 	This is very special, cause normally each instace would have its own private index.
 *  <li> <code>TrafficLoadManager</code> keeps all live cars in a <code>CarBagWithO1Remove</code> with 
 *  	<code>bagIndex = CarBagWithO1Remove.ALIVE_BAG</code>
 *  <li> Each road segment has a <code>SegmentInfo</code> that gets informed of the cars coming and going. 
 *  (For instance <code>CarCountSegmentInfo</code> lets the GUI display how many cars are on each road segment)
 *  CarBagSegmentInfo holds on to cars in a <code>CarBagWithO1Remove</code> with 
 *  	<code>bagIndex = CarBagWithO1Remove.SEGMENTINFO_BAG</code>
 * </ul>
 * 
 * <p>This class aggregates a <code>Bag</code> rather than extending <code>Bag</code>, 
 * so the user is not tempted to call <code>reverse</code> or other unsupported methods of <code>Bag</code>.
 * 
 * @author Gabriel Balan
 */
public class CarBagWithO1Remove 
{
	static public int NUM_BAGS = 3;
	static public int QUEUE_BAG = 0;
	static public int ALIVE_BAG = 1;
	static public int SEGMENTINFO_BAG = 2;
	
	
	final int bagIndex;
	
	//this is public against my better judgement, in the iterest of speed (direct access) 
	final public Bag bag;
	public CarBagWithO1Remove(int index) {
		bag = new Bag();
		bagIndex = index;
	}

	public CarBagWithO1Remove(int index, int capacity) {
		bag = new Bag(capacity);
		bagIndex = index;
	}

	public void add(Car c)
	{
		bag.add(c);
		c.indicesInBags[bagIndex]=bag.numObjs-1;
	}
	public void remove(Car c)
	{
		int c_index = c.indicesInBags[bagIndex];
		//TODO remove this after you're sure it works
		
		if(bag.numObjs<=c_index)
			throw new RuntimeException("car.index >= numObjs!");	
		if(bag.objs[c_index]!=c)
			throw new RuntimeException("bug in remove(Car)!");
		bag.remove(c_index);
		if(c_index<=bag.numObjs-1) 
			((Car)bag.objs[c_index]).indicesInBags[bagIndex] = c_index;
	}
	public Car get(int index)
	{
		return (Car)bag.objs[index];
	}
	
	public int size()
	{
		return bag.numObjs;
	}
	public void clear(){bag.numObjs=0;}
	
	public float compensateFixedAmount(float payoff)
	{
		for(int i=0;i<bag.numObjs;i++)
		{
			Car c = (Car)bag.objs[i];
			c.wealth+=payoff;
		}
		return bag.numObjs*payoff;
	}
	
	
	public float compensateFixedAmountWithDiary(float payoff)
	{
		for(int i=0;i<bag.numObjs;i++)
		{
			Car c = (Car)bag.objs[i];
			c.wealth+=payoff;
			if(c instanceof CreditCheckCar)
				((CreditCheckCar)c).addLogEntry("Compensated $"+payoff+" [$"+c.wealth+"]");
		}
		return bag.numObjs*payoff;
	}	

	/**
	 * I assume that Intersection refreshed tmpTotalBids by calling getWaitingCarsTotalBids()
	 * this time step to determine the winner 
	 * @param share = how much it pays per unit of bid.
	 */	
	public void compensateBidPropotionally(float share)
	{
		for(int i=0;i<bag.numObjs;i++)
		{
			Car c = (Car)bag.objs[i];
			c.wealth+=c.currentBid * share;
		}
	}

	public void compensateBidPropotionallyWithDiary(float share)
	{
		for(int i=0;i<bag.numObjs;i++)
		{
			Car c = (Car)bag.objs[i];
			float lumpSum = c.currentBid * share;
			c.wealth+=lumpSum;
			if(c instanceof CreditCheckCar)
				((CreditCheckCar)c).addLogEntry("Compensated $"+lumpSum+" [$"+c.wealth+"]");

		}
	}

	public void broadcastDiaryEntryWithQposition(String text)
	{
		for(int i=0;i<bag.numObjs;i++)
		{
			Object o = bag.objs[i];
			if(o instanceof CreditCheckCar)
			{
				CreditCheckCar c = (CreditCheckCar)o;
				int pos = c.getPosInQueue();
				c.addLogEntry(text+", "+(pos==-1? "still moving":" stopped in position "+pos));	
			}
		}		
	}

	public void broadcastDiaryEntry(String text)
	{
		for(int i=0;i<bag.numObjs;i++)
		{
			Object o = bag.objs[i];
			if(o instanceof CreditCheckCar)
				((CreditCheckCar)o).addLogEntry(text);	
		}		
	}

}
