package sim.app.traffic.grid;
import sim.app.traffic.*;
import sim.field.network.*;
import sim.util.*;
import ec.util.*;
/**
 * Generalized Intersection: there is a TrafficQueue for each pair
 * incoming edge - outgoing edge. 
 * 
 * This is more sensible than <code>OneQeueGroupIntersection</code>: 
 * a set of non-conflicting cars are allowed in the intersection, 
 * regardless of the street they are coming from. 
 * 
 * <p>There is a limitation, however:  two queues flowing into the same out edge
 * is considered a conflict. So the incoming edges are split into as many lanes as
 * outgoing edges, but every outgoing edge has only one lane.
 * 
 * <p>Unlike the simplified mechanism, this one allows two way traffic
 * (e.g. N and S)
 * 
 * <p>However, this one requires the planar graph information.
 * 
 * @author Gabriel Balan
 */
public class GeneralizedGridIntersection extends OneLaneOutEdgeIntersection
										implements GridNode 
{	
	
	boolean uturns=true;
	public static final String P_UTURNS = "U-turns";
	public static final String P_GENGRID_INTERSECTION = "generalized-grid-intersection";
	public Parameter defaultBase(){return new Parameter(P_GENGRID_INTERSECTION);}
	public void setup(ParamSimState state, Parameter base) 
	{
		super.setup(state, base);
		uturns = state.parameters.getBoolean(base.push(P_UTURNS),
												defaultBase().push(P_UTURNS),
												uturns);
	}
	
	Int2D coords;
	public Int2D getCoords(){return coords;}
	public void setPosition(int x, int y){this.coords = new Int2D(x, y);}
	int[] outEdgeIndices;
	int[] inEdgeIndices;
	public int[] getOutEdgeIndices(){return outEdgeIndices;}
	public int[] getInEdgeIndices(){return inEdgeIndices;}
	public Network.IndexOutIn getIOI(){return ioi;}

	public void netSetup(	TrafficSim ts, 
							Network.IndexOutIn ioi, 
							int nin, 
							int nout,
							int sin, 
							int sout,
							int ein, 
							int eout,
							int win, 
							int wout)
	{
		netSetup(ioi);
		outEdgeIndices = new int[4];
		outEdgeIndices[N] = nout;
		outEdgeIndices[S] = sout;
		outEdgeIndices[W] = wout;
		outEdgeIndices[E] = eout;

		inEdgeIndices = new int[4];
		inEdgeIndices[N] = nin;
		inEdgeIndices[S] = sin;
		inEdgeIndices[W] = win;
		inEdgeIndices[E] = ein;

		int ins = ioi.in.numObjs;
		int outs = ioi.out.numObjs;	
		

		EdgeIndex[] order = new EdgeIndex[ins+outs];
		int nextIndex = 0;
		if(nin>=0) order[nextIndex++] = new EdgeIndex(nin, true);
		if(nout>=0) order[nextIndex++] = new EdgeIndex(nout, false);
		if(ein>=0) order[nextIndex++] = new EdgeIndex(ein, true);
		if(eout>=0) order[nextIndex++] = new EdgeIndex(eout, false);
		if(sin>=0) order[nextIndex++] = new EdgeIndex(sin, true);
		if(sout>=0) order[nextIndex++] = new EdgeIndex(sout, false);
		if(win>=0) order[nextIndex++] = new EdgeIndex(win, true);
		if(wout>=0) order[nextIndex++] = new EdgeIndex(wout, false);		

		if(!uturns)
		{
			IntBag uturns = new IntBag(4);
			if(nin>=0 && nout>=0) uturns.add(getQueueId(nin, nout, outs));
			if(sin>=0 && sout>=0) uturns.add(getQueueId(sin, sout, outs));
			if(ein>=0 && eout>=0) uturns.add(getQueueId(ein, eout, outs));
			if(win>=0 && wout>=0) uturns.add(getQueueId(win, wout, outs));			
			netSetup(ts, ioi, getGroups(ioi.in.numObjs, ioi.out.numObjs, order, uturns.toArray()));
		}
		else
			netSetup(ts, ioi, order);
	}
}
