package sim.app.traffic.grid;

import sim.app.traffic.*;
import sim.app.traffic.car.Car;
import sim.app.traffic.route.*;
import sim.util.Int2D;
import sim.field.network.*;
import ec.util.*;
import sim.util.*;
/**
 * 
 * The itinerary could be any path inside the rectangle defined by 
 * the four coordinates (the source and destination), with the restriction 
 * the that path consists of the "minimum number of blocks"!
 * Expected intersection waiting times are of no concern here.
 * 
 * <p>Fewest turns (1 turn tops, since we're in the grid scenario).
 * There are three variations:
 * <ol>
 * <li> randomly choose which dimension to travel first.
 * <li> shorter leg first
 * <li> longer leg first
 * <li>	avoid center
 * </ol>
 * 
 * <p>The motivation for AVOID CENTER is that the center of the city is usually congested, so you want to avoid it.
 * It is implemented as "take the route with the turning point furthest from the center."
 * This implementation is not perfect; as shown by the following anomaly:
 * consider a 6x5 grid, S->D itinerary, 1 is closer to C than 2, yet the itinerary going through 1 is cleary better  
 * <font face="courier"><!-- I just needed a fixed-width font to draw the grid -->
 * <br>0-0-0-0-0
 * <br>| | | | |
 * <br>0-0-0-0-0
 * <br>| | | | |
 * <br>0-0-0-0-0
 * <br>| | C | |
 * <br>0-D-0-0-2
 * <br>| | | | |
 * <br>0-1-0-0-S
 * <br>| | | | |
 * <br>0-0-0-0-0
 * </font>
 * 
 * <p>TODO find a better "test" for the AVOID-CENTER behavior (i.e. address the anomaly)
 * <ul>
 * <li>max sum of distances to Center over all cells on a path; (although that will take a lot of SQRTs) OR
 * <li>furthest closest segment to Center. 
 * </ul>
 *
 * @author Gabriel Balan
 */
public class OneTurnTopsItineraryFactory extends ItineraryFactory 
{
	public static final String P_1TURNTOPS_ITINERARY_FACTORY = "1turntops-itinerary-factory";
	public static final String P_MODE = "mode";
	public static final String V_SF = "smaller-leg-first";
	public static final String V_LF = "bigger-leg-first";
	public static final String V_RF = "random";
	public static final String V_AC = "avoid-center";
	
	public static final int SF = 0;
	public static final int LF = 1;
	public static final int RF = 2;
	public static final int AC = 3;
	int mode;
	Double2D center=null; 
	
	public Parameter defaultBase(){return new Parameter(P_1TURNTOPS_ITINERARY_FACTORY);}
	public void setup(ParamSimState state, Parameter base)
	{
		super.setup(state, base);
		String smode = state.parameters.getStringWithDefault(base.push(P_MODE), defaultBase().push(P_MODE), V_AC);
		if(smode.equals(V_SF)) mode = SF;
		else
			if(smode.equals(V_LF)) mode = LF;
		else
			if(smode.equals(V_RF)) mode = RF;
		else
			if(smode.equals(V_AC)) mode = AC;
		else
			throw new RuntimeException("Invalid mode "+smode);
		if(mode == AC)
		{
			GridCity city = (GridCity)this.state.roadNet;
			center = new Double2D(0.5*(city.width-1),0.5*(city.height-1));
//			System.out.println("Grid Center "+center);
			//y means rows and x means columns. Just so things do not get mixed up 
		}
	}

	public Itinerary getItinerary(
		Intersection source,
		Intersection destination,
		Car car) 
	{
		GridNode currentNode;
		Int2D startingPoint = (currentNode = (GridNode)source).getCoords();
		Int2D destinationPoint = ((GridNode)destination).getCoords();
		int xBlocks= destinationPoint.x - startingPoint.x;
		int yBlocks= destinationPoint.y - startingPoint.y;
		int xDirection, yDirection;
		//assume (0,0) to be the upper left corner.

		//W is towards the smaller column indices
		if(xBlocks>=0)
			xDirection = GridNode.E;
		else
		{
			xDirection = GridNode.W;
			xBlocks= -xBlocks; 
		}
		
		//N is towards smaller row indices
		if(yBlocks>=0)
			yDirection = GridNode.S;
		else
		{
			yDirection = GridNode.N;
			yBlocks = -yBlocks; 
		}	
		int nblocks = xBlocks + yBlocks;
		int[] data = new int[nblocks];
		
		//I will go x, y and x again.
		//in each of the modes one of these three will be 0.
		int firstXLegBlocks=0;
		switch(mode)
		{
			case SF: 	if(xBlocks<yBlocks || (xBlocks==yBlocks && state.random.nextBoolean()))//ties broken randomly
						{
							firstXLegBlocks = xBlocks;
							xBlocks = 0;
						}
						else firstXLegBlocks = 0;
						break;
			case LF: 	if(xBlocks>yBlocks || (xBlocks==yBlocks && state.random.nextBoolean()))
						{
							firstXLegBlocks = xBlocks;
							xBlocks = 0;
						}
						else firstXLegBlocks = 0;
						break;
			case RF: 	if(state.random.nextBoolean())
						{
							firstXLegBlocks = xBlocks;
							xBlocks = 0;
						}
						else firstXLegBlocks = 0;
						break;
			case AC:	double tmp;
						double xfSq = (tmp = (center.x - destinationPoint.x))*tmp + 
									  (tmp = (center.y - startingPoint.y))*tmp;
						double yfSq = (tmp = (center.y - destinationPoint.y))*tmp + 
									  (tmp = (center.x - startingPoint.x))*tmp;
						if(xfSq>yfSq || (xfSq==yfSq && state.random.nextBoolean()))
						{
							firstXLegBlocks = xBlocks;
							xBlocks = 0;
						}
						else firstXLegBlocks = 0;
		}
		
		int index = 0;
		for(int i=0;i<firstXLegBlocks;i++)
		{
			int outIndex = data[index++] = currentNode.getOutEdgeIndices()[xDirection];
			Edge edge = (Edge)currentNode.getIOI().out.objs[outIndex];
			currentNode = (GridNode)edge.getOtherNode(currentNode);
		}
		for(int i=0;i<yBlocks;i++)
		{
			int outIndex = data[index++] = currentNode.getOutEdgeIndices()[yDirection];
			Edge edge = (Edge)currentNode.getIOI().out.objs[outIndex];
			currentNode = (GridNode)edge.getOtherNode(currentNode);
		}
		for(int i=0;i<xBlocks;i++)
		{
			int outIndex = data[index++] = currentNode.getOutEdgeIndices()[xDirection];
			Edge edge = (Edge)currentNode.getIOI().out.objs[outIndex];
			currentNode = (GridNode)edge.getOtherNode(currentNode);
		}
		return new FixedItinerary(data);
	}
}
