package sim.app.traffic.grid;
import sim.app.traffic.*;
import sim.field.network.*;
import sim.util.*;
import ec.util.*;


/**
 * This is a Two-Way-Street Intersection.
 * The phases are prespecified.
 * 
 * <p>Since this extends <code>OneLaneOutEdgeIntersection</code>, 
 * U-turns are only permited with left turns.
 * 
 * Groups:<ul>
 * <li>NS, SN, SE, NW//straight ahead and right turns</li>
 * <li>EW, WE, WS, EN//straight ahead and right turns</li>
 * <li>NE, SW [, NN, SS]//left & U turns</li>
 * <li>ES, WN [, EE, WW]//left & U turns</li>
 * </ul>
 * 
 * <p>Note that the right turns are not allowed with the left turns.
 * This satisfies the principe that two queues outputting in the same
 * edge are conflicting.
 * 
 * @author Gabriel Balan
 */
public class RegularGridIntersection extends OneLaneOutEdgeIntersection
										implements GridNode
{	
	boolean uturns=true;
	public static final String P_UTURNS = "U-turns";
	public static final String P_REGGRID_INTERSECTION = "regular-grid-intersection";
	public Parameter defaultBase(){return new Parameter(P_REGGRID_INTERSECTION);}
	public void setup(ParamSimState state, Parameter base) 
	{
		super.setup(state, base);
		uturns = state.parameters.getBoolean(base.push(P_UTURNS),
												defaultBase().push(P_UTURNS),
												uturns);
	}

	
	Int2D coords;
	public Int2D getCoords(){return coords;}
	public void setPosition(int x, int y){this.coords = new Int2D(x,y);}
	int[] outEdgeIndices;
	int[] inEdgeIndices;
	public int[] getOutEdgeIndices(){return outEdgeIndices;}
	public int[] getInEdgeIndices(){return inEdgeIndices;}
	public Network.IndexOutIn getIOI(){return ioi;}

	public void netSetup(	TrafficSim ts,
							Network.IndexOutIn ioi, 
							int nin, 
							int nout,
							int sin, 
							int sout,
							int ein, 
							int eout,
							int win, 
							int wout)
	{
		netSetup(ioi);
		outEdgeIndices = new int[4];
		outEdgeIndices[N] = nout;
		outEdgeIndices[S] = sout;
		outEdgeIndices[W] = wout;
		outEdgeIndices[E] = eout;
		
		inEdgeIndices = new int[4];
		inEdgeIndices[N] = nin;
		inEdgeIndices[S] = sin;
		inEdgeIndices[W] = win;
		inEdgeIndices[E] = ein;
		
		int ins = ioi.in.numObjs;
		int outs = ioi.out.numObjs;
		nQueues = ins*outs;

		int nGroups = 4;
		allocateData(nQueues, nGroups);

		for(int i=0;i<nQueues;i++)
				queues[i]= new TrafficQueue(ts, this, i, carsPassingPerFreshGreenPhasePerQueue, carsPassingPerOldGreenPhasePerQueue, takeOffTime);
		
		TrafficQueue[] o;
		
		Bag bag = new Bag(4);
		addQueue(nin, sout, bag);
		addQueue(sin, nout, bag);
		addQueue(sin, eout, bag);
		addQueue(nin, wout, bag);



		groups[0]= o = new TrafficQueue[bag.numObjs];
		System.arraycopy(bag.objs,0,o,0,bag.numObjs);
		bag.clear();							
		
		addQueue(ein, wout, bag);
		addQueue(win, eout, bag);
		addQueue(win, sout, bag);
		addQueue(ein, nout, bag);
		groups[1]= o = new TrafficQueue[bag.numObjs];
		System.arraycopy(bag.objs,0,o,0,bag.numObjs);
		bag.clear();
			
		addQueue(sin, wout, bag);
		addQueue(nin, eout, bag);
		if(uturns)
		{
			addQueue(sin, sout, bag);
			addQueue(nin, nout, bag);
		}
		groups[2]= o = new TrafficQueue[bag.numObjs];
		System.arraycopy(bag.objs,0,o,0,bag.numObjs);
		bag.clear();							

		addQueue(ein, sout, bag);
		addQueue(win, nout, bag);
		if(uturns)
		{
			addQueue(win, wout, bag);
			addQueue(ein, eout, bag);
		}
		groups[3]= o = new TrafficQueue[bag.numObjs];
		System.arraycopy(bag.objs,0,o,0,bag.numObjs);
		bag.clear();							
	}
	
//	//this is called only when the net is set up, so speed is not crucial
//	private TrafficQueue getQueue(int incomingEdgeIndex, int outgoingEdgeIndex)
//	{
//		int outs = ioi.out.numObjs;
//		return queues[incomingEdgeIndex*outs+outgoingEdgeIndex];
//	}
	
	private void addQueue(int incomingEdgeIndex, int outgoingEdgeIndex, Bag b)
	{
		if(incomingEdgeIndex<0 ||outgoingEdgeIndex<0) return;//on of these 
		int outs = ioi.out.numObjs;
		b.add(queues[incomingEdgeIndex*outs+outgoingEdgeIndex]);
	}	
}
