package sim.app.traffic.grid;

import sim.app.traffic.*;
import sim.app.traffic.car.Car;
import sim.app.traffic.route.*;
import sim.util.Int2D;
import sim.field.network.*;
import ec.util.*;
/**
 * This factory produces any shortest-path itinerary.  Since this is 
 * designed for grids, the itinerary could be any shortest path inside the bounding box 
 * containing the source and destination.
 * As name suggests it, all these shortest paths consist of the smallest number of 
 * blocks.
 * 
 * <p>Expected intersection waiting/crossing times are of no concern here.
 * 
 * <p>All such itineraries are equally likely to be produced.
 * <p>This only works with <code>GridCity</code>, 
 * not with <code>AlternatingGridCity</code>,
 * and not (yet) with <code>TorusCity</code>.
 * 
 * <p>If the bounding box is x-by-y blocks, there are x+y_choose_x
 *  = (x+y)!/x!y! possible itineraries (since you need to travel x+y blocks, 
 * and you get to choose x of them to be horizontal).
 * 
 * if you travel a block on the x axis, then you're left 
 * with (x+y-1)!/(x-1)!/y! itineraries, so the decision to
 * 'travel a block on the x axis' should have been made with 
 * probability (x+y-1)!/(x-1)!/y! / [(x+y)!/x!/y!] =  x/(x+y).
 * 
 * Therefore you can build a uniformly-distributed
 * itinerary by making independent decisions at each step: 
 * if the bounding box is x-by-y blocks, then travel 1 block on x with 
 * probability x/x+y and on y with probability y/(x+y).
 * Note that this works even when x=0 or y=0.
 * 
 * <p>TODO test this!
 * <p>TODO make sure the road net is a regular GridCity, but not a subclass, such as alternating or torus.
 * <p>TODO fix this for torus: take the smallest of |dx| and |torus.width-dx|. 
 * <p>TODO fix this for alternating grids: count how many times you can travel horizontally/vertically (at most dx/2);
 * You'd also need to watch the coin flip in the odd steps when there is no decision.
 * 
 * @author Gabriel Balan
 *
 */
public class UniformRandomMinBlockItineraryFactory extends SimpleRandomMinBlockItineraryFactory 
{
	public static final String P_UNIF_MINBLOCK_ITINERARY_FACTORY = "unif-minBlock-itinerary-factory";
	public Parameter defaultBase(){return new Parameter(P_UNIF_MINBLOCK_ITINERARY_FACTORY);}

	public Itinerary getItinerary(
		Intersection source,
		Intersection destination,
		Car car) 
	{
		GridNode currentNode;
		Int2D startingPoint = (currentNode = (GridNode)source).getCoords();
		Int2D destinationPoint = ((GridNode)destination).getCoords();
		xBlocks.value =destinationPoint.x - startingPoint.x;
		yBlocks.value =destinationPoint.y - startingPoint.y;
		int xDirection, yDirection;


		//W is towards the smaller column indices
		if(xBlocks.value>=0)
			xDirection = GridNode.E;
		else
		{
			xDirection = GridNode.W;
			xBlocks.value = -xBlocks.value; 
		}
		
		//N is towards smaller row indices
		if(yBlocks.value>=0)
			yDirection = GridNode.S;
		else
		{
			yDirection = GridNode.N;
			yBlocks.value = -yBlocks.value; 
		}	
		int nblocks = xBlocks.value+yBlocks.value;
		int[] data = new int[nblocks];
		for(int i=0;i<nblocks;i++)
		{
			assert(i+xBlocks.value+yBlocks.value == nblocks);
			int direction;
			double probx = ((double)xBlocks.value)/(xBlocks.value+yBlocks.value);
			
			if(state.random.nextBoolean(probx))
			{
				direction = xDirection;
				xBlocks.value--;
			}
			else
			{
				direction = yDirection;
				yBlocks.value--;
			}
			int outIndex = data[i] = currentNode.getOutEdgeIndices()[direction];
			Edge edge = (Edge)currentNode.getIOI().out.objs[outIndex];
			currentNode = (GridNode)edge.getOtherNode(currentNode);
		}
		assert(currentNode == destination);
		return new FixedItinerary(data);
	}
}
