package sim.app.traffic.load;

import sim.app.traffic.car.*;
import sim.app.traffic.*;
import sim.util.*;
import sim.util.stats.doubletime.TimeWeightedAverager;
import ec.util.*;
/**
 * This deploys cars according to a workload function: time->#cars; it is responsible for scheduling
 * itself to wake up more cars (if needed). If the function says "less cars", it waits for some to 
 * finish and then it does not wake up more. 
 * 
 * <p>
 * After a car reaches its destination, it reports back to the manager, and it gets "frozen".
 * The manager is responsible for keeping track of "sleeping" cars [and their locations, if 
 * waken-up cars should start from where they left off].
 * 
 * <p>
 * Cars start with the same wealth. All cars start the same, but then they do not have their wealth reset every day/week/year.
 * 	IT DEPENDS ON THE DESIRED TIME GRANULARITY: I WANT TO REDUCE THE VARIATION IN WAITING TIME OVER
 * ALL CARS, IN THE LAST X TIME UNITS. Depending on what is it that I want, even reseting after each
 * trip can make sense.
 * 
 * <p>
 * TODO A subclass can overwrite <code>destinationReached()</code>, so a car is rescheduled after a waiting time
 * drawn from some distribution. If the cars take trips after trips in an infinite loop (the same way 
 * as virtual browsers are used to test websites) without resetting their wealths, I might need a 
 * financially-balanced set of policies: if the "economy" leaks credits, everybody will have 0 out of the 
 * intersections, so it will be actually only 'counting-cars'!!
 * 
 * @author Gabriel Balan
 */
public abstract class TrafficLoadManager implements Parametrizable 
{
	
	public CarGenerator carGenerator;
	public ODSelector odSelector;
	public static final String P_CAR_GEN = "car-generator";
	public static final String P_OD_SELECTOR = "OD-selector";
	
	public static final String P_RESET_MODULO = "resetWealth";	

	final protected CarBagWithO1Remove aliveCars = new CarBagWithO1Remove(CarBagWithO1Remove.ALIVE_BAG)
	{
		public void add(Car c)
		{
			super.add(c);
			aliveCarCountAverager.addSample(aliveCars.bag.numObjs);
		}
		public void remove(Car c)
		{
			super.remove(c);
			aliveCarCountAverager.addSample(aliveCars.bag.numObjs);
		}
	};
	private TimeWeightedAverager aliveCarCountAverager;
	public double getAverageNumberOfLifeCars(){return aliveCarCountAverager.getAverage();}
	
	
	//, frozenCars; IF YOU USE THIS, ADD IT TO finish()
//	public WealthDistributionTrace aliveCarsWealthDistributionTrace = null;
	
	public void finish()
	{
		if(resetWealthModuloTrips==0)
		{
			int na = aliveCars.size();
			for(int i=na-1;i>0;i--)
				canCar(aliveCars.get(i));
				//initial wealth is irrelevant here, 
				//I just want the delay stats of all cars added to the global one.
				//since they were not added along the way.(no resets)
		}
		
		carGenerator.finish();
	}
	
	/** the number of trips after a car gets its wealth reset
	 * 0 if the wealths are never reset 
	 * 
	 * I put this here and not in CAR because it controls a global behavior
	 * across all kins of cars. 
	 * TODO it should go where the other traffic light policies are loaded!
	 * TODO I could also reset based on time (e.g. reset every day is a more resonably implementable idea)
	 * same holds for initialCarWealth
	 *  */
	protected int resetWealthModuloTrips = 1;
	public TrafficSim tSim;
		
	public void setup(ParamSimState state, Parameter base)
	{
		tSim = (TrafficSim)state;
		aliveCarCountAverager = new TimeWeightedAverager(tSim.schedule);
		Parameter p;
		carGenerator = (CarGenerator)state.parameters.getInstanceForParameter(p=base.push(P_CAR_GEN), null, CarGenerator.class);
		carGenerator.setup(state, p);
		
		odSelector = (ODSelector)state.parameters.getInstanceForParameter(p=base.push(P_OD_SELECTOR), null, ODSelector.class);
		odSelector.setup(state, p);
		
		resetWealthModuloTrips = state.parameters.getIntWithDefault(base.push(P_RESET_MODULO), null, resetWealthModuloTrips);
		if(resetWealthModuloTrips<0)
			throw new RuntimeException("reset wealth every "+resetWealthModuloTrips+" trips makes no sense!");

	}
	public Object simpleClone(){return null;}
	public void start()
	{
		aliveCarCountAverager.start();
	}

	static abstract public class WealthDistributionTrace implements TrafficSim.InfoDisplayUpdater
	{
		public abstract void addSlice(double time, sim.app.traffic.car.CarBagWithO1Remove cars);
		public void add(TrafficSim sim)
		{
			addSlice(sim.schedule.time(), sim.loadManager.aliveCars);
		}
	}
	static public class PrintWealths extends WealthDistributionTrace
	{
		public void reset(){}
		public void addSlice(double time, sim.app.traffic.car.CarBagWithO1Remove cars)
		{
			int n = cars.size();
			for(int i=0;i<n;i++) System.out.print(cars.get(i).wealth+"\t");
			System.out.println();

		}		
	}
	
	public abstract void destinationReached(Car car);
	
	
//	{
//		//TODO if you do stats on car.wealth, do it now, cause it might get reset.
//		if(resetWealthModuloTrips>0)
//		{
//			if(car.destinationsReached % resetWealthModuloTrips ==0)
//				car.resetWealthAndWaitingTimes(tSim, initialCarWealth);
//		}
//		allCars.add(car);
//		//I want to do SELECTION WITHOUT REPLACEMENT!, 
//		//but this would only work if the bag is not already empty (such as when the cars do tight loops)
//
//		//wake up allcars.objs[state.random.getInt(min(allCars.numObj-1, 1))];
//	}


	void check()
	{
		Bag intersections = tSim.roadNet.allNodes;
		for(int i=0;i<intersections.numObjs;i++)	
		{
			Intersection inter = (Intersection)intersections.objs[i];
			for(int j=0;j<inter.queues.length;j++)
			{
				TrafficQueue tq = inter.queues[j];
				if(tq.size()==0 && tq.currentWealth!=0)
					throw new RuntimeException("bummer.");				
			}
		}
	}

	protected void spawnACar()
	{
		Car c = carGenerator.getCar(tSim);
		aliveCars.add(c);
		c.destinationsReached = 0;
		c.wealth = tSim.initialWealth;

		RoadNet net = tSim.roadNet;

		int sourceIndex = odSelector.getRandomSourceIndex(tSim.random);
		int destinationIndex = odSelector.getRandomDestinationIndex(tSim.random, sourceIndex);	
//		System.out.println("Task: "+sourceIndex+" "+destinationIndex);

		Intersection source = (Intersection)net.allNodes.objs[sourceIndex];
		Intersection destination = (Intersection)net.allNodes.objs[destinationIndex];
		c.deploy(tSim, source, tSim.itineraryFactory.getItinerary(source, destination, c));		
	}

	protected void canCar(Car c)
	{
		if(resetWealthModuloTrips==0)
			c.resetWealthAndWaitingTimes(tSim, 0);
		aliveCars.remove(c);
	}
	
	
}
