package sim.app.traffic.queuewealth.delayincrease;
import ec.util.Parameter;
import sim.app.traffic.queuewealth.*;
import sim.app.traffic.*;
import sim.app.traffic.car.*;
import sim.util.*;

/**
 * @author Gabriel Balan
 * 
 * 
 * A queue's score is the sum of delay-increases due to
 * getting RED instead of GREEN. In other words, is the 
 * loss in efficiency produced by this queue if it gets
 * a RED instead of a GREEN
 * 
 * The value is based on a myopic analysis: it does not 
 * look for situations such as "if C1 does not stop here 
 * it will stop at the next intersection for 20 secs; but
 * if it waits here 5 more seconds, it's new
 * power/credit score/etc will allow it to go through 
 * unstopped at the next one.
 * 
 */
public class DelayIncrease_v1 extends QueueWealthMetric 
{
	//null accepted, we'll do Identity and just skip the function call.
	ExtraDelayContributionFunction fn=null;
	public static final String P_CONTRIBUTION_FN = "contribution-fn";
	public void setup(ParamSimState state, Parameter base)
	{
		Parameter p =base.push(P_CONTRIBUTION_FN);
		if(state.parameters.exists(p))
		{
			fn = (ExtraDelayContributionFunction)state.parameters.getInstanceForParameter(
				p, null, ExtraDelayContributionFunction.class);
			fn.setup(state, p);
		}
	}
	
	/**
	 * My intersection is considering giving me the RED/GREEN light. How much
	 * bigger is my delay at this intersection going to be if i get RED
	 * instead of GREEN?
	 */
	public float getQueueWealth(TrafficQueue tq) 
	{
		float totalExtraDelay = 0;
		sim.util.Bag inRangeCarBag = tq.inRangeCars.bag;
		if(inRangeCarBag.numObjs==0)
			return totalExtraDelay;
		
		double nextPhaseDuration = tq.owner.phaseDuration;

		// /upper limit on the number of cars I'll be able to
		// get through next phase (assuming I win)
		int crossingOpportunitiesNextPhase = (tq.color == TrafficQueue.GREEN) ? 
				tq.getMaxCarsPassingPerOldGreenPhase():
				tq.getMaxCarsPassingPerFreshGreenPhase();

		// for each car C already stopped
		// if C could made it through [if it gets green]
		// -add 5secs of yellow and 3secs*position of C.
		// else add 20secs

		// <code>takeoffDelay</code> includes the yellow duration and the
		// <code>takeoffTime</code>s
		double takeoffDelay = (tq.color == TrafficQueue.GREEN) ? tq.state.roadNet.intersectionCrossingDuration
				: 0;
		int qSize= tq.size();

		java.util.Iterator iter = tq.iterator();
		while(iter.hasNext())
		{
			Car c = (Car)iter.next();

			double delay_i = 0;
			if (crossingOpportunitiesNextPhase > 0)
			{
				delay_i = nextPhaseDuration;
			} else
			{
				crossingOpportunitiesNextPhase--;
				takeoffDelay += tq.takeOffTime;
				delay_i = takeoffDelay;
			}
			totalExtraDelay += (fn==null?delay_i:fn.delayContributuion(c, delay_i));
		}

		//
		// for each car C moving in-range
		// there's a delay of
		// (20secs-the time it takes for C to reach the intersection) if C would
		// make it through
		// //the first _crossing-opportunities_ cars
		// 0 secs for the rest (they won't make it through either way).
		double now = tq.state.schedule.time();
		int numberInRangeCars = inRangeCarBag.numObjs;
		int numberInRangeMovingCars = inRangeCarBag.numObjs-qSize;
		if (numberInRangeMovingCars != 0 && crossingOpportunitiesNextPhase != 0)
		{
			if(numberInRangeMovingCars <=crossingOpportunitiesNextPhase)
				//all cars in range could pass through.
				for (int i = 0; i < numberInRangeCars; i++)
				{
					Car c = (Car) inRangeCarBag.objs[i];
					if(c.isStoppedInQueue()) continue;
					double delay_i = Math.max(nextPhaseDuration-c.getTimeTillIntersection(now), 0);
					totalExtraDelay += (fn==null?delay_i:fn.delayContributuion(c, delay_i));
				}
			else
			{
				// only the first "crossingOpportunitiesNextPhase" cars
				// would be affected by the next RED/GREEN decision

				Double[] keys = tmpHeapKeys;
				Car[] values = tmpHeapValues;
				if(keys.length<numberInRangeCars)
				{//resize
					keys = new Double[numberInRangeCars];
					values = new Car[numberInRangeCars];
				}

				int index = 0;
				for (int i = 0; i < numberInRangeCars; i++)
				{
					Car c = (Car) inRangeCarBag.objs[i];
					if(c.isStoppedInQueue()) continue;
					values[index] = c;
					keys[index] = new Double(c.getTimeTillIntersection(now));
					index++;
				}
				Heap h = new Heap(keys, values, numberInRangeMovingCars);
				for (int x = 0; x < crossingOpportunitiesNextPhase; x++)
				{
					double timeTillIntersection = ((Double)h.getMinKey()).doubleValue();
					Car c = (Car)h.extractMin();
					double delay_i = Math.max(nextPhaseDuration
							- timeTillIntersection, 0);
					totalExtraDelay += (fn==null?delay_i:fn.delayContributuion(c, delay_i));
				}
			}
		}
		if(totalExtraDelay<0)
			throw new RuntimeException("Error: DelayIncrease<0");
		return totalExtraDelay;
	}
	static Double[] tmpHeapKeys = new Double[1];
	static Car[] tmpHeapValues = new Car[1];
}
