package sim.app.traffic.route;

import sim.app.traffic.*;
import sim.app.traffic.car.Car;
import sim.field.network.*;
import sim.util.*;
import ec.util.*;
import sim.field.network.stats.*;

/**
 * 
 * Choosing randomly a shortest path. All my experiments had all segments of equal length, so 
 * shortest path was essentially based on number of crossed intersections. 
 * 
 * Note that this is NOT necessarily a fastest path though 
 * (different intersections can be more crowded than others, so it would take longer to cross).
 * Moreover, future implementations should model the fact that 
 * "the more crowded the road, the slower you travel on it;"
 * moreover, different drivers could have different estimates of how crowded a street will be.
 * 
 * 
 * <p>This factory pre-computes and caches all-pairs shortest paths, so it assumes the network 
 * will not changed during the simulation (not the topology, nor the road lengths -i.e. edge weights).
 * 
 * <p>This factory builds an itinerary from source to destination, one intersection at a time.
 * At each intermediate point it randomly chooses between the 
 * next-step choices that lead to shortest paths. Each local decision is made using a uniform distribution,
 * but the resulting global itinerary distribution is not uniform.
 * 
 * 
 * <p>Consider the entire set of shortest-path itineraries represented as a tree:
 * <ul>
 * <li>each path is an itineary
 * <li>each node is labelled to an intersection
 * <li>the root is labelled with the the source intersection,
 * <li>every leaf is labelled with the destination intersection, 
 * <li>each non-leaf node is a posible decision point;
 * </ul>
 * This factory uses a uniform distribution to choose the child of the current node in the tree;
 * however, in order to choose an itinerary from a uniform distribution, one should choose
 * the between a child of a node with a probability proportional to the number of leaves in the 
 * subtree rooted in that child.
 * Each choice is  [locally] uniform, but the overall result is not [globally] uniform.
 * 
 * <p>A factory producing globally uniform distribution over shortest itineraries 
 * would need a Dynamic Programming table to compute the number of ways
 * you can get from an intersection to another; these values could grow exponentially in the size of the road network, too.
 * For grids there's no need for a Dynamic Programming table (see sim.app.traffic.grid.UniformRandomMinBlockItineraryFactory;
 * note that it only works for 2-way-street grids, no 1-way street alternating grids, no toroidal grids either).
 * 
 * <p>Although the uniform itinery distribution is intuitivelly more appealing, the factory in this class is simpler. 
 * More importantly, I already published a paper
 * using it, so I keep this class for backward compatilibty/result reproducibility.
 * 
 * @author Gabriel Balan
 */
public class SimpleRandomShortestPathItineraryFactory extends ItineraryFactory 
{
//	public static final String P_ALT_RND_MINBLOCK_ITINERARY_FACTORY = "alt-rnd-minBlock-itinerary-factory";
//	public Parameter defaultBase(){return new Parameter(P_ALT_RND_MINBLOCK_ITINERARY_FACTORY);}
	public Parameter defaultBase(){return null;}

	double[][] distances;
	TrafficSim state;
	
	WeightedEdgeMetric edgeMetric = WeightedEdgeMetric.defaultInstance;
	//TODO have this specified in the .params specs (set it in <code>setup</code>)
	
	public void setup(ParamSimState state, Parameter base) 
	{
		this.state = (TrafficSim)state;
		distances = NetworkStatistics.johnsonShortestPathsMatrix(this.state.roadNet, edgeMetric);
		//TODO transpose the matrix to speed things up 
		//(with destination as the first index, getItineray will work with 1 double[], not double[][]).
	}

	static IntBag ties = new IntBag();
	
	public Itinerary getItinerary(
		Intersection source,
		Intersection destination,
		Car car) 
	{
		int destInt = destination.ioi.index;
		int sourceInt = source.ioi.index;
//		System.out.println("Task: "+sourceInt+"\t"+destInt);
		IntBag iti = new IntBag(distances.length);
		
		while(destInt != sourceInt)
		{
			ties = getShortestPathNextStepCandidates(source, destInt, ties);
			
			int winningIndex = ties.objs[(ties.numObjs==1)? 0: state.random.nextInt(ties.numObjs)];
			iti.add(winningIndex);
			source = (Intersection)((Edge)source.ioi.out.objs[winningIndex]).getOtherNode(source);
			sourceInt = source.ioi.index;
			
		}
//		System.out.print("Itinerary:");
//		for(int i=0;i<iti.numObjs;i++)System.out.print(iti.objs[i]+" ");
//		System.out.println();		
		return new FixedItinerary(iti);
	}
	
	/**
	 * This does not return the indices of intersections (in the 0..#intersections-1),
	 * but the indices in current intersection's ioi.out bag.
	 */
	protected IntBag getShortestPathNextStepCandidates(Intersection intersection, int destinationIndex, IntBag candidateBag)
	{
		if(candidateBag==null) candidateBag = new IntBag();
		candidateBag.clear();
		Bag outEdges = intersection.ioi.out;
		int intersectionIndex = intersection.ioi.index;
		double intersectionToDestinationDistance = distances[intersectionIndex][destinationIndex];
		for(int i=0;i<outEdges.numObjs;i++)
		{
			Edge edge = (Edge)outEdges.objs[i];
			Intersection inter = (Intersection)(edge).getOtherNode(intersection);
			int neighborIndex = inter.ioi.index;
			double edgeWeight = edgeMetric.getWeight(edge);
			double neighborToDestinationDistance = distances[neighborIndex][destinationIndex];
			
			if(neighborToDestinationDistance + edgeWeight == intersectionToDestinationDistance)
				candidateBag.add(i);
			//people say "don't compare floating point numbers for equality" (use <, > and some epsilon)
			//e.g. Math.abs(intersectionToDestinationDistance - neighborToDestinationDistance - edgeWeight) <= 10E-9)
			//
			//]I only use the distance and edge-value values, so numerical rounding errors 
			// do not accumulate, so this shouldn't be much of an issue.
			//
			//It turns out that my roads' lengths are all powers of 2, so I didn't need an epsilon
			//using integer values is also safe.
			//
			//TODO I could use a comparator
		}
		assert(!candidateBag.isEmpty());
		return candidateBag;
	}
}
