package sim.app.traffic.route;

import sim.app.traffic.*;
import sim.app.traffic.car.*;
import sim.field.network.*;
import sim.field.network.stats.*;
import sim.util.*;
import sim.util.intdistribution.*;
import ec.util.*;

/**
 * Given a (Source, Destination) pair of intersections, this factory draws
 * an itinerary from a uniform distribution over all shortest 
 * paths from Source to Destination. 
 * 
 * @author Gabriel Balan
 *
 */
public class UniformRandomShortestPathItineraryFactory extends
		SimpleRandomShortestPathItineraryFactory
{
	public static final String P_CACHE_LOCAL_DISTRIBS = "cache-local-distributions";
	/**
	 * If this is null, then there's no caching.
	 * TODO use soft references instead (they let GC reclaim some of them when memory is tight).
	 */
	private IntDistribution[][] localDecisionDistributions = null;
	/** 
	 * Each cell contains the number of different shortest paths between 2 intersections in the network.
	 * Note that the number of paths is exponential in the size of the network.
	 * <ul>
	 * <li>For a 10x10 intersection grid (9x9 block grid) there are under 50k paths between each pair of 
	 * nodes, so <code>int</code> would have been enough, but I'm using <code>long</code> anyways
	 * (heck, we have 64bit cpus now)
	 * <li>there's always <code>BigInteger</code>.
	 * </ul>
	 */
	long[][] numberOfShortestPaths;
	/**
	 * @return a IntDistribution whose support consists of the indices of out-edges
	 * 	in the intersection's bag of out edges (AND NOT INTERSECTIONS' INDICES).
	 * 
	 * The distributions might be cached, so users must be careful not to break
	 * them: one should always call <code>getInt</code>[WithReplacement] on them 
	 * as opposed to <code>getIntWithoutReplacement</code>.
	 * 
	 */
	protected IntDistribution getDistribution(Intersection intersection, int destinationIndex)
	{
		assert(intersection.ioi.index != destinationIndex);
		if(localDecisionDistributions!= null)
		{
			IntDistribution cachedDistribution = localDecisionDistributions[intersection.ioi.index][destinationIndex];
			if(cachedDistribution!=null)
				return cachedDistribution;
		}
		IntBag indicesForNextIntersectionCandiates = getShortestPathNextStepCandidates(intersection, destinationIndex, ties);

		int nCandidates = indicesForNextIntersectionCandiates.numObjs;
		double[] probs = new double[nCandidates];
		for(int i = 0; i< nCandidates;i++)
		{
			Edge e = (Edge)intersection.ioi.out.objs[indicesForNextIntersectionCandiates.objs[i]];				                                   
			Intersection successor= (Intersection)(e.getOtherNode(intersection));
			int successorIntersectionIndex = successor.ioi.index;
			probs[i] = numberOfShortestPaths[successorIntersectionIndex][destinationIndex];
		}
		ArbitrarySupportArbitraryProbabilityDistribution distrib = 
			new ArbitrarySupportArbitraryProbabilityDistribution(
					new MersenneTwisterFast(state.random.nextLong()),
					probs,
					(localDecisionDistributions != null ? 
							indicesForNextIntersectionCandiates.toArray():
								//if I'm caching these distributions
								//then I need to clone this array 
								//(otherwise all distributions will share it)
							indicesForNextIntersectionCandiates.objs),
							//if I'm NOT caching these distributions
							//then at most a single distribution is alive at any time
							//so there's no problem
					nCandidates
					);
		//bummer, probs is ready for GC at this point.
		//Good thing GC is tuned for these sort of things.
		
		
		if(localDecisionDistributions!= null)
		{
			localDecisionDistributions[intersection.ioi.index][destinationIndex] = distrib;
		}
		return distrib;
	}
	
	public Itinerary getItinerary(
			Intersection source,
			Intersection destination,
			Car car) 
		{
			int destInt = destination.ioi.index;
			int sourceInt = source.ioi.index;
//			System.out.println("Task: "+sourceInt+"\t"+destInt);
			IntBag iti = new IntBag(distances.length);
			
			while(destInt != sourceInt)
			{
				IntDistribution distrib = getDistribution(source, destInt);
				int winningIndex = distrib.getInt();
				
				iti.add(winningIndex);
				source = (Intersection)((Edge)source.ioi.out.objs[winningIndex]).getOtherNode(source);
				sourceInt = source.ioi.index;
				
			}
//			System.out.print("Itinerary:");
//			for(int i=0;i<iti.numObjs;i++)System.out.print(iti.objs[i]+" ");
//			System.out.println();		
			return new FixedItinerary(iti);
		}
	
	/**
	 * <code>super.setup()</code> sets up <code>double[][] distances</code>, containing 
	 * all-pair shortest paths.
	 * Now I set up <code>long[][] numberOfShortestPaths</code>, containing the numer of 
	 * different ways you can get from every node to every node.
	 *
	 * O( V(V+E))
	 * 
	 * <p>I sure hope the road net is strongly connected.
	 */
	public void setup(ParamSimState _state, Parameter _base) 
	{
		super.setup(_state, _base);
		final int nIntersections = NetworkStatistics.getNumberNodes(this.state.roadNet);
		numberOfShortestPaths = new long[nIntersections][nIntersections];
		for(int destinationNodeIndex = 0; destinationNodeIndex < nIntersections; destinationNodeIndex++)
		{
//			System.out.println("\tDEST "+ destinationNodeIndex);

			long[]  numberOfShortestPathsToDestination = numberOfShortestPaths[destinationNodeIndex];
			//again, I should transpose <code>distances</code> to make the array accesses faster.
			
			
			//#ShortestPath( node ) = Sum_node' #ShortestPath( node ) 
			//where node' is a successor of node  (i.e. (node,node') \in V)
			//such that dist(node,node')+[shortest]distanceToDest(node') = [shortest]distanceToDest(node)
			//
			//You want to update #ShortestPath( node ) after you have the final values for #ShortestPath( node' ).
			//You want radiate out from destination... or use a stack.
			
			//A negative numberOfShortestPathsToDestination value means I have not computed #ShortestPath for that node yet.
			//(#ShortestPath is positive).
			//
			//-2 = ON_THE_STACK
			//-1 = NOT_YET_ON_THE_STACK
			
			for(int i = 0; i < nIntersections; i++)
				numberOfShortestPathsToDestination[i] = NOT_YET_ON_THE_STACK;
			numberOfShortestPathsToDestination[destinationNodeIndex] = 1;
			
			for(int i = 0; i < nIntersections; i++)
			{
				long nShortestPaths_i2d = numberOfShortestPathsToDestination[i];
				if(nShortestPaths_i2d == NOT_YET_ON_THE_STACK)
					computeNumberOfWays(i, destinationNodeIndex, numberOfShortestPathsToDestination);
				assert(numberOfShortestPathsToDestination[i] >0);
			}			
			//this takes O(V+E)
			//each node is pushed once and it's out-edges are considered at most twice.
//			System.out.println("======================");
		}
		
		if(_state.parameters.getBoolean(_base.push(P_CACHE_LOCAL_DISTRIBS), 
										null, 
										false))
		{
			localDecisionDistributions = new IntDistribution[nIntersections][nIntersections];
		}
	}
	static private IntBag _successorBag = new IntBag();//not thread safe.
	static private IntBag stack = new IntBag();
	/**
	 * Computes (if not computed already) the computeNumberOfWays (shortest paths) from source to destination.
	 * It also computes that for all the nodes along any shortest path --if not already.
	**/
	private void computeNumberOfWays(int sourceNodeIndex, int destinationNodeIndex, long[] numberOfShortestPathsToDestination)
	{
		assert(stack.isEmpty());
		stack.push(sourceNodeIndex);
		while(!stack.isEmpty())
		{
			int topNodeIndex = stack.top();
				
			Intersection top= (Intersection)state.roadNet.getAllNodes().objs[topNodeIndex];
			IntBag successorBag = getShortestPathNextStepCandidates(top, destinationNodeIndex, _successorBag);
			
			long sum = 0;
			boolean readyToEvaluate = true;
			
			for(int s=successorBag.numObjs-1;s>=0;s--)
			{
				Edge e = (Edge)top.ioi.out.objs[successorBag.objs[s]];				                                   
				Intersection successor= (Intersection)(e.getOtherNode(top));
				int successorIntersectionIndex = successor.ioi.index;
				long ns = numberOfShortestPathsToDestination[successorIntersectionIndex];
				if(ns>0)
				{
					sum+=ns;
					if(sum<=0)
						throw new RuntimeException("Overflow");
					continue;
				}
				readyToEvaluate = false;
				if(ns == NOT_YET_ON_THE_STACK)
					stack.push(successorIntersectionIndex);
				//else ns == ON_THE_STACK: nothing to do, 
				//successor is already on the stack.				
			}
			
			if(readyToEvaluate)
			{
				numberOfShortestPathsToDestination[topNodeIndex] = sum;
//				System.out.println("numberOfShortestPathsToDestination[\t"+topNodeIndex+"]="+sum);
				int t = stack.pop();
				assert(t==topNodeIndex);
			}
		}
		//the graph is finite and (I assume) strongly connected, so this won't cycle forever.
		//plus worst case is O(V+E)
		//a node is pushed in only once, and its out-edges are only iteraded over 
		//at most twice.
		//
		//It's important that a node's successor is not trapped under it on the stack
		//(since the successor won't be added to the stack, and the node would be stuck on the stack).
		//1.When this is called the stack has 1 node, 
		//2.The stack is empty at the end of the call
		//3.all node's successors (all the way to destination) would have been processed.
		//
	}
	
	
	private static final long NOT_YET_ON_THE_STACK = -1;
	private static final long ON_THE_STACK = -2;
}
