package sim.field.network.stats.actorcentrality;
import sim.field.network.*;

/**
 * 
 * Actor Centrality Index
 * 
 * <p>Normally this would be an interface containing 
 * <code>double getMeasure(final Network network, Object node)</code>.
 * BUT in some metrics (e.g. <code>InformationCentrality</code> and
 * <code>BetweennessCentrality</code>) all getMeasure calls use some common 
 * precomputed matrix.
 * 
 * <p>The downside of this is that one cannot have a single static instance of a certain centrality mesure, 
 * but as long as one does not have a large number of small Networks processed in a tight loop,
 * the overhead of allocating a new CentralityMeasure object for each graph is small. 
 * 
 * @author Gabriel Catalin Balan
 */
public abstract class NodeIndex 
{
	public final Network network;
	public NodeIndex(final Network network){this.network=network;}
	
	//I would call this getCentrality, but prestige measures extend this as well
	public abstract double getValue(final Object node);

	//For the measures computed in the constructor, overwriting this saves a hash-table look-up 
	public double getValue(final int nodeIndex)
	{
		return getValue(network.allNodes.objs[nodeIndex]);
	}
	/**
	 * The value of the metric might depend on the size of the graph.
	 * this function divides these values by the max value.
	 * 
	 * <br>This is final because this is what I do in <code>CentralizationStatistics</code>
	 * but without calling getValue(node) again.
	 */
	final public double getStandardizedValue(final Object node)
	{
		return getValue(node)/getMaxValue();
	}
	public abstract double getMaxValue();

	
	protected class DenseSquareMatrix 
    {
	    public int n;
	    public double[][] vals;
	        
	    public DenseSquareMatrix(int n)
        {
	        this.n = n;
	        this.vals = new double[n][n];
        }
	        
	    public DenseSquareMatrix(double[][] vals)
        {
	        this.vals = vals;
	        this.n = vals.length;
	        for(int i=0;i<n;i++)
		        if (vals[i].length != this.n)
		            throw new RuntimeException("This IS a square matrix");
        }

	    ///////////////////////////////////////////////////////////////////
	    // Got (and slightly adapted all the rest of the code code 
	    // from JAMA - http://math.nist.gov/javanumerics/jama/ 
	    ///////////////////////////////////////////////////////////////////    
	    public DenseSquareMatrix inverse () {
	
	        double[][] Idata = new double[n][n];
	
	        for (int i = 0; i < n; i++)
	           for (int j = 0; j < n; j++)
	           		Idata[i][j] = (i == j ? 1.0 : 0.0);
	        DenseSquareMatrix I = new DenseSquareMatrix(Idata);
	        return (new LUDecomposition(this)).solve(I);
	     }
	        
	    /**
	     * This reorders the rows of the matrix according
	     * to the array of indices <code>r</code>.<br>
	     * Intially this was 
	     * <code>getSubMatrix (int[] r, int j0, int j1)</code>
	     * and it would exctract the columns between <code>j0</code> 
	     * and <code>j1</code>,
	     * but I need all columns so here we are.
	     */ 
	    private DenseSquareMatrix getRowReorderedMatrix (int[] r) 
	        {
	        DenseSquareMatrix X = new DenseSquareMatrix(r.length);
	        double[][] B = X.vals;
	        try 
	            {
	            for (int i = 0; i < r.length; i++) 
	                {
	                for (int j = 0; j < n; j++) 
	                    B[i][j] = vals[r[i]][j];
	                }
	            } 
	        catch(ArrayIndexOutOfBoundsException e) 
	            {
	            throw new ArrayIndexOutOfBoundsException("Submatrix indices");
	            }
	        return X;
	        }
	
	        
	        
	    private class LUDecomposition implements java.io.Serializable 
	        {
	
	        /* ------------------------
	           Class variables
	           * ------------------------ */
	
	        /** Array for internal storage of decomposition.
	            @serial internal array storage.
	        */
	        private double[][] LU;
	        
	        /** Row and column dimensions, and pivot sign.
	            @serial column dimension.
	            @serial row dimension.
	            @serial pivot sign.
	        */
	        private int m, n, pivsign; 
	
	        /** Internal storage of pivot vector.
	            @serial pivot vector.
	        */
	        private int[] piv;
	
	        /* ------------------------
	           Constructor
	           * ------------------------ */
	
	        /** LU Decomposition
	            @param  A   Rectangular matrix
	            @return     Structure to access L, U and piv.
	        */
	        public LUDecomposition(DenseSquareMatrix A) 
	            {
	            // Use a "left-looking", dot-product, Crout/Doolittle algorithm.
	            LU = A.vals;
	            m = A.n;
	            n = A.n;
	            piv = new int[m];
	            for (int i = 0; i < m; i++)
	                piv[i] = i;
	                        
	            pivsign = 1;
	            double[] LUrowi;
	            double[] LUcolj = new double[m];
	
	            // Outer loop.
	            for (int j = 0; j < n; j++) 
	                {
	                // Make a copy of the j-th column to localize references.
	                for (int i = 0; i < m; i++) 
	                    LUcolj[i] = LU[i][j];
	                        
	                // Apply previous transformations.
	                for (int i = 0; i < m; i++)
	                    {
	                    LUrowi = LU[i];
	                    // Most of the time is spent in the following dot product.
	
	                    int kmax = Math.min(i,j);
	                    double s = 0.0;
	                    for (int k = 0; k < kmax; k++)
	                        s += LUrowi[k]*LUcolj[k];
	                    LUrowi[j] = LUcolj[i] -= s;
	                    }
	                   
	                // Find pivot and exchange if necessary.
	                int p = j;
	                for (int i = j+1; i < m; i++) 
	                    {
	                    if (Math.abs(LUcolj[i]) > Math.abs(LUcolj[p]))
	                        p = i;
	                    }
	                                
	                if (p != j) 
	                    {
	                    for (int k = 0; k < n; k++)
	                        {
	                        double t = LU[p][k]; LU[p][k] = LU[j][k]; LU[j][k] = t;
	                        }
	                    int k = piv[p]; piv[p] = piv[j]; piv[j] = k;
	                    pivsign = -pivsign;
	                    }
	                                
	                // Compute multipliers.
	                                     
	                if (j < m & LU[j][j] != 0.0) 
	                    {
	                    for (int i = j+1; i < m; i++)
	                        LU[i][j] /= LU[j][j];
	                    }
	                }
	            }
	
	        /* ------------------------
	           Public Methods
	           * ------------------------ */
	                
	        /** Is the matrix nonsingular?
	            @return     true if U, and hence A, is nonsingular.
	        */
	        public boolean isNonsingular() 
	            {
	            for (int j = 0; j < n; j++) 
	                {
	                if (LU[j][j] == 0)
	                    return false;
	                }
	            return true;
	            }
	                
	        /** Return lower triangular factor
	            @return     L
	        */
//	        public DenseSquareMatrix getL() 
//	            {
//	            DenseSquareMatrix X = new DenseSquareMatrix(m,n);
//	            double[][] L = X.vals;
//	            for (int i = 0; i < m; i++) 
//	                {
//	                for (int j = 0; j < n; j++) 
//	                    {
//	                    if (i > j)
//	                        L[i][j] = LU[i][j];
//	                    else if (i == j)
//	                        L[i][j] = 1.0;
//	                    else
//	                        L[i][j] = 0.0;
//	                    }
//	                }
//	            return X;
//	            }
	                
//	        /** Return upper triangular factor
//	            @return     U
//	        */
//	        public DenseSquareMatrix getU() 
//	            {
//	            DenseSquareMatrix X = new DenseSquareMatrix(n);
//	            double[][] U = X.vals;
//	            for (int i = 0; i < n; i++) 
//	                {
//	                for (int j = 0; j < n; j++) 
//	                    {
//	                    if (i <= j)
//	                        U[i][j] = LU[i][j];
//	                    else
//	                        U[i][j] = 0.0;
//	                    }
//	                }
//	            return X;
//	            }
	                
	//        /** Return pivot permutation vector
	//            @return     piv
	//        */
	//        public int[] getPivot()
	//            {
	//            int[] p = new int[m];
	//            for (int i = 0; i < m; i++) {
	//                p[i] = piv[i];
	//                }
	//            return p;
	//            }
	                
	        /** Return pivot permutation vector as a one-dimensional double array
	            @return     (double) piv
	        */
	        public double[] getDoublePivot() 
	            {
	            double[] vals = new double[m];
	            for (int i = 0; i < m; i++)
	                vals[i] = (double) piv[i];
	            return vals;
	            }
	                
	        /** Determinant
	            @return     det(A)
	            @exception  IllegalArgumentException  Matrix must be square
	        */
	        public double det() 
	            {
	            if (m != n)
	                throw new IllegalArgumentException("Matrix must be square.");
	            double d = (double) pivsign;
	            for (int j = 0; j < n; j++)
	                d *= LU[j][j];
	            return d;
	            }
	                
	        /** Solve A*X = B
	            @param  B   A Matrix with as many rows as A and any number of columns.
	            @return     X so that L*U*X = B(piv,:)
	            @exception  IllegalArgumentException Matrix row dimensions must agree.
	            @exception  RuntimeException  Matrix is singular.
	        */
	        public DenseSquareMatrix solve(DenseSquareMatrix B) 
	            {
//	            if (B.m != m)
//	                throw new IllegalArgumentException("Matrix row dimensions must agree.");
	            if (!this.isNonsingular())
	                throw new RuntimeException("Matrix is singular.");
	                        
	            // Copy right hand side with pivoting
	            int nx = B.n;
	            DenseSquareMatrix Xmat = B.getRowReorderedMatrix(piv);
	            double[][] X = Xmat.vals;
	                        
	            // Solve L*Y = B(piv,:)
	            for (int k = 0; k < n; k++) 
	                {
	                for (int i = k+1; i < n; i++) 
	                    {
	                    for (int j = 0; j < nx; j++)
	                        X[i][j] -= X[k][j]*LU[i][k];
	                    }
	                }
	            // Solve U*X = Y;
	            for (int k = n-1; k >= 0; k--) 
	                {
	                for (int j = 0; j < nx; j++)
	                    X[k][j] /= LU[k][k];
	                for (int i = 0; i < k; i++) 
	                    {
	                    for (int j = 0; j < nx; j++)
	                        X[i][j] -= X[k][j]*LU[i][k];
	                    }
	                }
	            return Xmat;
	            }
	        }
    }
}
