package sim.util;
import ec.util.*;
import java.io.*;

/**
 * @author Gabriel Balan
 *
 * The simplest most basic functionality of SSJ's Tally.
 * 
 * you add numbers, it maintains the count, sum and sum of squared values
 * in order to compute the sample mean and stdev.
 * 
 * What's new and special is that tallies can be combined.
 * 
 * I also maintain min and max, but they might be just dead wait slowing me down.
 */
public class Tally implements Parametrizable{
	public String name;
	public double sum;
	public int count;
	public double min;
	public double max;
	public double sumSq;
	
	public Tally(){reset();}
	public Tally(String name){reset();this.name=name;}
	public void reset()
	{ 
		sum=sumSq=0;
		count = 0;
		min = Double.POSITIVE_INFINITY;
		max = Double.NEGATIVE_INFINITY;
	}
	public void add(double value)
	{
		sum+=value;
		sumSq +=value*value;
		if(max<value) max=value;
		if(min>value) min=value;
		count++;
	}
	
	
	public double getAverage(){return sum/count;}
	public double getVariance(){return (sumSq - sum*sum/count)/(count-1);}
	public double getStDev(){return Math.sqrt(getVariance());}
	/**@return Coefficient of Variation*/
	public double getCoefVar(){return getStDev()/getAverage();}
	public double getMin(){return min;}
	public double getMax(){return max;}
	public int getCount(){return count;}
	public void setName(String name){this.name = name;}
	public String getName(){return name;}
	
	/**
	 * Fairness Index from (Jain, Chiu and Hawe, 84)
	 * (Sum_x)^2/[n Sum_(x^2)]
	 * 
	 * It does not woek well with negative values.
	 */
	public double getFairnessIndex()
	{
		if(min<0) throw new RuntimeException("Negative values!!!");
		return sum*sum/count/sumSq;
	}
	
	public String getFairnessIndexImage()
	{
		if(min<0) return "NA";
		return sum*sum/count/sumSq+"";
	}
	
	public double getConfidence( double confidenceLevel)
	{
		int n = getCount();
		if(n<30)	
			throw new RuntimeException("I've only implemented the '>=30' part");
			//I actually have a complete solution, but it envolves the ssj package I don't want to force you to use
		double alpha = 1-confidenceLevel;
		double sigma = Math.sqrt(getVariance());		
		return ec.multirun.ProbMath.inverseNormal(1-alpha/2)*sigma/Math.sqrt(n);
	}

	public void print()
	{
		System.out.println("\nStats for "+name);
		if(doCount)
			System.out.println("SampleSize...\t"+getCount());
		if(printAVG)
			System.out.println("Average......\t"+getAverage());
		if(printSTDEV)
			System.out.println("Stdev........\t"+getStDev());
		if(printMIN)
			System.out.println("Min..........\t"+getMin());
		if(printMAX)
			System.out.println("Max..........\t"+getMax());		
		if(printCV)
			System.out.println("Coef_Var.....\t"+getCoefVar());
		if(printFI)
			System.out.println("Fairness.....\t"+getFairnessIndexImage());
		if(printCI)
		{
			double avg = getAverage();
			double ci = getConfidence(confidenceLevel);
			System.out.println(confidenceLevel+"-Confidence interval for the mean:\t" + (avg-ci) + "\t" + (avg+ci));
		}
	}

	/** print on one line */
	public void println(final PrintWriter w)
	{
		String s = name;
		if(doCount) s += "\t" + getCount();
		if(printAVG) s+= "\t" + getAverage();
		if(printSTDEV) s+= "\t" + getStDev();
		if(printMIN) s+= "\t" + getMin();
		if(printMAX) s+= "\t" + getMax();
		if(printCV) s+= "\t" + getCoefVar();
		if(printCI)
		{
			double avg = getAverage();
			double ci = getConfidence(confidenceLevel);
			s += "\t" + (avg-ci) + "\t" + (avg+ci);
		}
		if(printFI) s+= "\t" + getFairnessIndex();
		w.println(s);
	}

	
	public boolean printAVG = true;
	public boolean printSTDEV = true;
	public boolean printMIN = true;
	public boolean printMAX = true;
	public boolean printCV = true;
	public boolean printCI = true;
	public boolean printFI = true;

	double confidenceLevel= 0.95;
	public boolean doCount = false;


	public static final String P_TALLY 		= "tally";
	public static final String P_NAME 		= "name";
	public static final String P_AVERAGE	= "average";
	public static final String P_STDEV 		= "stdev";
	public static final String P_MIN 		= "min";
	public static final String P_MAX 		= "max";
	public static final String P_CV 		= "cv";
	public static final String P_CI 		= "ci";
	public static final String P_FI 		= "fairness-index";

	public static final String P_CI_LEVEL 	= "ci-level";
	public static final String P_COUNT 		= "count";

	public void setup(ParamSimState state, Parameter base)
	{
		ParameterDatabase pdb = state.parameters;
		Parameter defp = defaultBase();
		name = pdb.getStringWithDefault(base.push(P_NAME), defp.push(P_NAME), name);
		doCount = pdb.getBoolean(base.push(P_COUNT), defp.push(P_COUNT), doCount);
		printAVG = pdb.getBoolean(base.push(P_AVERAGE), defp.push(P_AVERAGE), printAVG);
		printSTDEV = pdb.getBoolean(base.push(P_STDEV), defp.push(P_STDEV), printSTDEV);
		printMIN = pdb.getBoolean(base.push(P_MIN), defp.push(P_MIN), printMIN);
		printMAX = pdb.getBoolean(base.push(P_MAX), defp.push(P_MAX), printMAX);
		printCV = pdb.getBoolean(base.push(P_CV), defp.push(P_CV), printCV);
		printFI = pdb.getBoolean(base.push(P_FI), defp.push(P_FI), printFI);
		
		printCI = pdb.getBoolean(base.push(P_CI), defp.push(P_CI), printCI);
		if(printCI)
			confidenceLevel = pdb.getDouble(base.push(P_CI_LEVEL), defp.push(P_CI_LEVEL), 0, 1);
	}
	public Parameter defaultBase(){return new Parameter(P_TALLY);}
	/**	keeps the printing settings, reset the statistics.*/
	public Object simpleClone()
	{
		try
		{
			Tally newT = (Tally)clone();
			newT.reset();
			return newT;
		}catch(CloneNotSupportedException ex){return null;}
	}
}
