package sim.util.intdistribution;
import ec.util.*;

/**
 * This is an <code>IntDistribution</code> where a bunch of
 * integer values (not necesarily consecutive) are produced 
 * with arbitrary probabilities.
 * 
 * <p>The probabilities you specifiy don't have to be <=1; they just
 * have to be positive. (Imagine they get normalized).
 * 
 * @author Gabriel Balan
 *
 */
public class ArbitrarySupportArbitraryProbabilityDistribution extends IntDistribution
{
	//I'm not normalizing the probabilities!
	double[] cummulativeProbabilities;
	int[] supportValues;
	boolean uniform;
	
	int numElem;
	
	public ArbitrarySupportArbitraryProbabilityDistribution(){}
	public ArbitrarySupportArbitraryProbabilityDistribution(MersenneTwisterFast rng, 
															double[] probabilities, 
															int numElem)
	{
		this.rng = rng;
		this.numElem = numElem;
		cummulativeProbabilities = new double[numElem];
		uniform = true;
		double val_0 = cummulativeProbabilities[0];
		double sum = 0;
		for(int i=0;i<numElem;i++)
		{
			double val_i  = probabilities[i];
			if(val_i<0)
				throw new RuntimeException("Invalid 'probability' value: "+val_i);
			if(val_i!=val_0)
				uniform = false;
			cummulativeProbabilities[i] = sum += val_i;
		}
	}
	public ArbitrarySupportArbitraryProbabilityDistribution(MersenneTwisterFast rng, 
															double[] probabilities)
	{
		this(rng, probabilities, probabilities.length);
	}
	
	public ArbitrarySupportArbitraryProbabilityDistribution(MersenneTwisterFast rng, 
															double[] probabilities, 
															int[] support)
	{
		this(rng, probabilities, support, probabilities.length);
	}
	
	public ArbitrarySupportArbitraryProbabilityDistribution(MersenneTwisterFast rng, 
															double[] probabilities, 
															int[] support,
															int numElems)
	{
		this(rng, probabilities, numElems);
		if(numElems>support.length)
			throw new RuntimeException("Size mismatch");
		supportValues = support;
	}


	public static final String P_PROB = "probability";
	public static final String P_SUP = "support";
	public static final String P_COUNT = "count";
	public void setup(ParameterDatabase parameters, Parameter base, MersenneTwisterFast random)
	{	
		super.setup(parameters, base, random);
		Parameter p;
		int n = parameters.getInt(p=base.push(P_COUNT), null, 2);
		//no fun having n<2. (a distribution must have a support of at least 2 elements).
		if(n<2)
			throw new RuntimeException("Invalid value for "+p);
		int sum = 0;
		Parameter probP = base.push(P_PROB);
		cummulativeProbabilities = new double[n];
		for(int i=0;i<n;i++)
		{
			double val_i = parameters.getInt(p=probP.push(i+""), null, 0);
			if(val_i<0)
				throw new RuntimeException("Invalid value for "+p);			
			cummulativeProbabilities[i] = sum += val_i;
		}
		
		Parameter supP = base.push(P_SUP);
		if(parameters.exists(supP.push("0")))
		{
			supportValues = new int[n];
			for(int i=0;i<n;i++)
				supportValues[i] = parameters.getInt(supP.push(i+""), null);
		}
		numElem = n;
	}








	/**
	 * This will reduce the number of elements.
	 * It also takes O(n), cause it shifts elements to fill in the gap.
	 * (moving the last element in the gap will not buy you much, you still
	 * need to recompute the cummulative probs).
	 */
	public int getIntWithoutReplacement()
	{
		int result;
		if(uniform)
		{
			int index = rng.nextInt(numElem);
			result = index;
			if(supportValues!=null)
			{
				result = supportValues[index];
				supportValues[index] = supportValues[numElem-1];
			}
		}
		else
		{
			int index = getRandomIndex(rng.nextDouble()*cummulativeProbabilities[numElem-1], cummulativeProbabilities, 0, numElem-1);
			result = index;
			double prob = (index==0)? cummulativeProbabilities[0]:cummulativeProbabilities[index]-cummulativeProbabilities[index-1];
			for(int i=index;i<numElem-1;i++)
			{
				cummulativeProbabilities[index]=cummulativeProbabilities[index+1]-prob;
			}
			if(supportValues!=null)
			{
				result = supportValues[index];
				for(int i=index;i<numElem-1;i++)
					supportValues[index]=supportValues[index+1];
			}
		}
		numElem--;
		return result;
	}

	public int getInt()
	{
		
		int index = uniform?  
			rng.nextInt(numElem):
			getRandomIndex(rng.nextDouble()*cummulativeProbabilities[numElem-1], cummulativeProbabilities, 0, numElem-1);
		if(supportValues==null)
			return index;
		return supportValues[index];
	}


	static protected final int getRandomIndex(double prob, double[] cdf, int bottom, int top)	
	{
		int cur;
		while(top!=bottom)
		{
			cur = (top + bottom) / 2; // integer division
			if (cdf[cur] > prob)
				if (cur==0 || cdf[cur-1] <= prob)
					return exemptZeroes(cdf,cur);
				else // step down
					top = cur;
			else if (cur==cdf.length-1) // oops
				return exemptZeroes(cdf,cur);
			else if (bottom==cur) // step up
				bottom++;  // (8 + 9)/2 = 8
			else
				bottom = cur;  // (8 + 10) / 2 = 9
			}
		return exemptZeroes(cdf,bottom);  // oops		
	}
	
	/**
	 * copied from Sean Luke's RandomChoice.java
	 */
	private static final int exemptZeroes(final double[] probs, int index)
	{
		while(index > 0 && probs[index]==probs[index-1]) index--;
		return index;
	}

}
