package sim.util.jfreechart;

import java.util.*;

import org.jfree.data.*;
import org.jfree.data.xy.*;
import org.jfree.data.statistics.*;

/**
 * This was copied from DefaultBoxAndWhiskerXYDataset,
 * @see http://www.jfree.org/jfreechart/api/javadoc/src-html/org/jfree/data/statistics/DefaultBoxAndWhiskerXYDataset.html
 * but I changed the date thing on the x axis.
 * 
 * dates and items are ArrayLists not Lists, for speed.
 * 
 * <p> * I offer the option to throw away half the data points (every other one) when their number goes over
 * a specified number.
 * 
 * @author Gabriel Balan
 */
public class BoxAndWhiskerSeries extends AbstractXYDataset
										   implements BoxAndWhiskerXYDataset,
													  RangeInfo 
{
	
	/** The series name. */
	private Comparable seriesKey;
	
	/** Storage for the dates. */
	private ArrayList dates;
	
	/** Storage for the box and whisker statistics. */
	private ArrayList items;
	
	/** The minimum range value. */
	private Number minimumRangeValue;
	
	/** The maximum range value. */
	private Number maximumRangeValue;
	
	/** The range of values. */
	private Range rangeBounds;
	
	/** The coefficient used to calculate outliers. Tukey's default value is 1.5 (see EDA)
	 * Any value which is greater than Q3 + (interquartile range * outlier coefficient) is
	 * considered to be an outlier.
	 * Can be altered if the data is particularly skewed.
	 */
	private double outlierCoefficient = 1.5;
	
	/** The coefficient used to calculate farouts. Tukey's default value is 2 (see EDA)
	 * Any value which is greater than Q3 + (interquartile range * farout coefficient) is
	 * considered to be a farout.
	 *  Can be altered if the data is particularly skewed.
	 */
	private double faroutCoefficient = 2.0;
	
	public final int maxNumberPoints;
	public int modulo = 1;
	public int insertCount = 0;
	ArrayList tmpList;
	
	/**
	 * Constructs a new box and whisker dataset.
	 * <p>
	 * The current implementation allows only one series in the dataset.
	 * This may be extended in a future version.
	 *
	 * @param seriesName  the name of the series.
	 */	
	public BoxAndWhiskerSeries(Comparable seriesKey, final int maxNumberPoints) {
	
		this.seriesKey = seriesKey;
		if(maxNumberPoints<=0)
		{//unlimited # of points
			this.dates = new ArrayList();
			this.items = new ArrayList();
			tmpList = new ArrayList();
		}
		else
		{//limited number of data points
			this.dates = new ArrayList(maxNumberPoints);
			this.items = new ArrayList(maxNumberPoints);
			tmpList = new ArrayList(maxNumberPoints);
		}
		this.minimumRangeValue = null;
		this.maximumRangeValue = null;
		this.rangeBounds = null;
		this.maxNumberPoints = maxNumberPoints;

	}
	
	/**
	 * Same as the other constructor, but keep all points
	 * (no limit on the number of points).
	 */
	public BoxAndWhiskerSeries(Comparable seriesKey) 
	{
		this(seriesKey, -1);
	}

	public void add(final Number x, final BoxAndWhiskerItem item) {add(x, item, true);}

	/**
	 * Adds an item to the dataset.
	 * 
	 * @param x  the x coordinate.
	 * @param item  the item.
	 */
	public void add(final Number x, final BoxAndWhiskerItem item, boolean fireDataChanged) {
		insertCount++;
		if(insertCount%modulo!=0) return;
		int n = dates.size();
		if (n==maxNumberPoints)
		{
			modulo*=2;
			insertCount=0;
			//delete every other element.
			tmpList.clear();
			for(int i=0;i<n;i+=2)
				tmpList.add(dates.get(i));
			ArrayList switchTmp = tmpList;
			tmpList = dates;
			dates = switchTmp;
			
			tmpList.clear();
			for(int i=0;i<n;i+=2)
				tmpList.add(items.get(i));
			switchTmp = tmpList;
			tmpList = items;
			items = switchTmp;

		}
		
		this.dates.add(x);
		this.items.add(item);
		if (this.minimumRangeValue == null) {
			this.minimumRangeValue = item.getMinRegularValue();
		}
		else {
			if (item.getMinRegularValue().doubleValue() < this.minimumRangeValue.doubleValue()) {
				this.minimumRangeValue = item.getMinRegularValue();
			}
		}
		if (this.maximumRangeValue == null) {
			this.maximumRangeValue = item.getMaxRegularValue();
		}
		else {
			if (item.getMaxRegularValue().doubleValue() > this.maximumRangeValue.doubleValue()) {
				this.maximumRangeValue = item.getMaxRegularValue();
			}
		}
		this.rangeBounds = new Range(this.minimumRangeValue.doubleValue(), 
									this.maximumRangeValue.doubleValue());
		if(fireDataChanged)
			fireDatasetChanged();
	}
	
    /**
     * Returns the name of the series stored in this dataset.
     *
     * @param i  the index of the series. Currently ignored.
     * 
     * @return The name of this series.
     */
    public Comparable getSeriesKey(int i) {
        return this.seriesKey;
    }
	
	/**
	 * Returns the x-value for one item in a series.
	 * <p>
	 *
	 * @param series  the series (zero-based index).
	 * @param item  the item (zero-based index).
	 *
	 * @return the x-value.
	 */
	public Number getX(final int series, final int item) {
		return  (Number)this.dates.get(item);
	}
	
	/**
	 * Returns the y-value for one item in a series.
	 * <p>
	 * This method (from the XYDataset interface) is mapped to the getMaxNonOutlierValue(...) 
	 * method.
	 *
	 * @param series  the series (zero-based index).
	 * @param item  the item (zero-based index).
	 *
	 * @return the y-value.
	 */
	public Number getY(final int series, final int item) {
		return new Double(this.getMeanValue(series, item).doubleValue());  
	}
	
	/**
	 * Returns the mean for the specified series and item.
	 *
	 * @param series  the series (zero-based index).
	 * @param item  the item (zero-based index).
	 *
	 * @return the mean for the specified series and item.
	 */
	public Number getMeanValue(final int series, final int item) {
		Number result = null;
		BoxAndWhiskerItem stats = (BoxAndWhiskerItem) this.items.get(item);
		if (stats != null) {
			result = stats.getMean();
		}
		return result;
	}
	
	/**
	 * Returns the median-value for the specified series and item.
	 *
	 * @param series  the series (zero-based index).
	 * @param item  the item (zero-based index).
	 *
	 * @return the median-value for the specified series and item.
	 */
	public Number getMedianValue(final int series, final int item) {
		Number result = null;
		BoxAndWhiskerItem stats = (BoxAndWhiskerItem) this.items.get(item);
		if (stats != null) {
			result = stats.getMedian();
		}
		return result;
	}
	
	/**
	 * Returns the Q1 median-value for the specified series and item.
	 *
	 * @param series  the series (zero-based index).
	 * @param item  the item (zero-based index).
	 *
	 * @return the Q1 median-value for the specified series and item.
	 */
	public Number getQ1Value(final int series, final int item) {
		Number result = null;
		BoxAndWhiskerItem stats = (BoxAndWhiskerItem) this.items.get(item);
		if (stats != null) {
			result = stats.getQ1();
		}
		return result;
	}
	
	/**
	 * Returns the Q3 median-value for the specified series and item.
	 *
	 * @param series  the series (zero-based index).
	 * @param item  the item (zero-based index).
	 *
	 * @return the Q3 median-value for the specified series and item.
	 */
	public Number getQ3Value(final int series, final int item) {
		Number result = null;
		BoxAndWhiskerItem stats = (BoxAndWhiskerItem) this.items.get(item);
		if (stats != null) {
			result = stats.getQ3();
		}
		return result;
	}
	
	/**
	 * Returns the min-value for the specified series and item.
	 *
	 * @param series  the series (zero-based index).
	 * @param item  the item (zero-based index).
	 *
	 * @return the min-value for the specified series and item.
	 */
	public Number getMinRegularValue(final int series, final int item) {
		Number result = null;
		BoxAndWhiskerItem stats = (BoxAndWhiskerItem) this.items.get(item);
		if (stats != null) {
			result = stats.getMinRegularValue();
		}
		return result;
	}
	
	/**
	 * Returns the max-value for the specified series and item.
	 *
	 * @param series  the series (zero-based index).
	 * @param item  the item (zero-based index).
	 *
	 * @return the max-value for the specified series and item.
	 */
	public Number getMaxRegularValue(final int series, final int item) {
		Number result = null;
		BoxAndWhiskerItem stats = (BoxAndWhiskerItem) this.items.get(item);
		if (stats != null) {
			result = stats.getMaxRegularValue();
		}
		return result;
	}
	
	/**
	 * Returns the minimum value which is not a farout.
	 * @param series  the series (zero-based index).
	 * @param item  the item (zero-based index).
	 *
	 * @return a <code>Number</code> representing the maximum non-farout value.
	 */
	public Number getMinOutlier(final int series, final int item) {
		Number result = null;
		BoxAndWhiskerItem stats = (BoxAndWhiskerItem) this.items.get(item);
		if (stats != null) {
			result = stats.getMinOutlier();
		}
		return result;
	}
	 
															/**
	 * Returns the maximum value which is not a farout, ie Q3 + (interquartile range * farout 
	 * coefficient).
	 * 
	 * @param series  the series (zero-based index).
	 * @param item  the item (zero-based index).
	 *
	 * @return a <code>Number</code> representing the maximum non-farout value.
	 */
	public Number getMaxOutlier(final int series, final int item) {
		Number result = null;
		BoxAndWhiskerItem stats = (BoxAndWhiskerItem) this.items.get(item);
		if (stats != null) {
			result = stats.getMaxOutlier();
		}
		return result;
	}
	
	/**
	 * Returns an array of outliers for the specified series and item.
	 *
	 * @param series  the series (zero-based index).
	 * @param item  the item (zero-based index).
	 *
	 * @return the array of outliers for the specified series and item.
	 */
	public List getOutliers(final int series, final int item) {
		List result = null;
		BoxAndWhiskerItem stats = (BoxAndWhiskerItem) this.items.get(item);
		if (stats != null) {
			result = stats.getOutliers();
		}
		return result;
	}
	
	/**
	 * Returns the value used as the outlier coefficient. The outlier coefficient
	 * gives an indication of the degree of certainty in an unskewed distribution.
	 * Increasing the coefficient increases the number of values included.
	 * Currently only used to ensure farout coefficient is greater than the outlier coefficient
	 *
	 * @return a <code>double</code> representing the value used to calculate outliers
	 */
	public double getOutlierCoefficient() {
		return this.outlierCoefficient;
	}
	
	/**
	 * Returns the value used as the farout coefficient. The farout coefficient
	 * allows the calculation of which values will be off the graph.
	 *
	 * @return a <code>double</code> representing the value used to calculate farouts
	 */
	public double getFaroutCoefficient() {
		return this.faroutCoefficient;
	}
	
	/**
	 * Returns the number of series in the dataset.
	 * <p>
	 * This implementation only allows one series.
	 *
	 * @return the number of series.
	 */
	public int getSeriesCount() {
		return 1;
	}
	
	/**
	 * Returns the number of items in the specified series.
	 *
	 * @param series  the index (zero-based) of the series.
	 *
	 * @return  the number of items in the specified series.
	 */
	public int getItemCount(final int series) {
		return this.dates.size();
	}
	
	/**
	 * Sets the value used as the outlier coefficient
	 *
	 * @param outlierCoefficient being a <code>double</code> representing the value used to 
	 *        calculate outliers
	 */
	public void setOutlierCoefficient(final double outlierCoefficient) {
		this.outlierCoefficient = outlierCoefficient;
	}
	
	/**
	 * Sets the value used as the farouts coefficient. The farout coefficient must b greater than 
	 * the outlier coefficient.
	 * 
	 * @param faroutCoefficient being a <code>double</code> representing the value used to 
	 *        calculate farouts
	 * 
	 */
	public void setFaroutCoefficient(final double faroutCoefficient) {
	
		if (faroutCoefficient > this.getOutlierCoefficient()) {
			this.faroutCoefficient = faroutCoefficient;
		} 
		else {
			throw new IllegalArgumentException("Farout value must be greater " 
				+ "than the outlier value, which is currently set at: (" 
				+ getOutlierCoefficient() + ")");
		}
	}
	
    /**
     * Returns the minimum y-value in the dataset.
     *
     * @param includeInterval  a flag that determines whether or not the
     *                         y-interval is taken into account.
     * 
     * @return The minimum value.
     */
    public double getRangeLowerBound(boolean includeInterval) {
        double result = Double.NaN;
        if (this.minimumRangeValue != null) {
            result = this.minimumRangeValue.doubleValue();
        }
        return result;        
    }

    /**
     * Returns the maximum y-value in the dataset.
     *
     * @param includeInterval  a flag that determines whether or not the
     *                         y-interval is taken into account.
     * 
     * @return The maximum value.
     */
    public double getRangeUpperBound(boolean includeInterval) {
        double result = Double.NaN;
        if (this.maximumRangeValue != null) {
            result = this.maximumRangeValue.doubleValue();
        }
        return result;        
    }

    /**
     * Returns the range of the values in this dataset's range.
     *
     * @param includeInterval  a flag that determines whether or not the
     *                         y-interval is taken into account.
     * 
     * @return The range.
     */
    public Range getRangeBounds(boolean includeInterval) {
        return this.rangeBounds;
    }

}

