package sim.util.jfreechart;

import java.io.Serializable;
import java.util.*;

import org.jfree.data.general.DatasetChangeEvent;
import org.jfree.data.xy.*;
import org.jfree.util.ObjectUtilities;
import org.jfree.data.statistics.*;

import sim.util.DoubleBag;
/**
 * @author Gabriel Balan
 * This is a copy of HistogramDataset, plus the clear() method.
 * All because their data is private, so extending it would not work.
 * 
 * I also needed to call fireDatasetChanged() after addSeries.
 * (extending HistogramDataset would have been the answer here too).
 * 
 * I changed addSeries so I can use it with values in a DoubleBag. 
 * Could not do this with just extending.
 * 
 */
public class ClearableHistogramDataset  extends AbstractIntervalXYDataset 
										implements IntervalXYDataset, Cloneable, Serializable
{
	/** A list of maps. */
	  private List list;
    
	  /** The histogram type. */
	  private HistogramType type;

	  /**
	   * Creates a new (empty) dataset with a default type of {@link HistogramType}.FREQUENCY.
	   */
	  public ClearableHistogramDataset() {
		  this.list = new ArrayList();
		  this.type = HistogramType.FREQUENCY;
	  }
    
	  /**
	   * Returns the histogram type. 
	   * 
	   * @return the type (never <code>null</code>).
	   */
	  public HistogramType getType() { 
		  return this.type; 
	  }

	  /**
	   * Sets the histogram type and sends a {@link DatasetChangeEvent} to all registered
	   * listeners.
	   * 
	   * @param type  the type (<code>null</code> not permitted).
	   */
	  public void setType(final HistogramType type) {
		  if (type == null) {
			  throw new IllegalArgumentException("Null 'type' argument");
		  }
		  this.type = type;   
		  notifyListeners(new DatasetChangeEvent(this, this));
	  }

	  /**
	   * Adds a series to the dataset, using the specified number of bins.
	   * 
	   * @param name  the series name (<code>null</code> not permitted).
	   * @param values the values (<code>null</code> not permitted).
	   * @param bins  the number of bins (must be at least 1).
	   */
	  public void addSeries(final String name, final double[] values, final int bins) {
		  // defer argument checking...
		  final double minimum = getMinimum(values);
		  final double maximum = getMaximum(values);
		  addSeries(name, values, bins, minimum, maximum);
	  }


	  /**
	   * Adds a series to the dataset. Any data value falling on a bin boundary will be assigned to
	   * the lower value bin, with the exception of the lower bound of the bin range which is always
	   * assigned to the first bin.
	   * 
	   * @param name  the series name (<code>null</code> not permitted).
	   * @param values  the raw observations.
	   * @param bins  the number of bins.
	   * @param minimum  the lower bound of the bin range.
	   * @param maximum  the upper bound of the bin range.
	   */
	  public void addSeries(final String name, 
							final double[] values, 
							final int bins, 
							final double minimum, 
							final double maximum) 
	{addSeries(name, values, values.length, bins, minimum, maximum);}

	
	public void addSeries(final String name, 
						  final DoubleBag values, 
						  final int bins, 
						  final double minimum, 
						  final double maximum) 
	{addSeries(name, values.objs, values.numObjs, bins, minimum, maximum);}
	

	
	/**
	 * Adds a series to the dataset. Any data value falling on a bin boundary will be assigned to
	 * the lower value bin, with the exception of the lower bound of the bin range which is always
	 * assigned to the first bin.
	 * 
	 * @param name  the series name (<code>null</code> not permitted).
	 * @param values  the raw observations.
	 * @param valueCount the number of observations.
	 * @param bins  the number of bins.
	 * @param minimum  the lower bound of the bin range.
	 * @param maximum  the upper bound of the bin range.
	 */							
	public void addSeries(final String name, 
						  final double[] values,
						  final int valueCount,
						  final int bins, 
						  final double minimum, 
						  final double maximum) {

        
		  if (name == null) {
			  throw new IllegalArgumentException("Null 'name' argument.");   
		  }
		  if (values == null) {
			  throw new IllegalArgumentException("Null 'values' argument.");
		  }
		  if(values.length < valueCount)
		  if (bins < 1) {
			  throw new IllegalArgumentException("Fewer values than you thought.");
		  }
		  double binWidth = (maximum - minimum) / bins;	  

		  double tmp = minimum;
		  List binList = new ArrayList(bins);
		  for (int i = 0; i < bins; i++) {
			  HistogramBin bin;
			  // make sure bins[bins.length]'s upper boundary ends at maximum
			  // to avoid the rounding issue. the bins[0] lower boundary is
			  // guaranteed start from min
			  if (i == bins - 1) {
				  bin = new HistogramBin(tmp, maximum);
			  }
			  else {
				  bin = new HistogramBin(tmp, tmp + binWidth);
			  }
			  tmp = tmp + binWidth;
			  binList.add(bin);
		  }
		  		  
		  // fill the bins
		  int lastBinIndex = 0;
		  for (int i = 0; i < valueCount; i++) 
		  {
			double val_i = values[i];
		  	 
			int binIndex = bins - 1;
			if (values[i] < maximum) 
			{
				double fraction = (values[i] - minimum) / (maximum - minimum);
				binIndex = (int) (fraction * bins);
			}
			HistogramBin bin = (HistogramBin) binList.get(binIndex);
			assert(val_i<= bin.getEndBoundary() || val_i >= bin.getStartBoundary());
			bin.incrementCount();
//			System.out.println(i+":\t"+val_i+"\tbin="+binIndex+"\tBinCount:"+bin.getCount());
		  }
		  // generic map for each series
		  final Map map = new HashMap();
		  map.put("name", name);
		  map.put("bins", binList);
		  //TODO what should happen to "values.length" now that I use valueCount?
		  
		  map.put("values.length", new Integer(valueCount));
		  map.put("bin width", new Double(binWidth));
		  this.list.add(map);
		  fireDatasetChanged();
	  }
    
	  /**
	   * Returns the minimum value in an array of values.
	   * 
	   * @param values  the values (<code>null</code> not permitted and zero-length array 
	   *                not permitted).
	   * 
	   * @return  the minimum value.
	   */
	  private double getMinimum(final double[] values) {
		  if (values == null || values.length < 1) {
			  throw new IllegalArgumentException("Null or zero length 'values' argument.");
		  }
		  double min = Double.MAX_VALUE;
		  for (int i = 0; i < values.length; i++) {
			  if (values[i] < min) {
				  min = values[i];
			  }
		  }
		  return min;
	  }

	  /**
	   * Returns the maximum value in an array of values.
	   * 
	   * @param values  the values (<code>null</code> not permitted and zero-length array 
	   *                not permitted).
	   * 
	   * @return the maximum value.
	   */
	  private double getMaximum(final double[] values) {
		  if (values == null || values.length < 1) {
			  throw new IllegalArgumentException("Null or zero length 'values' argument.");
		  }
		  double max = -Double.MAX_VALUE;
		  for (int i = 0; i < values.length; i++) {
			  if (values[i] > max) {
				  max = values[i];
			  }
		  }
		  return max;
	  }

	  /**
	   * Returns the bins for a series.
	   * 
	   * @param series  the series index.
	   * 
	   * @return An array of bins.
	   */
	 public  List getBins(final int series) {
		  Map map = (Map) this.list.get(series);
		  return (List) map.get("bins"); 
	  }

	  /**
	   * Returns the total number of observations for a series.
	   * 
	   * @param series  the series index.
	   * 
	   * @return the total.
	   */
	  private int getTotal(final int series) {
		  Map map = (Map) this.list.get(series);
		  return ((Integer) map.get("values.length")).intValue(); 
	  }

	  /**
	   * Returns the bin width for a series.
	   * 
	   * @param series  the series index (zero based).
	   * 
	   * @return the bin width.
	   */
	  private double getBinWidth(final int series) {
		  Map map = (Map) this.list.get(series);
		  return ((Double) map.get("bin width")).doubleValue(); 
	  }

	  /**
	   * Returns the number of series in the dataset.
	   * 
	   * @return The series count.
	   */
	  public int getSeriesCount() { 
		  return this.list.size(); 
	  }
    
	  /**
	   * Returns the name for a series.
	   * 
	   * @param series  the series index (zero based).
	   * 
	   * @return The series name.
	   */
	  public String getSeriesName(final int series) {
		  final Map map = (Map) this.list.get(series);
		  return (String) map.get("name"); 
	  }
	    /**
	     * Returns the key for a series.
	     * 
	     * @param series  the series index (zero based).
	     * 
	     * @return The series key.
	     */
	    public Comparable getSeriesKey(int series) {
	        Map map = (Map) this.list.get(series);
	        return (String) map.get("name"); 
	    }


	  /**
	   * Returns the number of data items for a series.
	   * 
	   * @param series  the series index (zero based).
	   * 
	   * @return the item count.
	   */
	  public int getItemCount(final int series) {
		  return getBins(series).size(); 
	  }

	  /**
	   * Returns the X value for a bin. 
	   * <p>
	   * This value won't be used for plotting histograms, since the renderer will ignore it.
	   * But other renderers can use it (for example, you could use the dataset to create a line
	   * chart).
	   * 
	   * @param series  the series index (zero based).
	   * @param item  the item index (zero based).
	   * 
	   * @return The start value.
	   */
	  public Number getX(int series, int item) {
		  List bins = getBins(series);
		  HistogramBin bin = (HistogramBin) bins.get(item);
		  double x = (bin.getStartBoundary() + bin.getEndBoundary()) / 2.;
		  return new Double(x);
	  }

	  /**
	   * Returns the y-value for a bin (calculated to take into account the histogram type).
	   * 
	   * @param series  the series index (zero based).
	   * @param item  the item index (zero based).
	   * 
	   * @return The y-value.
	   */
	  public Number getY(int series, int item) {
		  List bins = getBins(series);
		  HistogramBin bin = (HistogramBin) bins.get(item);
		  double total = getTotal(series);
		  double binWidth = getBinWidth(series);

		  if (this.type == HistogramType.FREQUENCY) {
			  return new Double(bin.getCount());
		  }
		  else if (this.type == HistogramType.RELATIVE_FREQUENCY) {
			  return new Double(bin.getCount() / total);
		  }
		  else if (this.type == HistogramType.SCALE_AREA_TO_1) {
			  return new Double(bin.getCount() / (binWidth * total));
		  }
		  else { // pretty sure this shouldn't ever happen
			  throw new IllegalStateException();
		  }
	  }

	  /**
	   * Returns the start value for a bin.
	   * 
	   * @param series  the series index (zero based).
	   * @param item  the item index (zero based).
	   * 
	   * @return The start value.
	   */
	  public Number getStartX(int series, int item) {
		  final List bins = getBins(series);
		  final HistogramBin bin = (HistogramBin) bins.get(item);
		  return new Double(bin.getStartBoundary());
	  }

	  /**
	   * Returns the end value for a bin.
	   * 
	   * @param series  the series index (zero based).
	   * @param item  the item index (zero based).
	   * 
	   * @return The end value.
	   */
	  public Number getEndX(int series, int item) {
		  List bins = getBins(series);
		  HistogramBin bin = (HistogramBin) bins.get(item);
		  return new Double(bin.getEndBoundary());
	  }

	  /**
	   * Returns the start y-value for a bin (which is the same as the y-value, this method 
	   * exists only to support the general form of the {@link IntervalXYDataset} interface).
	   * 
	   * @param series  the series index (zero based).
	   * @param item  the item index (zero based).
	   * 
	   * @return The y-value.
	   */
	  public Number getStartY(int series, int item) {
		  return getY(series, item);
	  }

	  /**
	   * Returns the end y-value for a bin (which is the same as the y-value, this method 
	   * exists only to support the general form of the {@link IntervalXYDataset} interface).
	   * 
	   * @param series  the series index (zero based).
	   * @param item  the item index (zero based).
	   * 
	   * @return The Y value.
	   */    
	  public Number getEndY(int series, int item) {
		  return getY(series, item);
	  }

	  /**
	   * Tests this dataset for equality with an arbitrary object.
	   * 
	   * @param obj  the object to test against (<code>null</code> permitted).
	   * 
	   * @return A boolean.
	   */
	  public boolean equals(final Object obj) {
        
		  if (obj == null) {
			  return false;   
		  }
        
		  if (obj == this) {
			  return true;   
		  }
        
		  if (obj instanceof ClearableHistogramDataset) {
			  final ClearableHistogramDataset dataset = (ClearableHistogramDataset) obj;
			  final boolean b0 = ObjectUtilities.equal(dataset.type, this.type);
			  final boolean b1 = ObjectUtilities.equal(dataset.list, this.list);
			  return b0 && b1;
		  }
        
		  return false;
        
	  }

	  /**
	   * Returns a clone of the dataset.
	   * 
	   * @return A clone of the dataset.
	   * 
	   * @throws CloneNotSupportedException if the object cannot be cloned.
	   */
	  public Object clone() throws CloneNotSupportedException {
		  return super.clone();   
	  }

		public void clear(){list.clear();}
		public void clearAndReport(){list.clear();fireDatasetChanged();}
}
