package sim.util.stats;
import umontreal.iro.lecuyer.stat.*;
import umontreal.iro.lecuyer.probdist.*;
import cern.colt.list.DoubleArrayList;
/**
 * 
 * <p>It offers quantile capabilities. So if you have the data already sorted, feed it in in ASCENDING order!!!
 * 
 * <p>In the "confidence interval of the median", the ranks are
 * already sorted, but I got to have something more general
 * without compromizing the good time-complexity of the 
 * best-case (yet the only one I need).
 *
 * <p>
 * <font color="red">
 * I am using this instead of TallyInABag cause the goodness-of-fit static methods (in umontreal.iro.lecuyer.gof.GofStat),
 * such as: <code>andersonDarling</code> and <code>kolmogorovSmirnov</code> all use TallyInArray.
 * </font>
 * 
 * <p>TallyInArray is now TallyStore.
 * 
 * @author Gabriel Balan
 */
public class MyTallyInArray extends TallyStore
{
	boolean sorted = true;//ASCENDING
	
	public MyTallyInArray(){super();}
	public MyTallyInArray(int initialCapacity){super(initialCapacity);}	
	public MyTallyInArray(DoubleArrayList dal)
	{
		super(dal);
		if(dal.size()>1)sorted=false;
	}
	public String toString(){return name;}
	public void init()
	{
		super.init();
		sorted=true;
	}

	double last;
	public void add(double d)
	{
		if(sorted && numberObs()>0) sorted = (last<=d);
		//I don't want to do insertion sort cause that has a quadratic worst case complexity. 
		super.add(last = d);	
	}
	
	public void sort()
	{
		if(sorted) return;
		sorted=true;
		getArray().sort();
	}
	
	// If u is not a multiple of 1/(n - 1), PERCENTILE interpolates to determine the value at the kth percentile.
	public double quantile(double u)
	{
		DoubleArrayList data = getArray();
		if(!sorted) data.sort();
		if(u<0 || u>1) throw new RuntimeException("Invalid quantile "+u);
		u*=(numberObs()-1);
		int index1 = (int)u;
		double w1 = u - index1;//frac{u}
		if(w1 ==0) return data.getQuick(index1);
		int index2 = index1+1;
		double w2 = (w1==0)? 0 : index2 - u;
		return data.getQuick(index1)*w2 + w1 * data.getQuick(index2);
	}	

	
	
	/**
	 * Thomson-Savur: confidence interval around the median.
	 * 
	 * According to Wineberg & Christensen (GECCO 03).
	 * binomial quantile /w cumulatice upper
	 * tail = alpha/2		b = excel's CRITBINOM(n, 0.5, alpha/2)
	 * lower percentile		l = b/(n-1) 
	 * upper percentile 	u = 1-l
	 * lower bound			value_l = excel's PERCENTILE(data, l)
	 * upper bound			value_u = excel's PERCENTILE(data, u)
	 * 
	 * Advantages:		It gives a better understandind of the system
	 * 						-median  /w error bounds
	 * 						-easy to draw on a graph.
	 * (my own:				-I can compare more than 2 treatments without n^2/2 re-rankings, like the ranked T-test)
	 * 
	 * Disadvantage:	Not as sensitive as a ranked T test
	 */
	public void  confidenceIntervalMedian(double alpha, double[] centerAndRadius)
	{
		DoubleArrayList data = getArray();
		int n  = data.size();
		double b = BinomialDist.inverseF(n, .5, alpha/2);

//		double l = b/(n-1);
//		double u = 1-l;
//		double value_l = quantile(l);
//		double value_u = quantile(u);
//		double center = centerAndRadius[0]=(value_l+value_u)/2;
//		centerAndRadius[1] = center - value_l;
		
		//the quantile of a multiple of 1/(n-1) is one of the sample values (no interpolation)
		//so here is the same thing, but faster.
		if(!sorted) data.sort();
		int b_fast = (int)b;
		double value_l_fast = data.getQuick(b_fast);
		double value_u_fast = data.getQuick(n-1-b_fast);
		double center = centerAndRadius[0]=(value_l_fast+value_u_fast)/2;
		centerAndRadius[1] = center - value_l_fast;
	}

	/**
	 * Tally.confidenceIntervalStudent uses the confidence coefficient!
	 * TallyInABag.confidenceIntervalMean uses the significance level
	 * 
	 * This function borrows the meaning from EXCEL:
	 * 			"Alpha is the significance level used to compute the confidence level. 
	 * 			The confidence level equals 100*(1 - alpha)%, or in other words, 
	 * 			an alpha of 0.05 indicates a 95 percent confidence level."

	 * @param alpha = significance level
	 * 1-alpha = confidence coefficient
	 * 100(1-alpha) = confidence level
	 * @param centerAndRadius
	 */
	public void  confidenceIntervalMean(double alpha, double[] centerAndRadius)
	{
		centerAndRadius[0]=average();
		int n  = numberObs();
		centerAndRadius[1] = StudentDist.inverseF(n - 1, 1- 0.5* alpha) * 
								Math.sqrt(variance() / (double)n);
	}


	public static MyTallyInArray pairwiseMax(MyTallyInArray t1, MyTallyInArray t2)
	{
		int n = t1.numberObs(); 
		assert(n ==t2.numberObs());
		MyTallyInArray t = new MyTallyInArray(n);
		double[] t1l = t1.getArray().elements();
		double[] t2l = t2.getArray().elements();
		for(int i=0;i<n;i++)
			t.add(Math.max(t1l[i], t2l[i]));
		return t;
	}

	public void print(double confidenceLevel)
	{
		System.out.println("\nStats for "+getName());
		System.out.println("SampleSize...\t"+numberObs());
		System.out.println("Average......\t"+average());
		if(numberObs()>=2)
			System.out.println("Stdev........\t"+standardDeviation());
		System.out.println("Min..........\t"+minValue);
		System.out.println("Max..........\t"+maxValue);		
		double[] centerAndRadius = new double[2];
		if(numberObs()>=2)
		{
			confidenceIntervalMean(confidenceLevel, centerAndRadius);
			System.out.println(confidenceLevel+"-Confidence interval for the mean:\t" + (centerAndRadius[0]-centerAndRadius[1]) + "\t" + (centerAndRadius[0]+centerAndRadius[1]));
			confidenceIntervalMedian(confidenceLevel, centerAndRadius);
			System.out.println(confidenceLevel+"-Confidence interval for the median:\t" + (centerAndRadius[0]-centerAndRadius[1]) + "\t" + (centerAndRadius[0]+centerAndRadius[1]));	
//			System.out.println("Median.......\t\t"+quantile(0.5));
//			System.out.println("Median-CI-center\t"+centerAndRadius[0]);
		}
	}
	public static final void main(String[] args)
	{
//		//instead of CRITBINOM
//		//the static call is doing linear search
//		//the other one does binary search.
//		//although the treatments have usually the same size, 
//		//unfortunatelly the code is more general. On the other hand, 
//		//speed is not a issue, even in the online case (racing).
//		BinomialDist bd = new BinomialDist(10, 1d/2);
//		double p = 0;
//		for(int i=0;i<10;i++)
//		{
//			System.out.println(p+"\t"+bd.inverseF(p)+"\t"+ BinomialDist.inverseF(10, .5, p));
//			p+=0.1;
//		}
		MyTallyInArray t = new MyTallyInArray();
		ec.util.MersenneTwisterFast rng = new ec.util.MersenneTwisterFast();
		for(int i=0;i<100;i++)
			t.add(rng.nextDouble());
		t.confidenceIntervalMedian(0.05, new double[2]);
	}
}
