package sim.util.stats;

import umontreal.iro.lecuyer.probdist.*;
import umontreal.iro.lecuyer.stat.*;
/**
 * @author Catalin Balan
 *
 * I am using some ideas from Bryan Lewis' code
 * @see http://nimitz.mcs.kent.edu/~blewis/stat/tTest.html 
 * btw, there is a bug in his T-test formulas for unequal variances
 */
public /*strictfp*/ class StatisticUtilities
{
	static public boolean tTest(Tally treatmentA, Tally treatmentB, double alpha)
	{
		return tTest(treatmentA, treatmentB, alpha, true, false);
	}

	static public boolean tTest(Tally treatmentA, Tally treatmentB, double alpha, boolean verbosity)
	{
		return tTest(treatmentA, treatmentB, alpha, true, verbosity);
	}
	static public boolean tTest(Tally treatmentA, Tally treatmentB, double alpha, boolean doFTest, boolean verbosity)
	{
		return alpha>tTestCriticalValue(treatmentA, treatmentB, doFTest, verbosity);
	}

	/**
	 * @return true if the two treatments are significantly different (with confidence 1-alpha).
	 * if it returns false, the meaning is not "they're the same", but "I cannot say that they're different".
	 */
	static public double tTestCriticalValue(Tally treatmentA, Tally treatmentB, boolean doFTest, boolean verbosity)
	{
		int nA = treatmentA.numberObs();
		int nB = treatmentB.numberObs();
		double meanA = treatmentA.average();
		double meanB = treatmentB.average();
		double varA = treatmentA.variance();
		double varB = treatmentB.variance();
		double varOverN_A = varA/nA;
		double varOverN_B = varB/nB;
		
		double fAlpha =0;
		//if !doFTest, that fAlpha=0 will take you right to the "unequal variances" branch
		// which is the default behavior.
		if(doFTest)
			fAlpha = (varA>varB)?	fDist(nA - 1, nB - 1, varA/varB):
									fDist(nB - 1, nA - 1, varB/varA);
		
		double tScore;
		int df;
		if (fAlpha <= 0.005) 
		{
			/* 
			 * Unequal variances: Assumption:
			 * 1. The samples (n1 and n2) from two normal populations are independent
			 * 2. One or both sample sizes are less than 30
			 * 3. The appropriate sampling distribution of the test statistic is the t distribution
			 * 4. The unknown variances of the two populations are not equal
 
			 * @see http://projectile.is.cs.cmu.edu/research/public/talks/t-test.htm#unequal
			 * @see /http://bmj.bmjjournals.com/collections/statsbk/7.shtml
			 */
			df = (int)Math.round(Math.pow(varOverN_A+varOverN_B, 2) /
				(Math.pow(varOverN_A, 2)/(nA - 1)	+ Math.pow(varOverN_B, 2)/(nB - 1)));
			//standard error of the difference between the means = sqrt(stdev1^2/n1+stdev2^2/n2).
			double se = Math.sqrt(varOverN_A+varOverN_B);
			tScore = Math.abs(meanA - meanB) / se;
		}
		else 
		{
			/* 
			 * Equal variances
			 * 
			 * @see http://projectile.is.cs.cmu.edu/research/public/talks/t-test.htm#equalvariance
			 * @see /http://bmj.bmjjournals.com/collections/statsbk/7.shtml
			 */
		  df = 	nA + nB - 2;
		  /*
		   * To compute the two-sample t-test two major computations are needed before computing
		   * the t-test. First, you need to estimate the pooled standard deviation of the two 
		   * samples. The pooled standard deviation gives an weighted average of the standard
		   * deviations of the two samples. The  pooled standard deviation is going to be between
		   * the two standard deviations, with greater weight given to the standard deviation from
		   * a larger sample.
		   */
		  double sp = Math.sqrt( ((nA - 1)*varA + (nB - 1)*varB) / df );
		  tScore = Math.abs(meanA - meanB) * Math.sqrt(nA*nB/(nA+nB)) / sp;
		}
//		double pValue = StudentDist.cdf(df, -tScore);  //1 tailed.
		double pValue = StudentDist.cdf(df, -tScore)*2;//2 tailed.
		//The good news is that this is the exact pValue ANOVA would have
		//returned for these two samples.

		
		if(verbosity)
		{
			System.out.println();
			System.out.println("\t\t"+treatmentA.getName()+"\t"+treatmentB.getName());
			System.out.println("n\t\t"+nA+"\t"+nB);
			System.out.println("Avg\t\t"+meanA+"\t"+meanB);
			System.out.println("Var\t\t"+varA+"\t"+varB);
			System.out.println("StDev\t"+Math.sqrt(varA)+"\t"+Math.sqrt(varB));
			System.out.println("\ndf\t\t"+df);
			System.out.println("t-value\t"+tScore);
		// TODO 1-tailed???? I want two tailed right?
		// It seems that TDIST(tStat, df, 1) * 2 = TDIST(tStat, df, 2)
		// i.e. 1tail *2 = 2tail. 
			if(doFTest)
			{
				System.out.println("fAlpha\t"+fAlpha);
				if (fAlpha <= 0.005) 
				{
					System.out.println("An F test on the sample variances indicates that they are ");
					System.out.println("probably not from the same population (the variances ");
					System.out.println("can't be pooled), at an alpha level of " + fAlpha + ".");
					System.out.println("Thus, the t-test was set up for samples with unequal variances. ");
					System.out.println("(The degrees of freedom were adjusted.)");
				}
				else
				{
					System.out.println("An F test on the sample variances indicates that they could be ");
					System.out.println("from the same population, (alpha level of 0.005).");
					System.out.println("Accordingly, the t-test was set up for samples with equal population variance.");
				}
				System.out.println();
			}
			System.out.println("2-tailed pValue\t"+((pValue <= 0.0001)?"< 0.0001":""+pValue));
			if (pValue <= 0.01)  
				System.out.println("This probability indicates that there is a difference in sample means.");
			else if (pValue <= 0.05) 
					System.out.println("This probability indicates that there may be a difference in sample means.");
				else 
				{
					System.out.println("There is not a significant difference in the sample means.");
					System.out.println("A difference could not be detected due to large variability, small sample size,");
					System.out.println("or both. Of course, it's possible that the samples really are from the same population!");
				}
		}
		//TODO I should return the 1tailed pValue for more generality
		// and somebody else should do alpha>2*XX?
		return pValue;
	}
		
		
		
/*
 * @see http://www.itl.nist.gov/div898/handbook/eda/section3/eda359.htm
 * An F-test ( Snedecor and Cochran, 1983) is used to test if the standard deviations of two populations are equal. 
 */
 static public double fDist(double v1, double v2, double f) {
	 /* 	F distribution with v1, v2 deg. freedom P(x>f) */
	 return BetaDist.cdf(v1/2, v2/2, 8 ,v1/(v1 + v2*f));
 }

 static public double fDist(double v1, double v2, double f, int d) {
	 /* 	F distribution with v1, v2 deg. freedom P(x>f) */
	 return BetaDist.cdf(v1/2, v2/2, d ,v1/(v1 + v2*f));
 }
	//SSJ: 
	// GofStat:  chi2 kolmogorovSmirnov
	public static final void main(String[] args)
	{
		
//		StudentDist sd = new StudentDist(10);
//		for(int i=1;i<10;i++)
//		{
//			double t = i*.5;
//			System.out.println(t+"\t"+(1-sd.cdf(t))+"\t"+sd.cdf(-t));
//		} 	

//		//P(T<=t), 1 tail
//		System.out.println(StudentDist.cdf(18, -2.234014381));
//		
//		//P(T<=t), 2 tails
//		System.out.println(2*StudentDist.cdf(18, -2.234014381));
//		//P(T<=t)_1tail x2 =P(T<=t)_2tails
//		//because the first thing is cdf(-t), the prob area of the
//		//left forbiden area. The right forbiden area is equal to the left one, 
//
//		//t critical 1tail.
//		System.out.println(StudentDist.inverseF(18, 0.05));
//		//t critical, 2tails.
//		System.out.println(StudentDist.inverseF(18, 1-0.05/2));		

//		System.out.println("lewis\t"+SpecialFunctions.fDist(9, 9, 28.39166667/27.89166667));
//		System.out.println("ssj 5\t"+fDist(9, 9, 28.39166667/27.89166667, 5));
//		System.out.println("ssj 8\t"+fDist(9, 9, 28.39166667/27.89166667, 8));
//		System.out.println("ssj 10\t"+fDist(9, 9, 28.39166667/27.89166667, 10));
//		System.out.println("ssj 12\t"+fDist(9, 9, 28.39166667/27.89166667, 12));

//		MyTallyInArray x = new MyTallyInArray();
//		for(int i=1;i<=10;i++) x.add(i);
//		MyTallyInArray y = new MyTallyInArray();
//		y.add(.092);
//		y.add(.134);y.add(.165);y.add(.211);
//		y.add(.242);y.add(.302);y.add(.357);
//		y.add(.401);y.add(.405);y.add(.442);
//		double[] eq = new double[2];
//		System.out.println("R^2 " + linearRegression(x, y, eq));
//		System.out.println("Y="+eq[0]+"X+"+eq[1]);

//		MyTallyInArray th_50 = getNormalDistQuantile(50);
//		MyTallyInArray th_500 = getNormalDistQuantile(500);
//
//		NormalGen gen = new NormalGen(new umontreal.iro.lecuyer.rng.RandMrg(), new NormalDist(0.4,2.9));
//		MyTallyInArray t = new MyTallyInArray(500);	
//
//
//		Tally t_50 = new Tally("50samples"), t_500 = new Tally("500samples");
//		for(int k=0;k<100;k++)
//		{
//			t.init();
//			for(int i=0;i<500;i++) t.add(gen.nextDouble());
//			t.sort();
//			t_500.add(linearRegression(t, th_500, null));
//			
//			t.init();
//			for(int i=0;i<50;i++) t.add(gen.nextDouble());
//			t.sort();
//			t_50.add(linearRegression(t, th_50, null));
//		}
//		System.out.println(t_50.reportAndConfidenceIntervalStudent(.05));
//		System.out.println(t_500.reportAndConfidenceIntervalStudent(.05));

//		for(int df=2;df<10;df++)
//		{
//			for(int f=2;f<12;f++)
//				System.out.print(fDist(df, df, f)+"\t");
//			System.out.println();
//		}

		System.out.println(fDist(19, 19, 4.880155833));
	}
	
	/**
	 * Use this for QQplots
	 * @return R^2
	 * 
	 * @param slopeAndInterception  if it's not null, then 
	 * 	slopeAndInterception[0]<- slope
	 * 	slopeAndInterception[1]<- intercept
	 * 
	 * Y = X * slope + intercept.
	 * 
	 * SST (sum of square total) = sum (Yi- AVG_Y)^2 = VARIANCE_Y * n-1.
	 */
	public static double linearRegression(	MyTallyInArray x, 
											MyTallyInArray y,
											double[] slopeAndInterception)
	{
		int n = y.numberObs();
		if(n!=x.numberObs()) throw new RuntimeException("X and Y tallies have different sizes.");
		x.sort();
		y.sort();
		double[] dataX = x.getArray().elements();
		double[] dataY = y.getArray().elements();
		double meanX = x.average();
		double meanY = y.average();
		double sumXY=0;
		double sumXX=0;
		for(int i=0;i<n;i++)
		{
			double x_i = dataX[i];
			double y_i = dataY[i];
			sumXY +=x_i*y_i;
			sumXX +=x_i*x_i;
		}
		double b1 = (sumXY-n*meanX*meanY)/(sumXX-n*meanX*meanX);	//slope
		double b0 = meanY-b1*meanX;									//intercept
		if(slopeAndInterception!=null)
		{
			slopeAndInterception[0] = b1;//I know, I know, don't start
			slopeAndInterception[1] = b0;
		}
		double SSE = 0;
		for(int i=0;i<n;i++)
		{
			double x_i = dataX[i];
			double y_i = dataY[i];
			double err = y_i-b0-b1*x_i;
			SSE += err*err;
		}
		double SST = y.variance()*(n-1);
		double SSR = SST-SSE;
		return SSR/SST;		
	}
	

	/**
	 * Use this for a theoretical QQplot against a sample of size n.
	 * 
	 * Speaking of R^2 in the theoretical QQplot as a measure to assess 
	 * the normality of a distribution:
	 * 
	 * I generated 50 (500) samples from a true normal distribution
	 * Here's the R^2 stas over 100 repetitions:
	 * 
	 * REPORT on Tally stat. collector ==> 50samples 
	 * (R^2 of a 50 point QQplot)
	 * min        max      average    standard dev.  num. obs
	 * 0.924       0.993      0.976        0.012          100
	 * 5.0% confidence interval for mean: (     0.976,     0.976 )
	 * 
	 * REPORT on Tally stat. collector ==> 500samples 
	 * (R^2 of a 500 point QQplot)
	 * min        max      average    standard dev.  num. obs
	 * 0.990       0.999      0.997       1.6E-3          100
	 * 5.0% confidence interval for mean: (     0.997,     0.997 )
  	 * 
	 * My point is that a sample of 50 points coming from a perfect 
	 * target distribution "might not always score a perfect 1" (for R^2). 
	 * 
	 * That's good to know when assessing the normality of the empirical data.
	 * 
	 */
	static MyTallyInArray getNormalDistQuantile(int n)
	{
		MyTallyInArray normQ = new MyTallyInArray(n);

		for(int i=1;i<=n;i++)
		{
			normQ.add(NormalDist.inverseF01( (i-0.5)/n));
		}
		return normQ;
	}
} 
